﻿<#
.SYNOPSIS
  Vadlidate an XML file against XSD schema.
.DESCRIPTION
  Check the validity of an XML file against an XSD schema file, throw exception if not valid
.EXAMPLE
  ValidateXmlAgainstXsd .\HPCIaaSClusterConfig.xml .\HPCIaaSClusterConfig.xsd
#>
function ValidateXmlAgainstXsd
{
    Param
    (
        # The xml file path
        [Parameter(Mandatory=$true)]
        [String] $XmlFile,
		
        # The schema file path
        [Parameter(Mandatory=$true)]
        [String] $SchemaFile		
    )
	

    [string[]] $Script:XmlValidationErrorLog = @()
    [scriptblock] $xmlValidationEventHandler = {
        $Script:XmlValidationErrorLog += $args[1].Exception.Message
    }

    $xml = New-Object System.Xml.XmlDocument
    $schemaReader = New-Object System.Xml.XmlTextReader $SchemaFile
    try
    {
        $schema = [System.Xml.Schema.XmlSchema]::Read($schemaReader, $xmlValidationEventHandler)
        $xml.Schemas.Add($schema) | Out-Null
        $xml.Load($XmlFile)
        $xml.Validate($xmlValidationEventHandler)
    }
    catch
    {
        throw
    }
    finally
    {
        $schemaReader.Dispose()
    }

    if($Script:XmlValidationErrorLog.Length -gt 0)
    {
        throw ($Script:XmlValidationErrorLog -join "`r`n")
    }
}

function ValidatePasswordComplexity
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Password
    )

    if($Password.Length -lt 8)
    {
        throw "The password must contain at least 8 characters"
    }

    $chCategories = 0
    if($Password -cmatch "[^0-9a-zA-Z]")
    {
        $chCategories++
    }

    if($Password -cmatch "[0-9]")
    {
        $chCategories++
    }

    if($Password -cmatch "[a-z]")
    {
        $chCategories++
    }

    if($Password -cmatch "[A-Z]")
    {
        $chCategories++
    }

    if($chCategories -lt 3)
    {
        throw "The password must contain 3 of the following: `n a lowercase character `n an uppercase character `n a number `n a special character"
    }
}

function ValidateUserName
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $UserName
    )

    $valid = $false
    do
    {
        if($UserName -match "^admin[1-2]{0,1}$")
        {
            break
        }

        if($UserName -match "^user[1-5]{0,1}$")
        {
            break
        }

        if($UserName -eq "a")
        {
            break
        }

        if($UserName -eq "1")
        {
            break
        }

        if($UserName -match '[\s, \\, /, \[, \], :, |, <, >, +, =, ;, \,, ?, *, @]')
        {
            break
        }

        $valid = $true
    } while ($false)

    if(-not $valid)
    {
        throw "The user name $UserName is not allowed. Please specify another one"
    }
}

function ValidateVNetNaming
{
    Param
    (
        # The VNet Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        # The Subnet Name
        [Parameter(Mandatory=$true)]
        [String] $SubnetName
    )

    $regex = "^[^0-9].{1,62}$"
    if($VNetName -notmatch $regex)
    {
        throw "The virtual network name $VNetName is invalid. It must contain between 2 and 63 characters, and cannot start with a number."
    }
    if($SubnetName -notmatch $regex)
    {
        throw "The subnet Name $SubnetName is invalid. It must contain between 2 and 63 characters, and cannot start with a number."
    }
}

function ValidateServiceNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceName
    )
    
    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z-]{1,63}$"
        if($ServiceName -cnotmatch $regex)
        {
            break;
        }
        if($ServiceName[0] -eq '-')
        {
            break;
        }
        if($ServiceName[-1] -eq '-')
        {
            break;
        }

        $matched = $true
    } while($false)
    
    if(-not $matched)
    {
        throw "The cloud service name $ServiceName is invalid, It must meet the following criteria: `n 1. It must contain fewer than 63 characters;`n 2. It can contain only letters, numbers, and hyphens, and the first and last character must be a letter or number."
    }
}

function ValidateServiceNamingPattern
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count      
    )

    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z][0-9a-zA-Z-]{0,61}%[0-9]{1,62}%$"
        if($ServiceNamePattern -cnotmatch $regex)
        {
            break
        }
        if($ServiceNamePattern.Length -gt 65)
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The cloud service naming pattern $ServiceNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 4 and 65 characters including the percent signs;`n 2. The root_name can contain only letters, numbers, and hyphens, and it must not start with a hyphen; `n 3. The start_number can contain only numbers; `nExample: AzureService%100%"
    }

    $result = ParseNamePattern -NamePattern $ServiceNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $Count)
    {
        throw "Not enough name space in service naming pattern $ServiceNamePattern, $Count names required, only $availNameCount available"
    }
}

function ValidateStorageAccountNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName
    )
    
    $regex = "^[0-9a-z]{3,24}$"
    if($StorageAccountName -cnotmatch $regex)
    {
        throw "The storage account name $StorageAccountName is invalid. It must contain between 3 and 24 characters, and can contain only lowercase letters and numbers."
    }
}

function ValidateStorageAccountNamingPattern
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count      
    )

    $matched = $false
    do 
    {
        $regex = "^[0-9a-z]{1,23}%[0-9]{1,23}%$"
        if($StorageAccountNamePattern -cnotmatch $regex)
        {
            break
        }
        if(($StorageAccountNamePattern.Length -gt 26) -or ($StorageAccountNamePattern.Length -lt 5))
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The storage account naming pattern $StorageAccountNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 5 and 26 characters including the percent signs;`n 2. The root_name can contain only lower-case letters and numbers; `n 3. The start_number can contain only numbers; `nExample: azurestorage%100%"
    }

    $result = ParseNamePattern -NamePattern $StorageAccountNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $Count)
    {
        throw "Not enough name space in storage account naming pattern $StorageAccountNamePattern, $Count names required, only $availNameCount available"
    }
}

function ValidateAffinityGroupNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $AffinityGroupName
    )
    
    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z][0-9a-zA-Z-]{0,62}$"
        if($AffinityGroupName -cnotmatch $regex)
        {
            break;
        }
        if($AffinityGroupName[-1] -eq '-')
        {
            break;
        }

        $matched = $true
    } while($false)
    
    if(-not $matched)
    {
        throw "The affinity group name $AffinityGroupName is invalid. It must meet the following criteria: `n 1. It must contain fewer than 63 characters;`n 2. It can contain only letters, numbers, and hyphens, and the first and last character must be a letter or number."
    }
}

function ValidateVMNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VMName
    )

    $regex = "^[a-zA-Z][a-zA-Z0-9-]{1,13}[a-zA-Z0-9]$"
    if($VMName -cnotmatch $regex)
    {
        throw "The VM Name $VMName is invalid. it must meet the following criteria: `n 1. It must contain between 3 and 15 characters; `n 2. It can contain only letters, numbers, and hyphens. `n 3. It must start with a letter and must end with a letter or a number."
    }
}


function ValidateVMPatternNaming
{
    Param
    (
        # The VM Name pattern
        [Parameter(Mandatory=$true)]
        [String] $VMNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $NodeCount      
    )

    $matched = $false
    do 
    {
        $regex = "^[a-zA-Z][a-zA-Z0-9-]{0,13}%[0-9]{1,14}%$"
        if($VMNamePattern -cnotmatch $regex)
        {
            break
        }
        if(($VMNamePattern.Length -gt 17) -or ($VMNamePattern.Length -lt 5))
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The VM naming pattern $VMNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 5 and 17 characters including the percent signs;`n 2. The root_name can contain only letters, numbers, and hyphens, and it must start with a letter `n 3. The start_number can contain only numbers; `nExample: AzureCN%1000%"
    }

    $result = ParseNamePattern -NamePattern $VMNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $NodeCount)
    {
        throw "Not enough name space in VM naming pattern $VMNamePattern, $NodeCount names required, only $availNameCount available"
    }
}

function ValidateVMSize
{
    Param
    (
        # The VM Name
        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$true)]
        [String] $VMSize,
        
        [Parameter(Mandatory=$true)]
        [String[]] $AvailableSizes              
    )

    if($AvailableSizes -cnotcontains $VMSize)
    {
        throw "The VM size '$VMSize' of VM $VMName is invalid. Valid sizes are: $([String]::Join(',', $AvailableSizes))"
    }
}


function ValidateVMExtensionVersion
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Version
    )

    if(($ext.Version -notmatch "^[0-9]{1,5}\.\*$") -and ($ext.Version -notmatch "^[0-9]{1,5}\.[0-9]{1,5}$"))
    {
        throw "The Version $Version is invalid. It must be <major>.<minor>, and the minor can be '*'"
    }
}

function GetRegionalVNetLocation
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName
    )

    $uniqueID = [Guid]::NewGuid().ToString()
    $vnetCfgFile = "$env:temp\VNetCfg_$uniqueID.xml"

    InvokeAzureCmdWithRetry -Command "Get-AzureVNetConfig -ExportToFile `"$vnetCfgFile`"" | Out-Null
    if(Test-Path -Path $vnetCfgFile)
    {
        $xmlDoc = [xml](Get-Content $vnetCfgFile)
        Remove-Item $vnetCfgFile -Force

        $vnetSite = $xmlDoc.GetElementsByTagName("VirtualNetworkSite") | Where-Object {$_.name -eq $VNetName}
        if($null -ne $vnetSite)
        {
            if($vnetSite.HasAttribute("Location"))
            {
                return $vnetSite.Location
            }
            else
            {
                throw "Virtual network $VNetName is not a regional virtual network." 
            }
        }
    }

    throw "Virtual network $VNetName doesn't exist"
}

function CheckService
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String[]] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceAffinityGroupTable,
        
        # If not specified, we don't check the deployment label
        [Parameter(Mandatory=$false)]
        [String] $DeploymentLabel=""
    )

    $svcStateTable = @{}
    $errmsgs = @()
    $existingServices = @(InvokeAzureCmdWithRetry -Command "Get-AzureService")
    foreach($svcName in $ServiceName)
    {
        try
        {
            TraceInfo "Checking cloud service $svcName ..."
            $service = $existingServices | ?{$_.ServiceName -eq $svcName}
            if($null -ne $service)
            {
                $agName = $ServiceAffinityGroupTable[$svcName]
                if([String]::IsNullOrEmpty($agName))
                {
                    if($null -ne $service.AffinityGroup)
                    {
                        throw "The cloud service $svcName already exists but in an affinity group $($service.AffinityGroup), not directly in the region $Location."
                    }
                    elseif($service.Location -ne $Location)
                    {
                        throw "The cloud service $svcName already exists but in the region $($service.Location), not in the region $Location."
                    }
                }
                else
                {
                    if($null -eq $service.AffinityGroup)
                    {
                        throw "The cloud service $svcName already exists but not in the affinity group $agName."
                    }
                    if($service.AffinityGroup -ne $agName)
                    {
                        throw "The cloud service $svcName already exists but in affinity group $($service.AffinityGroup), not in affinity group $agName."
                    }
                }

                $deployment = InvokeAzureCmdWithRetry -Command "Get-AzureDeployment -ServiceName $svcName -WarningAction SilentlyContinue"
                if($null -ne $deployment)
                {
                    # Check VNet
                    if($deployment.VNetName -ne $VNetName)
                    {
                        throw "The production deployment of cloud service $ServiceName is not in virtual network $VNetName, but in virtual network $($deployment.VNetName)."
                    }
            
                    if((-not [String]::IsNullOrEmpty($DeploymentLabel)) -and ((DecodeLabelIfBase64Encoded $deployment.Label) -ne $DeploymentLabel))
                    {
                         throw "The cloud service $svcName is already in use by other production deployment."
                    }

                    TraceInfo "The cloud service $svcName already exists."
                    $svcStateTable[$svcName] = "Deployed"
                }
                else
                {
                    TraceInfo "The cloud service $svcName already exists."
                    $svcStateTable[$svcName] =  "Created"
                }
            }
            else
            {
                if((InvokeAzureCmdWithRetry -Command "Test-AzureName -Service -Name $svcName") -eq $true)
                {
                    throw "The cloud service name $svcName is not available, it is already in use."
                }

                TraceInfo "Cloud service $svcName doesn't exist"
                $svcStateTable[$svcName] = "NotCreated"
            }
        }
        catch
        {
            $errmsgs += $_.Exception.Message
        }
    }

    if($errmsgs.Count -ne 0)
    {
        throw ($errmsgs -join "`r`n")
    }

    return $svcStateTable
}

function ValidateVMConfig
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName
    )

    $vmexists = $false
    $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -Name $VMName -ServiceName $ServiceName"
    if($null -ne $vm)
    {
        $vmexists = $true
        $vmInSubnet = Get-AzureSubnet -VM $vm -ErrorAction SilentlyContinue
        if($vmInSubnet -ne $SubnetName)
        {
            throw "The virtual machine with same name $VMName already exists in cloud service $ServiceName, but not in Subnet $SubnetName"
        }
        TraceInfo "The virtual machine $VMName already exists"
    }
    else
    {
        TraceInfo "The virtual machine $VMName doesn't exist"
    }

    return $vmexists
}

<#
.SYNOPSIS
  Vadlidate input Cluster Config file
.DESCRIPTION
  Validate whether the input cluster config file is valid, throw exception if not valid
.EXAMPLE
  ConfigFileSemanticsCheck .\HPCIaaSClusterConfig.xml
#>
function ConfigFileSemanticsCheck
{
    Param
    (
        # Prefix name to name resources
        [Parameter(Mandatory=$true)]
        [String] $ConfigFile
    )
	
    if (-not (Test-Path -Path $ConfigFile))
    {
        throw "Config File $ConfigFile does not exist"
    }

    if(-not (Test-Path -Path "$PSScriptRoot\HPCIaaSClusterConfig.xsd"))
    {
        throw "XML Schema File HPCIaaSClusterConfig.xsd does not exist"
    }
    
    $xmlFullPath = (Resolve-Path -Path $ConfigFile).Path

    ## Validate the config file against schema
    ValidateXmlAgainstXsd -XmlFile $xmlFullPath -SchemaFile "$PSScriptRoot\HPCIaaSClusterConfig.xsd"

    $errormsgs = @()
    $warningmsgs = @()
    $config = ([Xml] (Get-Content $xmlFullPath)).IaaSClusterConfig
    $commonConfig = GetCommonConfig $config
    $domainConfig = GetDomainConfig $config.Domain
    $dbConfig = GetDatabaseConfig $config.Database
    $hnConfig = GetHeadNodeConfig $config.HeadNode
    $cnConfig = $null
    $bnConfig = $null
    $lnConfig = $null
    if($null -ne $config["ComputeNodes"])
    {
        $cnConfig = GetNodeGroupConfig $config.ComputeNodes
    }
    if($null -ne $config["BrokerNodes"])
    {
        $bnConfig = GetNodeGroupConfig $config.BrokerNodes
    }
    if($null -ne $config["LinuxComputeNodes"])
    {
        $lnConfig = GetNodeGroupConfig $config.LinuxComputeNodes
    }

    ## Validate the subscription
    $subNameExists = $false
    if($null -ne $commonConfig.SubscriptionName)
    {
        $subNameExists = $true 
    }

    if($null -eq $commonConfig.SubscriptionId)
    {
        if(-not $subNameExists)
        {
            $errormsgs += "Either 'SubscriptionName' or 'SubscriptionId' must be specified in 'Subscription'"
        }
    }
    else
    {
        if($subNameExists)
        {
            $errormsgs += "You cannot specify both 'SubscriptionName' and 'SubscriptionId', specify either of them in 'Subscription'"
        }
        else
        {
            try
            {
                [System.Guid]::Parse($commonConfig.SubscriptionId) | Out-Null
            }
            catch
            {
                $errormsgs += "The 'SubscriptionId' in 'Subscription' is not a valid Guid: $($commonConfig.SubscriptionId)"
            }
        }
    }

    ## Validate Domain options
    if($domainConfig.DCOption -eq "NewDC")
    {
        if($null -eq $domainConfig.DomainController)
        {
            $errormsgs += "'DomainController' must be specified when 'DCOption' is specified as 'NewDC'"
        }
    }
    # Workaround we allow ExistingDC to give DC VM information
    elseif($domainConfig.DCOption -eq "HeadNodeAsDC")
    {
        if($null -ne $domainConfig.DomainController)
        {
            $errormsgs += "'DomainController' must not be specified when 'DCOption' is specified as 'ExistingDC' or 'HeadNodeAsDC'"
        }
    }

    if($null -ne $domainConfig.DNSForwarder)
    {
        if($domainConfig.DCOption -eq "ExistingDC")
        {
            $errormsgs += "'DNSForwarder' must not be specified when 'DCOption' is specified as 'ExistingDC'"
        }
        else
        {
            try
            {
                 [System.Net.IPAddress]::Parse($domainConfig.DNSForwarder) | Out-Null
            }
            catch
            {
                $errormsgs += "'DNSForwarder' under 'Domain' is not a valid IP address: $($domainConfig.DNSForwarder)"
            }
        }
    }

    if($domainConfig.NoWindowsAutoUpdate -and ($domainConfig.DCOption -eq "ExistingDC"))
    {
        $errormsgs += "'NoWindowsAutoUpdate' must not be specified when 'DCOption' is specified as 'ExistingDC'"
    }
    

    ## Validate Database options

    if ($dbConfig.DBOption -eq "ExistingDB")
    {
        if($domainConfig.DCOption -ne "ExistingDC")
        {
            $errormsgs += "'ExistingDB' can only be specified when you already have an existing domain forest"
        }
        if($null -ne $dbConfig.DBVersion)
        {
            $errormsgs += "'DBVersion' must not be specified when 'DBOption' is specified as 'ExistingDB'"
        }
        if($null -eq $dbConfig.DBConnection)
        {
            $errormsgs += "'DBConnection' must be specified when 'DBOption' is specified as 'ExistingDB'"
        }
        else
        {
            if([String]::IsNullOrEmpty($dbConfig.DBConnection.ServerInstance))
            {
                $errormsgs += "'ServerInstance' must be specified when 'DBOption' is specified as 'ExistingDB'"
            }
        }
    }
    elseif($dbConfig.DBOption -eq "NewRemoteDB")
    {
        if($null -eq $dbConfig.DBServer)
        {
            $errormsgs += "'DBServer' must be specified when 'DBOption' is specified as 'NewRemoteDB'"
        }
        else
        {
            if(($dbConfig.DBServer.DataDiskSizeInGB -lt 100) -or ($dbConfig.DBServer.DataDiskSizeInGB -gt 1000))
            {
                $errormsgs += "The allowed value range of 'DataDiskSizeInGB' is 100 - 1000."
            }
        }

        if($null -eq $dbConfig.DBVersion)
        {
            $errormsgs += "'DBVersion' must be specified when 'DBOption' is specified as 'NewRemoteDB'"
        }
    }
    elseif($dbConfig.DBOption -eq "LocalDB")
    {
        if($null -ne $dbConfig.DBVersion)
        {
            $errormsgs += "'DBVersion' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
        if($null -ne $dbConfig.DBServer)
        {
            $errormsgs += "'DBServer' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
        if($null -ne $dbConfig.DBConnection)
        {
            $errormsgs += "'DBConnection' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
    }

    ## Validate head node options
    if($null -ne $hnConfig.ImageFamily)
    {
        if($hnConfig.ImageFamily -notmatch "HPC Pack")
        {
            $errormsgs += "The 'ImageFamily' in 'HeadNode' is not a valid HPC image family: $($hnConfig.ImageFamily)"
        }
    }

    ## Validate the VM Size Combination in Affinity Group
    $AGVMSizeMap = @{}
    $configDefaultAGName = ""
    if($null -ne $config["AffinityGroup"])
    {
        $configDefaultAGName = $config.AffinityGroup.Trim()
    }

    $vmConfigList = @($domainConfig.DomainController, $dbConfig.DBServer, $hnConfig, $cnConfig, $bnConfig, $lnConfig)
    foreach($vmConfig in $vmConfigList)
    {
        if($null -eq $vmConfig)
        {
            continue
        }

        $vmAGName = $vmConfig.AffinityGroup
        if($null -eq $vmConfig.AffinityGroup)
        {
            $vmAGName = $configDefaultAGName
        }

        $AGVMSizeMap[$vmAGName] += @($vmConfig.VMSize)
        $AGVMSizeMap[$vmAGName] = @($AGVMSizeMap[$vmAGName] | select -Unique)
    }
    
    $agErr = @()
    $RDMASizes = @(GetVMSizesWithRDMA)
    foreach($agname in $AGVMSizeMap.Keys)
    {
        if($agname -eq "")
        {
            continue
        }

        # All in RDMASizes or none in RDMASizes
        if(@($AGVMSizeMap[$agname] | % { $_ -in $RDMASizes } | select -Unique).Count -ne 1)
        {
            $agErr += ("$agname : {0}`n" -f ($AGVMSizeMap[$agname] -join " "))
        }
    }

    if($agErr.Count -gt 0)
    {
        $errormsgs += "VM size combination error in the following affinity group(s), A8/A9 sizes cannot co-exist with other sizes: `n$agErr"
    }

    ## Validate the naming of the config file
    $agCount = 0
    foreach($agname in $AGVMSizeMap.Keys)
    {
        if($agname -eq "")
        {
            continue
        }

        $agCount++
        try
        {
            ValidateAffinityGroupNaming -AffinityGroupName $agname
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($agCount -gt 0)
    {
        $warningmsgs += "Affinity group is no longer recommended as a result of Azure network architectural improvement. You can remove 'AffinityGroup' from the configuration file."
    }

    try
    {
        ValidateStorageAccountNaming -StorageAccountName $config.Subscription.StorageAccount.Trim()
    }
    catch
    {
        $errormsgs += $_.Exception.Message
    }

    if($null -ne $config["VNet"])
    {
        $configVNetName = $config.VNet.VNetName.Trim()
        $configSubnetName = $config.VNet.SubnetName.Trim()
        try
        {
            ValidateVNetNaming -VNetName $configVNetName -SubnetName $configSubnetName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($null -ne $domainConfig.DomainController)
    {
        try
        {
            ValidateVMNaming -VMName $domainConfig.DomainController.VMName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $domainConfig.DomainController.ServiceName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($null -ne $dbConfig.DBServer)
    {
        try
        {
            ValidateVMNaming -VMName $dbConfig.DBServer.VMName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $dbConfig.DBServer.ServiceName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($null -ne $hnConfig)
    {
        try
        {
            ValidateVMNaming -VMName $hnConfig.VMName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $hnConfig.ServiceName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        if($null -ne $hnConfig.PreConfigScript)
        {
            $configHNPreConfigScript = $hnConfig.PreConfigScript
            if (-not (Test-Path -Path $configHNPreConfigScript -PathType Leaf))
            {
                $errormsgs += "The 'PreConfigScript' file does not exist: $configHNPreConfigScript"
            }
            elseif(-not $configHNPreConfigScript.ToLower().EndsWith(".ps1"))
            {
                $errormsgs += "The 'PreConfigScript' file is not a PowerShell script file: $configHNPreConfigScript"
            }
        }

        if($null -ne $hnConfig.PostConfigScript)
        {
            $configHNPostConfigScript = $hnConfig.PostConfigScript
            if (-not (Test-Path -Path $configHNPostConfigScript -PathType Leaf))
            {
                $errormsgs += "The 'PostConfigScript' file does not exist: $configHNPostConfigScript"
            }
            elseif(-not $configHNPostConfigScript.ToLower().EndsWith(".ps1"))
            {
                $errormsgs += "The 'PostConfigScript' file is not a PowerShell script file: $configHNPostConfigScript"
            }
        }
    }

    foreach($nodeGroupConfig in @($cnConfig, $bnConfig, $lnConfig))
    {
        if($null -eq $nodeGroupConfig)
        {
            continue
        }

        try
        {
            ValidateVMPatternNaming -VMNamePattern $nodeGroupConfig.VMNamePattern -NodeCount $nodeGroupConfig.NodeCount
        }
        catch
        {
            $errormsgs += $_.Exception.Message            
        }

        if($null -ne $nodeGroupConfig.ServiceName)
        {
            try
            {
                ValidateServiceNaming -ServiceName $nodeGroupConfig.ServiceName
            }
            catch
            {
                $errormsgs += $_.Exception.Message            
            }

            if($nodeGroupConfig.NodeCount -gt 50)
            {
                $errormsgs += "The 'NodeCount' of $($nodeGroupConfig.NodeGroupName) has exceeded the maximum value 50: $($nodeGroupConfig.NodeCount)"
            }
        }
        else
        {
            # Currently, 'ServiceNamePattern' not supported for 'BrokerNodes'
            if($nodeGroupConfig.NodeGroupName -eq 'BrokerNodes')
            {
                $errormsgs += "'ServiceNamePattern' is not supported in $($nodeGroupConfig.NodeGroupName), please specify 'ServiceName' instead."
            }
            else
            {
                $svcNamePattern = $nodeGroupConfig.ServiceNamePattern
                $maxNodeCountPerSvc = $nodeGroupConfig.MaxNodeCountPerService
                $storageNamePattern = $nodeGroupConfig.StorageAccountNamePattern
                if(($maxNodeCountPerSvc -eq 0) -or ($maxNodeCountPerSvc -gt 50))
                {
                    $errormsgs += "The 'MaxNodeCountPerService' of $($nodeGroupConfig.NodeGroupName) must be between 1 and 50"
                }
                else
                {
                    $nodeSetCount = [Math]::Ceiling($nodeGroupConfig.NodeCount/$maxNodeCountPerSvc)
                    try
                    {
                        ValidateServiceNamingPattern -ServiceNamePattern $svcNamePattern -Count $nodeSetCount
                    }
                    catch
                    {
                        $errormsgs += $_.Exception.Message            
                    }
                    try
                    {
                        ValidateStorageAccountNamingPattern -StorageAccountNamePattern $storageNamePattern -Count $nodeSetCount
                    }
                    catch
                    {
                        $errormsgs += $_.Exception.Message            
                    }
                }
            }
        }

        if($null -ne $nodeGroupConfig.VMExtensions)
        {
            foreach($ext in $nodeGroupConfig.VMExtensions)
            {
                try
                {
                    ValidateVMExtensionVersion -Version $ext.Version
                }
                catch
                {
                    $errormsgs += ("Error in VMExtension $($ext.ExtensionName): " + $_.Exception.Message)
                }
            }
        }
    }

    if($null -ne $lnConfig)
    {
        if($null -ne $lnConfig.OSVersion)
        {
            $errormsgs +=  "'OSVersion' not supported for LinuxNodes."
        }
        if($null -eq $lnConfig.SSHKeyPairForRoot)
        {
            $warningmsgs += "'SSHKeyPairForRoot' is not specified in 'LinuxComputeNodes', the script will generate an SSH key pair for test purpose. It is recommended to specify 'SSHKeyPairForRoot' for production deployments."
        }
    }

    ## Validate the Service/VM Name conflict
    $services = @{}
    $vmNames = @()
    # Collect the service/VM name info of DC, DB, HN
    foreach($vmConfig in @($domainConfig.DomainController, $dbConfig.DBServer, $hnConfig))
    {
        if($null -ne $vmConfig)
        {
            $vmAGName = $vmConfig.AffinityGroup
            if($null -eq $vmAGName)
            {
                $vmAGName = $configDefaultAGName
            }

            $services[$vmConfig.ServiceName] += @($vmAGName)
            $services[$vmConfig.ServiceName] = @($services[$vmConfig.ServiceName] | select -Unique)
            if($vmNames -contains $vmConfig.VMName)
            {
                $errormsgs += "There is more than one VM with same name: $($vmConfig.VMName)."
            }
            else
            {
                $vmNames += $vmConfig.VMName
            }
        }
    }

    # Validate the services /VM name for Node groups
    $svcNamePatterns = @()
    $vmNamePatterns = @()
    $nodeGroupSingleSvcs = @()
    foreach($nodeGroupConfig in @($cnConfig, $bnConfig, $lnConfig))
    {
        if($null -ne $nodeGroupConfig)
        {
            # Check VM Name confliction
            foreach($vmname in $vmNames)
            {
                if(NameMatchesPattern -Name $vmname -NamePattern $nodeGroupConfig.VMNamePattern)
                {
                    $errormsgs += "The VM name $vmname conflicts with 'VMNamePattern' of $($nodeGroupConfig.NodeGroupName): $($nodeGroupConfig.VMNamePattern)."
                }                
            }
            
            foreach($pattern in $vmNamePatterns)
            {
                if(NamePatternOverlapped -NamePattern1 $pattern -NamePattern2 $nodeGroupConfig.VMNamePattern)
                {
                    $errormsgs += "The 'VMNamePattern' of $($nodeGroupConfig.NodeGroupName) may overlap with other VMNamePattern: $($nodeGroupConfig.VMNamePattern) vs $pattern."
                }
            }

            $vmNamePatterns += $nodeGroupConfig.VMNamePattern

            # Check service confliction
            $vmAGName = $nodeGroupConfig.AffinityGroup
            if($null -eq $vmAGName)
            {
                $vmAGName = $configDefaultAGName
            }

            if($null -ne $nodeGroupConfig.ServiceName)
            {
                if($nodeGroupSingleSvcs -contains $nodeGroupConfig.ServiceName)
                {
                    $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other. The 'ServiceName' $($nodeGroupConfig.ServiceName) for $($nodeGroupConfig.NodeGroupName) is already used by another node group."
                }
                else
                {
                    if($services.Keys -contains $nodeGroupConfig.ServiceName)
                    {
                        $warningmsgs += "The 'ServiceName' $($nodeGroupConfig.ServiceName) for $($nodeGroupConfig.NodeGroupName) is already used by other node(s). It is recommended to specifiy a dedicated cloud service for $($nodeGroupConfig.NodeGroupName)"
                    }

                    $nodeGroupSingleSvcs += $nodeGroupConfig.ServiceName
                    $services[$nodeGroupConfig.ServiceName] += @($vmAGName)
                    $services[$nodeGroupConfig.ServiceName] = @($services[$nodeGroupConfig.ServiceName] | select -Unique)
                }

                foreach($pattern in $svcNamePatterns)
                {
                    if(NameMatchesPattern -Name $nodeGroupConfig.ServiceName -NamePattern $pattern)
                    {
                        $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other. The 'ServiceName' for $($nodeGroupConfig.NodeGroupName) conflicts with the 'ServiceNamePattern' for another node group: $($nodeGroupConfig.ServiceName) vs $pattern."
                    }
                }
            }
            else
            {
                foreach($svc in $services.Keys)
                {
                    if(NameMatchesPattern -Name $svc -NamePattern $nodeGroupConfig.ServiceNamePattern)
                    {
                        if($nodeGroupSingleSvcs -contains $svc)
                        {
                            $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other. The 'ServiceNamePattern' for $($nodeGroupConfig.NodeGroupName) conflicts with the 'ServiceName' for another node group: $($nodeGroupConfig.ServiceNamePattern) vs $svc."
                        }
                        else
                        {
                            $warningmsgs += "The 'ServiceNamePattern' for $($nodeGroupConfig.NodeGroupName) conflicts with the 'ServiceName' for another node group: $($nodeGroupConfig.ServiceNamePattern) vs $svc. It is recommended to specify dedicated cloud services for $($nodeGroupConfig.NodeGroupName)."
                        }
                    }
                }

                foreach($pattern in $svcNamePatterns)
                {
                    if(NamePatternOverlapped -NamePattern1 $pattern -NamePattern2 $nodeGroupConfig.ServiceNamePattern)
                    {
                        $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other, the 'ServiceNamePattern' for $($nodeGroupConfig.NodeGroupName) may overlap with that for another node group: $($nodeGroupConfig.ServiceNamePattern) vs $pattern."
                    }
                }

                $svcNamePatterns += $nodeGroupConfig.ServiceNamePattern
            }
        }
    }

    foreach($svc in $services.Keys)
    {
        if($services[$svc].Count -gt 1)
        {
            $errormsgs += ("The cloud service $svc cannot be specified in multiple affnity groups: " + $services[$svc])
        }
    }

    ### Check certificates
    $certList = @(GetCertificatesConfig -IaaSClusterConfig $config)
    $certIds = @($certList | %{$_.Id})
    $pfxFiles = @($certList | %{$_.PfxFile.ToLower()})
    foreach($group in ($certIds | group))
    {
        if($group.Count -gt 1)
        {
            $errormsgs += "[Certificate] There are two or more certificates with the same Id: " + $group.Name
        }
    }

    foreach($group in ($pfxFiles | group))
    {
        if($group.Count -gt 1)
        {
            $errormsgs += "[Certificate] The following certificate is specified multiple times: " + $group.Name
        }
    }

    if($certList.Count -gt 1)
    {
        if($null -ne $config["Certificates"] -and ($null -ne $config["AzureBurst"]) -and ($null -ne $config.AzureBurst["Certificate"]))
        {
            $warningmsgs += "The 'Certificate' elements are specified in both 'Certificates' and 'AzureBurst'. You can move all certificates from 'AzureBurst' to 'Certificates'"
        }
    }

    ### Check AzureBurst
    if($null -ne $config["AzureBurst"])
    {
        $errormsgs += CheckAzureBurstConfig -AzureBustConfig $config["AzureBurst"] -Certificates $certList
    }

    ### Check AutoGrowShink
    if($null -ne $config["AutoGrowShrink"])
    {
        $autoGrownShrinkConfig =  GetAutoGrowShrinkConfig $config["AutoGrowShrink"]
        if($certIds -notcontains $autoGrownShrinkConfig.CertificateId)
        {
            $errormsgs += "[AutoGrowShink] The 'CertificateId' $($autoGrownShrinkConfig.CertificateId) cannot be found."
        }
        if($autoGrownShrinkConfig.GrowInterval -ne 0 -and $autoGrownShrinkConfig.GrowInterval -lt 1)
        {
            $errormsgs += "[AutoGrowShink] The 'GrowInterval' should be no less than 1 minutes."
        }
        if($autoGrownShrinkConfig.ShrinkInterval -ne 0 -and $autoGrownShrinkConfig.ShrinkInterval -lt 2)
        {
            $errormsgs += "[AutoGrowShink] The 'ShrinkInterval' should be no less than 2 minutes."
        }
    }

    ### Conclusion
    if($errormsgs.Count -gt 0)
    {
        $msg = ""
        $i = 1
        while($i -le $errormsgs.Count)
        {
            $msg += ("[Error $i] {0}`r`n" -f $errormsgs[$i-1])
            $i++
        }

        throw $msg
    }

    foreach($msg in $warningmsgs)
    {
        Write-Warning $msg
    }
}

function CheckStorageAccount
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $StorageAccountName,

        [Parameter(Mandatory=$true)]
        [String] $Location
    )

    $accountStatus = @{}
    $errMsg = @()
    $existingStorageAccounts = @(InvokeAzureCmdWithRetry -Command "Get-AzureStorageAccount")
    foreach($stg in $StorageAccountName)
    {
        TraceInfo "Checking storage account $stg"
        $storageAccount = $existingStorageAccounts | ?{$_.StorageAccountName -eq $stg}
        if($null -ne $storageAccount)
        {
            if($storageAccount.GeoPrimaryLocation -ne $Location)
            {
                $errMsg += "The storage account $stg already exists, but in region $($storageAccount.GeoPrimaryLocation), not in region $Location."
                continue
            }

            if($null -eq $storageAccount.AffinityGroup)
            {
                TraceInfo "The storage account $stg already exists in region $Location."
            }
            else
            {
                TraceInfo "The storage account $stg already exists in Affinity Group $($storageAccount.AffinityGroup)."
            }
        }
        else
        {
            if((InvokeAzureCmdWithRetry -Command "Test-AzureName -Storage -Name $stg") -eq $true)
            {
                $errMsg += "The storage account name $stg is not available. It is already in use."
                continue
            }
        }

        $accountStatus[$stg] = ($null -ne $storageAccount)
    }

    if($errMsg.Count -gt 0)
    {
        throw ($errMsg -join "`r`n")
    }

    return $accountStatus
}

function CheckLocationAndAffinityGroup
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [String[]] $AGNames
    )

    $locations = InvokeAzureCmdWithRetry -Command "Get-AzureLocation" | ForEach-Object {$_.Name}
    if($locations -notcontains $Location)
    {
        throw "Invalid Location: $Location."
    }
    else
    {
        TraceInfo "$Location is a valid Location."
    }

    $agStates = @{}
    if(($null -ne $AGNames) -and ($AGNames.Count -gt 0))
    {
        $errmsgs = @()
        foreach($agname in $AGNames)
        {
            $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $agname"
            if($null -ne $ag)
            {
                if($ag.Location -ne $Location)
                {
                    $errmsgs += "The affinity group $agname already exists in region $($ag.Location), but not in region $Location."
                }
                else
                {
                    TraceInfo "The affinity group $agname already exists."
                    $agStates[$agname] = $true
                }
            }
            else
            {
                TraceInfo "The affinity group $agname doesn't exist yet."
                $agStates[$agname] = $false
            }
        }
        if($errmsgs.Count -gt 0)
        {
            throw ($errmsgs -join "`r`n")
        }
    }

    return $agStates
}

function CheckVNet
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName,

        [Parameter(Mandatory=$true)]
        [String] $Location
    )

    $vnetExists = $false
    $vnet = Get-AzureVNetSiteWithRetry -VNetName $VNetName -ErrorAction SilentlyContinue
    if ($null -ne $vnet)
    {
        $vnetExists = $true
        # The VNet already exists, check whether the settings are consistent
        # Decide whether it is a Regional VNet by Property AffinityGroup, for Regional VNet, AffinityGroup is $null
        if($null -eq $vnet.AffinityGroup)
        {
            # Check the location information for Regional VNet
            # Currently, we cannot get the Location by Get-AzureVNetSite command for Regional VNet
            $vnetLocation = GetRegionalVNetLocation -VNetName $VNetName
            if($vnetLocation -ne $Location)
            {
                throw "The virtual network $VNetName already exists but in region $vnetLocation, not in $Location."
            }

            TraceInfo "The virtual network $VNetName already exists in region $Location."                    
        }
        else
        {
            $vnetAG = $vnet.AffinityGroup
            throw "The virtual network $VNetName already exists but is associated with an affinity group $vnetAG."
        }

        ### Validate Subnet
        $subnet = $vnet.Subnets | Where-Object Name -eq $SubnetName
        if($null -eq $subnet)
        {
            throw "The virtual network $VNetName already exists but no subnet $SubnetName configured in it."
        }
        else
        {
            TraceInfo "The subnet $SubnetName already configured in $VNetName."
        }
    }
    else
    {
        TraceInfo "The virtual network $VNetName doesn't exist."
    }

    return $vnetExists
}

function CheckAzureVMImage
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$true)]
        [ValidateSet("Windows","Linux")]
        [String] $OS,

        [Parameter(Mandatory=$false)]
        [String] $Location = "",

        # The image list to be searched, if not specified, use Get-AzureVMImage to get the image.
        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @()
    )

    if(($null -eq $FromImageList) -or ($FromImageList.Count -eq 0))
    {
        $image = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $ImageName"
    }
    else
    {
        $image = $FromImageList | ?{$_.ImageName -eq $ImageName}
    }

    if($null -eq $image)
    {
        throw "The image $ImageName does not exist"
    }
    
    if($image.OS -ne $OS)
    {
        throw "The image $ImageName is not a $OS image, but a $($image.OS) image."
    }
    
    if(-not [String]::IsNullOrEmpty($Location))
    {
        $imgLocations = @()        
        if($null -ne $image.Location)
        {
            $imgLocations += $image.Location.Split(';')
        }
        elseif($null -ne $image.AffinityGroup)
        {
            $agName = $image.AffinityGroup
            $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $agName"
            if($null -ne $ag)
            {
                $imgLocations += $ag.Location
            }
        }

        if($imgLocations -notcontains $Location)
        {
            throw "The image $ImageName found, but not in region $Location"
        }
    }
}

function CheckNodesInService
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String[]] $TargetNodeNames,
        
        [Parameter(Mandatory=$true)]
        [ValidateSet("ComputeNode","BrokerNode", "LinuxNode")]
        [String] $NodeType,

        [Parameter(Mandatory=$false)]
        [String] $ImageName = "",
                
        [Parameter(Mandatory=$false)]
        [Switch] $ImageHPCInstalled
    )

    TraceInfo "Checking virtual machines in cloud service $ServiceName"
    $VMsInService = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName")
    # Check whether there is any VM other than the target VMs in the deployment
    $otherVMNames = @($VMsInService | ?{$TargetNodeNames -notcontains $_.Name} | %{$_.Name})
    if($otherVMNames.Count -gt 0)
    {
        TraceWarning "There are some other virtual machines in the cloud service $ServiceName : $otherVMNames"
    }

    $existingNodes = @($VMsInService | ?{$TargetNodeNames -contains $_.Name})
    # Check the source image of the existing VMs
    $imgNotMatch = @()
    $srcImgNames = @()
    foreach($vm in $existingNodes)
    {
        # There are two types of VM images: VMImage and OSImage
        # If the VM was created from VMImage, the SourceImageName is VMImageContext.OSDiskConfiguration.Name (typically as <ImageName>-os-yyyy-MM-dd)
        # If the VM was created from OSImage, the SourceImageName is OSImageContext.ImageName
        $srcImgName = $vm.VM.OSVirtualHardDisk.SourceImageName
        $srcImgNames += $srcImgName

        if($NodeType -eq "LinuxNode")
        {
            # For Linux Node, we just check whether VM was created from the given image or its replicas.
            if(-not [String]::IsNullOrEmpty($ImageName))
            {
                # if the given image is HPC installed, the source image name shall be <ImageName>[_<StorageAccountName>] or <ImageName>-os-yyyy-MM-dd
                if(($srcImgName -notmatch "^$ImageName(_[0-9a-z]{3,24})?$") -and ($srcImgName -notmatch "^$ImageName-os-"))
                {
                    $imgNotMatch += $vm.Name
                }                
            }
        }
        else
        {
            if(-not [String]::IsNullOrEmpty($ImageName) -and $ImageHPCInstalled.IsPresent)
            {
                # if the given image is HPC installed, the source image name shall be <ImageName>[_<StorageAccountName>] or <ImageName>-os-yyyy-MM-dd
                if(($srcImgName -notmatch "^$ImageName(_[0-9a-z]{3,24})?$") -and ($srcImgName -notmatch "^$ImageName-os-"))
                {
                    $imgNotMatch += $vm.Name
                }
            }
            else
            {
                # if the given image is not HPC installed, means we will create the base HPC node image and create nodes with this base image.
                # The source image name shall be HPC-Pack-<NodeType>_<DateTime>[_<StorageAccountName>] or HPC-Pack-<NodeType>_<DateTime>-os-yyyy-MM-dd
                if(-not (IsBaseHPCNodeImage -VMSrcImageName $srcImgName -NodeType $NodeType))
                {
                    $imgNotMatch += $vm.Name
                }
            }
        }
    }
    if($imgNotMatch.Count -gt 0)
    {
        throw ("The following virtual machines already exist in cloud service $ServiceName but not created by this script: " + ($imgNotMatch -join ","))
    }

    $readyVMNames = @()
    if($existingNodes.Count -gt 0)
    {
        $srcImgNames = @($srcImgNames | select -Unique | Sort-Object -Descending)
        $readyVMNames = @($existingNodes | %{$_.Name})
    }

    return @{"ReadyVMNames"=$readyVMNames;"VMImages"=$srcImgNames}
}

function CheckNodeGroup
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $NamePattern, 

        [Parameter(Mandatory=$true)]
        [int] $NodeCount, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize,

        [Parameter(Mandatory=$true)]
        [String[]] $AvailableVMSizes,
        
        [Parameter(Mandatory=$true)]
        [Object[]] $AvailableVMImages,
                
        [Parameter(Mandatory=$true, ParameterSetName="SingleService")]
        [String] $ServiceName, 
        
        [Parameter(Mandatory=$true, ParameterSetName="ServicePattern")]
        [String] $ServiceNamePattern, 

        [Parameter(Mandatory=$true, ParameterSetName="ServicePattern")]
        [String] $StorageAccountNamePattern,

        [Parameter(Mandatory=$true, ParameterSetName="ServicePattern")]
        [int] $MaxNodeCountPerService,

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceStateTable, 
        
        [Parameter(Mandatory=$true)]
        [ValidateSet("ComputeNode","BrokerNode", "LinuxNode")]
        [String] $NodeType,
                
        [Parameter(Mandatory=$false)]
        [String] $ImageName = "",

        [Parameter(Mandatory=$true)]
        [String] $Location,
                
        [Parameter(Mandatory=$false)]
        [Switch] $ImageHPCInstalled
    )

    try
    {
        ValidateVMSize -VMName $NamePattern -VMSize $VMSize -AvailableSizes $AvailableVMSizes
        if(-not [String]::IsNullOrEmpty($ImageName))
        {
            if($NodeType -eq "LinuxNode")
            {
                CheckAzureVMImage -ImageName $ImageName -OS Linux -Location $Location -FromImageList $AvailableVMImages
            }
            else
            {
                CheckAzureVMImage -ImageName $ImageName -OS Windows -Location $Location -FromImageList $AvailableVMImages
            }            
        }

        $srcImageNames = @()
        $serviceVMTable = @{}
        if($PsCmdlet.ParameterSetName -eq "ServicePattern")
        {
            $nodeNameSets = @(GetNameSetsByPattern -NamePattern $NamePattern -Count $NodeCount -MaxCountPerSet $MaxNodeCountPerService)
            $nodeSetCount = $nodeNameSets.Count
            $serviceNames = @(GetNamesByPattern -NamePattern $ServiceNamePattern -Count $nodeSetCount)
            $storageNames = @(GetNamesByPattern -NamePattern $StorageAccountNamePattern -Count $nodeSetCount)
        }
        else
        {
            $nodeNameSets = @(, @(GetNamesByPattern -NamePattern $NamePattern -Count $NodeCount))
            $nodeSetCount = 1
            $serviceNames = @($ServiceName)
        }

        $totalReadyVMCount = 0
        for($i=0;$i -lt $nodeSetCount; $i++)
        {
            $serviceName = $serviceNames[$i]
            $targetNodeNames = $nodeNameSets[$i]
            if($ServiceStateTable[$serviceName] -eq "Deployed")
            {
                $result = CheckNodesInService -ServiceName $serviceName -TargetNodeNames $targetNodeNames -NodeType $NodeType `
                                -ImageName $ImageName -ImageHPCInstalled:$ImageHPCInstalled
                $nodeNamesToCreate = @($targetNodeNames | ?{$result["ReadyVMNames"] -notcontains $_})
                $srcImageNames += $result["VMImages"]
                if($nodeNamesToCreate.Count -gt 0)
                {
                    $serviceVMTable[$serviceName] = $nodeNamesToCreate
                }

                $totalReadyVMCount += $result["ReadyVMNames"].Count
            }
            else
            {
                $serviceVMTable[$serviceName] = $targetNodeNames
            }
        }

        $hpcNodeImageName = ""
        if($ImageHPCInstalled -or ($NodeType -eq "LinuxNode"))
        {
            $hpcNodeImageName = $ImageName
        }
        elseif($srcImageNames.Count -gt 0)
        {
            $imageList = @(Get-BaseHPCNodeImage -NodeType $NodeType -Location $Location)
            $srcImages = @()
            if($imageList.Count -gt 0)
            {
                $srcImageNames = $srcImageNames | select -Unique
                foreach($srcImgName in $srcImageNames)
                {
                    $img = GetVMImageByVMSourceImageName -SourceImageName $srcImgName -FromImageList $imageList
                    if(($null -ne $img) -and (IsBaseHPCNodeImage -ImageName $img.ImageName -NodeType $NodeType))
                    {
                        $srcImages += $img
                    }
                }
            }

            if($srcImages.Count -gt 0)
            {
                # The format of the hpc node base image name is HPC-Pack-<NodeType>_<DateTime>[_<StorageAccountName>]
                # Sort the images in descending order by name(<DateTime>[_<StorageAccountName>] part), so that the latest image can be selected.
                $srcImages = $srcImages | Sort-Object -Property ImageName -Descending
                # The first image is probably with tailing _<StorageAccountName>, try to find an image with same <DateTime> but without _<StorageAccountName>.
                $firstImgStg = GetStorageAccountFromMediaLink -MediaLink (GetVMImageMediaLink $srcImages[0])
                $hpcNodeImageName = $srcImages[0].ImageName
                if($hpcNodeImageName -match "_$firstImgStg$")
                {
                    $firstImgNameRoot = $hpcNodeImageName.Substring(0, $hpcNodeImageName.Length - $firstImgStg.Length - 1)
                    $imgWithRootName = $srcImages | ?{$_.ImageName -eq $firstImgNameRoot}
                    if($null -ne $imgWithRootName)
                    {
                        $hpcNodeImageName = $imgWithRootName.ImageName
                    }                
                }
            }
        }

        TraceInfo "$totalReadyVMCount node(s) already exist, $($NodeCount - $totalReadyVMCount) node(s) will be created."

        return @($serviceVMTable, $hpcNodeImageName)
    }
    catch
    {
        throw
    }
}


function GetVMImageByVMSourceImageName
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String]$SourceImageName,

        [Parameter(Mandatory=$true)]
        [Object[]] $FromImageList
    )
  
    if($null -ne $FromImageList -and $FromImageList.Count -gt 0)
    {
        foreach($img in $FromImageList)
        {
            if($img.ImageName -eq $SourceImageName)
            {
                return $img
            }
            if($img.PSObject.Properties.Name -contains "OSDiskConfiguration" -and $img.OSDiskConfiguration.PSObject.Properties.Name -contains "Name")
            {
                if($img.OSDiskConfiguration.Name -eq $SourceImageName)
                {
                    return $img
                }
            }
        }
    }

    return $null
}

function CheckAzureBurstConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $AzureBustConfig,

        [Parameter(Mandatory=$false)]
        [Object[]] $Certificates = @()
    )

    $errMsgs = @()
    $certIds = @()
    if($null -ne $Certificates)
    {
        $certIds = @($Certificates | %{$_.Id})
    }
    $tempNames = @()
    $svcNames = @()
    foreach($xmlTemp in $AzureBustConfig.AzureNodeTemplate)
    {
        $tempErrMsgs = @()
        $tempName = $xmlTemp.TemplateName.Trim()
        $tempNames += $tempName

        $svcName = $xmlTemp.ServiceName.Trim()
        try
        {
            ValidateServiceNaming -ServiceName $svcName | Out-Null
        }
        catch
        {
            $tempErrMsgs += $_.Exception.Message
        }

        try
        {
            ValidateStorageAccountNaming -StorageAccountName $xmlTemp.StorageAccount.Trim() | Out-Null
        }
        catch
        {
            $tempErrMsgs += $_.Exception.Message
        }

        $subId = $xmlTemp.SubscriptionId.Trim()
        try
        {
            [System.Guid]::Parse($subId) | Out-Null
        }
        catch
        {
            $tempErrMsgs += "The 'SubscriptionId' is not a valid Guid: $subId"
        }

        $svcNames += $svcName

        $certId = [Int]$xmlTemp.CertificateId.Trim()
        if($certIds -notcontains $certId)
        {
            $tempErrMsgs += "The 'CertificateId' $certId cannot be found."
        }

        if(($null -ne $xmlTemp["NodeCount"]) -and ($null -eq $xmlTemp["RoleSize"]))
        {
            $tempErrMsgs += "'RoleSize' must be specified when 'NodeCount' is specified."
        }

        if($null -ne $xmlTemp["Proxy"])
        {
            if($xmlTemp.Proxy.UsesStaticProxyCount.Trim() -eq "true")
            {
                if($null -ne $xmlTemp.Proxy["ProxyMultiplicity"])
                {
                    $proxyNum = [int]$xmlTemp.Proxy.ProxyMultiplicity.Trim()
                    if($proxyNum -gt 10 -or $proxyNum -lt 2)
                    {
                        $tempErrMsgs += "The value of 'ProxyMultiplicity' is invalid: $proxyNum (allowed range 2 - 10)."
                    }
                }
            }
            else
            {
                if($null -ne $xmlTemp.Proxy["ProxyRatio"])
                {
                    $prxoyRatio = [int]$xmlTemp.Proxy.ProxyRatio.Trim()
                    if($prxoyRatio -gt 400 -or $prxoyRatio -lt 100)
                    {
                        $tempErrMsgs += "The value of 'ProxyRatio' is invalid: $prxoyRatio (allowed range 100 - 400)."
                    }
                }
            
                if($null -ne $xmlTemp.Proxy["ProxyRatioBase"])
                {
                    $prxoyRatioBase = [int]$xmlTemp.Proxy.ProxyRatioBase.Trim()
                    if($prxoyRatioBase -gt 600 -or $prxoyRatioBase -lt 200)
                    {
                        $tempErrMsgs += "The value of 'ProxyRatioBase' is invalid: $prxoyRatioBase (allowed range 200 - 600)."
                    }
                }
            }
        }

        foreach($msg in $tempErrMsgs)
        {
            $errMsgs += "[AzureNodeTemplate $tempName] $msg"
        }

        $tempErrMsgs.Clear()
    }

    foreach($group in ($tempNames | group))
    {
        if($group.Count -gt 1)
        {
            $errMsgs += "[AzureNodeTemplate] There are two or more AzureNodeTemplates with the same 'TemplateName': " + $group.Name
        }
    }

    foreach($group in ($svcNames | group))
    {
        if($group.Count -gt 1)
        {
            $errMsgs += "[AzureNodeTemplate] The cloud service name $($group.Name) is specified in two or more AzureNodeTemplates."
        }
    } 
    
    return $errMsgs
}

function GetAutoGrowShrinkConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $AutoGrowShrinkConfig
    )
    $configInfo = @{}
    $configInfo["CertificateId"] = [Int]$AutoGrowShrinkConfig.CertificateId.Trim()
    ## 0 means default grow/shrink interval
    $configInfo["GrowInterval"] = 0
    $configInfo["ShrinkInterval"] = 0
    if($null -ne $AutoGrowShrinkConfig["GrowInterval"])
    {
        $configInfo["GrowInterval"] = [Int]$AutoGrowShrinkConfig.GrowInterval.Trim()
    }
    if($null -ne $AutoGrowShrinkConfig["ShrinkInterval"])
    {
        $configInfo["ShrinkInterval"] = [Int]$AutoGrowShrinkConfig.ShrinkInterval.Trim()
    }
    $configInfo['Enable'] = $true
    if(($null -ne $AutoGrowShrinkConfig["Enable"]) -and ($AutoGrowShrinkConfig.Enable.Trim() -eq "false"))
    {
        $configInfo['Enable'] = $false
    }

    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetCertificatesConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $IaaSClusterConfig
    )
    $certList = @()
    $certElements = @()
    if($null -ne $IaaSClusterConfig["Certificates"])
    {
        $certElements += $IaaSClusterConfig.Certificates.Certificate
    }
    if(($null -ne $IaaSClusterConfig["AzureBurst"]) -and ($null -ne $IaaSClusterConfig.AzureBurst["Certificate"]))
    {
        $certElements += $IaaSClusterConfig.AzureBurst.Certificate
    }
    
    foreach($elem in $certElements)
    {
        $cert = @{}
        $cert["Id"] = [Int]$elem.Id.Trim()
        $cert["PfxFile"] = $elem.PfxFile.Trim()
        $cert["Password"] = $null
        if($null -ne $elem["Password"])
        {
            $cert["Password"] = $elem.Password.Trim()
        }

        $certConfig = New-Object psobject -Property $cert
        $certList += $certConfig
    }

    return $certList  
}


function PromptForPfxFilePassword
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $FilePath, 

        [Parameter(Mandatory=$false)]
        [String] $Password = ""
    )
    
    $retry = 0
    $pswToTry = $Password
    while($true)
    {
        try
        {
            $pfxCert = GetCertificate $FilePath $pswToTry
            break
        }
        catch
        {
            if($_.Exception.HResult -eq 0x80070056)
            {
                if([String]::IsNullOrEmpty($pswToTry))
                {
                    $prompt = "The certificate file $FilePath is password protected. Please input the password"
                }
                else
                {
                    $prompt = "The password for the certificate file $FilePath is incorrect. Please input the correct password"
                }
                    
                if($retry -lt 3)
                {
                    $secPsw = Read-Host -Prompt $prompt -AsSecureString
                    $pswToTry = ConvertSecureStrToPlain -SecurePassword $secPsw 
                    $retry++
                }
                else
                {
                    throw "The password for the certificate file $FilePath is incorrect."
                }
            }
            elseif($_.Exception.HResult -eq 0x80092009)
            {
                throw "The file $FilePath is not a valid PFX file."
            }
            else
            {
                throw
            }
        }
    }
    
    if(-not $pfxCert.HasPrivateKey)
    {
        throw "The certificate file $pfxFile doesn't contain a private key."
    }
    return $pswToTry
}

function ValidateCertificates
{
    param
    (
        [Parameter(Mandatory=$true)]
        [Object[]] $Certificates
    )
    
    $certInfoList = @()
    foreach($cert in $Certificates)
    {   
        $pfxFile = $cert.PfxFile
        $password = ""
        if(-not [string]::IsNullOrEmpty($cert.Password))
        {
            $password = $cert.Password
        }

        $retry = 0
        while($true)
        {
            try
            {
                $pfxCert = GetCertificate $pfxFile $password
                break
            }
            catch
            {
                if($_.Exception.HResult -eq 0x80070056)
                {
                    if([String]::IsNullOrEmpty($password))
                    {
                        $prompt = "The certificate file $pfxFile is password protected. Please input the password"
                    }
                    else
                    {
                        $prompt = "The password for the certificate file $pfxFile is incorrect. Please input the correct password"
                    }
                    
                    if($retry -lt 3)
                    {
                        $secPsw = Read-Host -Prompt $prompt -AsSecureString
                        $password = ConvertSecureStrToPlain -SecurePassword $secPsw 
                        $retry++
                    }
                    else
                    {
                        throw "The password for the certificate file $pfxFile is incorrect."
                    }
                }
                elseif($_.Exception.HResult -eq 0x80092009)
                {
                    throw "The file $pfxFile is not a valid PFX file."
                }
                else
                {
                    throw
                }
            }
        }

        if(-not $pfxCert.HasPrivateKey)
        {
            throw "The certificate file $pfxFile doesn't contain a private key."
        }
        
        $certInfo = @{}
        $certInfo["Id"] = $cert.Id
        $certInfo["CertName"] = [IO.Path]::GetFileNameWithoutExtension($pfxFile)
        $certInfo["PfxContent"] = [IO.File]::ReadAllBytes($pfxFile)
        $certInfo["Password"] = $password
        $certInfo["Thumbprint"] = $pfxCert.Thumbprint
        $certInfoPSObject = New-Object psobject -Property $certInfo
        $certInfoList += $certInfoPSObject
    }

    return $certInfoList
}

function GetCommonConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $IaaSClusterConfig
    )
    
    $configInfo = @{}
    $configInfo['SubscriptionId'] = $null
    $configInfo['SubscriptionName'] = $null
    if($null -ne $IaaSClusterConfig.Subscription["SubscriptionName"])
    {
        $configInfo['SubscriptionName'] = $IaaSClusterConfig.Subscription.SubscriptionName.Trim()
    }

    if($null -ne $IaaSClusterConfig.Subscription["SubscriptionId"])
    {
        $configInfo['SubscriptionId'] = $IaaSClusterConfig.Subscription.SubscriptionId.Trim()
    }
    
    $configInfo['DefaultStorageAccount'] = $IaaSClusterConfig.Subscription.StorageAccount.Trim()
    $configInfo['Location'] = $IaaSClusterConfig.Location.Trim()
    $configInfo['Language'] = $null
    if($null -ne $config["Language"])
    {
        $configInfo['Language'] = $IaaSClusterConfig.Language.Trim()
    }
    $configInfo['DefaultAffinityGroup'] = $null
    if($null -ne $config["AffinityGroup"])
    {
        $configInfo['DefaultAffinityGroup'] = $IaaSClusterConfig.AffinityGroup.Trim()
    }

    $configInfo['VNetName'] = $IaaSClusterConfig.VNet.VNetName.Trim()
    $configInfo['SubnetName'] = $IaaSClusterConfig.VNet.SubnetName.Trim()

    # We put the DomainFQDN in 'Domain' to CommonConfig
    $configInfo['DomainFQDN'] = $IaaSClusterConfig.Domain.DomainFQDN.Trim()
    $configInfo['DomainNetBios'] = $configInfo['DomainFQDN'].Split(".")[0].ToUpper()

    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetDomainConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $DomainConfig
    )
    
    $configInfo = @{}
    $configInfo['DomainFQDN'] = $DomainConfig.DomainFQDN.Trim()
    $configInfo['DomainNetBios'] = $configInfo['DomainFQDN'].Split(".")[0].ToUpper()
    $configInfo['DCOption'] = $DomainConfig.DCOption.Trim()
    $configInfo['DomainController'] = $null
    if($null -ne $DomainConfig['DomainController'])
    {
        $dcInfo = @{}
        $dcInfo['VMName'] = $DomainConfig.DomainController.VMName.Trim()
        $dcInfo['ServiceName'] = $DomainConfig.DomainController.ServiceName.Trim()
        $dcInfo['VMSize'] = $DomainConfig.DomainController.VMSize.Trim()
        $dcInfo['AffinityGroup'] = $null
        if($null -ne $DomainConfig.DomainController["AffinityGroup"])
        {
            $dcInfo['AffinityGroup'] = $DomainConfig.DomainController.AffinityGroup.Trim()
        }

        $configInfo['DomainController'] = New-Object psobject -Property $dcInfo
    }

    $configInfo['DNSForwarder'] = $null
    if($null -ne $DomainConfig['DNSForwarder'])
    {
        $configInfo['DNSForwarder'] = $DomainConfig.DNSForwarder.Trim()
    }

    $configInfo['NoWindowsAutoUpdate'] = $false
    if(($null -ne $DomainConfig["NoWindowsAutoUpdate"]) -and ($DomainConfig.NoWindowsAutoUpdate.Trim() -ne "false"))
    {
        $configInfo['NoWindowsAutoUpdate'] = $true
    }
    
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetDatabaseConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $DBConfig
    )
    
    $configInfo = @{}
    $configInfo['DBOption'] = $DBConfig.DBOption.Trim()
    $configInfo['DBServer'] = $null
    $configInfo['DBConnection'] = $null
    $configInfo['DBVersion'] = $null
    if($null -ne $DBConfig['DBVersion'])
    {
        $configInfo['DBVersion'] = $DBConfig.DBVersion.Trim()
    }

    if($null -ne $DBConfig['DBServer'])
    {
        $dbServerInfo = @{}
        $dbServerInfo['VMName'] = $DBConfig.DBServer.VMName.Trim()
        $dbServerInfo['ServiceName']  = $DBConfig.DBServer.ServiceName.Trim()
        $dbServerInfo['VMSize']  = $DBConfig.DBServer.VMSize.Trim()
        $dbServerInfo['AffinityGroup']  = $null
        if($null -ne $DBConfig.DBServer["AffinityGroup"])
        {
            $dbServerInfo['AffinityGroup'] = $DBConfig.DBServer.AffinityGroup.Trim()
        }

        $dbServerInfo['DataDiskSizeInGB'] = 200
        if($null -ne $DBConfig.DBServer["DataDiskSizeInGB"])
        {
            $dbServerInfo['DataDiskSizeInGB'] = [int]$DBConfig.DBServer.DataDiskSizeInGB.Trim()
        }

        $configInfo['DBServer'] = New-Object psobject -Property $dbServerInfo
    }
    
    if($null -ne $DBConfig['DBConnection'])
    {
        $dbConnectionInfo = @{}
        $dbConnectionInfo['ServerInstance'] = $DBConfig.DBConnection.ServerInstance.Trim()
        $dbConnectionInfo['ManagementDB'] = "HpcManagement"
        $dbConnectionInfo['SchedulerDB'] = "HpcScheduler"
        $dbConnectionInfo['MonitoringDB'] = "HpcMonitoring"
        $dbConnectionInfo['ReportingDB'] = "HpcReporting"
        $dbConnectionInfo['DiagnosticsDB'] = "HpcDiagnostics"
        if($null -ne $DBConfig.DBConnection["ManagementDB"])
        {
            $dbConnectionInfo['ManagementDB'] = $DBConfig.DBConnection.ManagementDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["SchedulerDB"])
        {
            $dbConnectionInfo['SchedulerDB'] = $DBConfig.DBConnection.SchedulerDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["MonitoringDB"])
        {
            $dbConnectionInfo['MonitoringDB'] = $DBConfig.DBConnection.MonitoringDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["ReportingDB"])
        {
            $dbConnectionInfo['ReportingDB'] = $DBConfig.DBConnection.ReportingDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["DiagnosticsDB"])
        {
            $dbConnectionInfo['DiagnosticsDB'] = $DBConfig.DBConnection.DiagnosticsDB.Trim()
        }

        $configInfo['DBConnection'] = New-Object psobject -Property $dbConnectionInfo
    }
    
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}


function GetHeadNodeConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $HeadNodeConfig
    )
    
    $configInfo = @{}
    $configInfo['VMName'] = $HeadNodeConfig.VMName.Trim()
    $configInfo['ServiceName'] = $HeadNodeConfig.ServiceName.Trim()
    $configInfo['VMSize'] = $HeadNodeConfig.VMSize.Trim()
    $configInfo['AffinityGroup'] = $null
    if($null -ne $HeadNodeConfig["AffinityGroup"])
    {
        $configInfo['AffinityGroup'] = $HeadNodeConfig.AffinityGroup.Trim()
    }

    $configInfo['ImageFamily'] = $null
    if($null -ne $HeadNodeConfig["ImageFamily"])
    {
        $configInfo['ImageFamily'] = $HeadNodeConfig.ImageFamily.Trim()
    }

    $configInfo['EnableRESTAPI'] = $false
    if(($null -ne $HeadNodeConfig["EnableRESTAPI"]) -and ($HeadNodeConfig.EnableRESTAPI.Trim() -ne "false"))
    {
        $configInfo['EnableRESTAPI'] = $true
    }

    $configInfo['EnableWebPortal'] = $false
    if(($null -ne $HeadNodeConfig["EnableWebPortal"]) -and ($HeadNodeConfig.EnableWebPortal.Trim() -ne "false"))
    {
        $configInfo['EnableWebPortal'] = $true
    }

    $configInfo['PreConfigScript'] = $null
    if($null -ne $HeadNodeConfig["PreConfigScript"])
    {
        $configInfo['PreConfigScript'] = $HeadNodeConfig.PreConfigScript.Trim()
    }

    $configInfo['PostConfigScript'] = $null
    if($null -ne $HeadNodeConfig["PostConfigScript"])
    {
        $configInfo['PostConfigScript'] = $HeadNodeConfig.PostConfigScript.Trim()
    }
    
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetNodeGroupConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $NodeGroupConfig
    )
    
    $configInfo = @{}
    $configInfo['NodeGroupName'] = $NodeGroupConfig.Name
    if($NodeGroupConfig.Name -eq "ComputeNodes")
    {
        $configInfo['NodeType'] = "ComputeNode"
    }
    elseif($NodeGroupConfig.Name -eq "BrokerNodes")
    {
        $configInfo['NodeType'] = "BrokerNode"
    }
    elseif($NodeGroupConfig.Name -eq "LinuxComputeNodes")
    {
        $configInfo['NodeType'] = "LinuxNode"
    }

    $configInfo['VMNamePattern'] = $NodeGroupConfig.VMNamePattern.Trim()
    $configInfo['VMSize'] = $NodeGroupConfig.VMSize.Trim()
    $configInfo['NodeCount'] = [int]$NodeGroupConfig.NodeCount.Trim()
    if($null -ne $NodeGroupConfig["ServiceName"])
    {
        $configInfo['ServiceName'] = $NodeGroupConfig.ServiceName.Trim()
        $configInfo['ServiceNamePattern'] = $null
    }
    else
    {
        $configInfo['ServiceName'] = $null
        $configInfo['ServiceNamePattern'] = $NodeGroupConfig.ServiceNamePattern.Trim()
        $configInfo['MaxNodeCountPerService'] = [int]$NodeGroupConfig.MaxNodeCountPerService.Trim()
        $configInfo['StorageAccountNamePattern'] = $NodeGroupConfig.StorageAccountNamePattern.Trim()
    }

    $configInfo['AffinityGroup'] = $null
    if($null -ne $NodeGroupConfig["AffinityGroup"])
    {
        $configInfo['AffinityGroup'] = $NodeGroupConfig.AffinityGroup.Trim()
    }

    $configInfo['ImageName'] = $null
    $configInfo['ImageHPCPackInstalled'] = $false
    $configInfo['OSVersion'] = $null
    if($null -ne $NodeGroupConfig["ImageName"])
    {
        $configInfo['ImageName'] = $NodeGroupConfig["ImageName"].InnerText.Trim()
        if($NodeGroupConfig["ImageName"].HasAttribute("HPCPackInstalled"))
        {
            if($NodeGroupConfig["ImageName"].GetAttribute("HPCPackInstalled").Trim() -eq "true")
            {
                $configInfo['ImageHPCPackInstalled'] = $true
            }
        }
    }
    elseif($null -ne $NodeGroupConfig["OSVersion"])
    {
        $configInfo['OSVersion'] = $NodeGroupConfig.OSVersion.Trim()
    }
    
    $configInfo['VMExtensions'] = $null
    if($null -ne $NodeGroupConfig["VMExtensions"])
    {
        $extensions = @()
        foreach($xmlExt in $NodeGroupConfig.VMExtensions.VMExtension)
        {
            $extName = $xmlExt.ExtensionName.Trim()
            $extPublisher = $xmlExt.Publisher.Trim()
            $extVersion = $xmlExt.Version.Trim()
            $extPubConfig = $null
            $extPrivateConfig = $null
            if($null -ne $xmlExt["PublicConfiguration"])
            {
                $extPubConfig = $xmlExt.PublicConfiguration.Trim()
            }

            if($null -ne $xmlExt["PrivateConfiguration"])
            {
                $extPrivateConfig = $xmlExt.PrivateConfiguration.Trim()
            }
           
            $extensions += NewVMExtensionConfig -ExtensionName $extName -Publisher $extPublisher -Version $extVersion -PublicConfiguration $extPubConfig -PrivateConfiguration $extPrivateConfig
        }

        $configInfo['VMExtensions'] = $extensions
    }

    $configInfo['SSHKeyPairForRoot'] = $null
    if($null -ne $NodeGroupConfig["SSHKeyPairForRoot"])
    {
        $sshCertificate = @{}
        $sshCertificate["PfxFile"] = $NodeGroupConfig.SSHKeyPairForRoot.PfxFile.Trim()
        if($null -ne $NodeGroupConfig.SSHKeyPairForRoot["Password"])
        {
            $sshCertificate["Password"] = $NodeGroupConfig.SSHKeyPairForRoot.Password.Trim()
        }
        else
        {
            $sshCertificate["Password"] = $null
        }

        $configInfo['SSHKeyPairForRoot'] = New-Object psobject -Property $sshCertificate
    }
        
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetNodeGroupServiceStorageTable
{
    param
    (
        [Parameter(Mandatory=$true)]
        [psobject] $NodeGroup,

        [Parameter(Mandatory=$true)]
        [string] $DefaultStorageAccount
    )

    $serviceStorageTable = @{}
    if($null -ne $NodeGroup.ServiceName)
    {
        $serviceStorageTable[$NodeGroup.ServiceName] = $DefaultStorageAccount
    }
    else
    {
        $nodeSetCount = [Math]::Ceiling($NodeGroup.NodeCount/$NodeGroup.MaxNodeCountPerService)
        $serviceNames = @(GetNamesByPattern -NamePattern $NodeGroup.ServiceNamePattern -Count $nodeSetCount)
        $storageAccounts = @(GetNamesByPattern -NamePattern $NodeGroup.StorageAccountNamePattern -Count $nodeSetCount)
        for($i=0; $i -lt $nodeSetCount; $i++)
        {
            $serviceStorageTable[$serviceNames[$i]] = $storageAccounts[$i]
        }
    }

    return $serviceStorageTable
}
# SIG # Begin signature block
# MIIaxwYJKoZIhvcNAQcCoIIauDCCGrQCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUqFL8fROmCjtQegi0iq62PNhx
# OCCgghWCMIIEwzCCA6ugAwIBAgITMwAAAHJp9pDCFDtJoQAAAAAAcjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUwMzIwMTczMjA0
# WhcNMTYwNjIwMTczMjA0WjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OjMxQzUtMzBCQS03QzkxMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoxPVJ2iBlQVE
# sJuqKGMDs9En3ra2ZyP2mieV2SfpWI7AXCdUfRHZ7B+712Ff5z/jWtu5iNakH+yw
# OsPZy911xkYMU2T9mm1IXZcHoKEMvj6SL5H3q2dq0SX5n3iqXJcxf4wwjDu9cKit
# 5v8t+rrQxeVIuseUK34MXOH/NYpAv1H0FU0vm0MjOh7ltNRzua86yAgrQBSj7gML
# NybNLvq+M9b3JBVF6cFLmII8syS9LSf7drlwWzPRJ71Q/6+TVq59UtTQO217Juvk
# 9BdtStzFmMyR+7VdVxybCwxPr6dW/2U7oouf4UxtszHDMtHw7XEnA2jg2iq6ODkZ
# frNXUQCMhQIDAQABo4IBCTCCAQUwHQYDVR0OBBYEFND9obr0IhS4jv6FL059LwBC
# pwFTMB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZI
# hvcNAQEFBQADggEBACJEd3U6Z3Bw2w+djrTKkrwYGl8Ys2vdoryZDDm9ybp6jjCk
# QCKLBGssgmRVJcEaYjUTkqGomLk5Mt4r5X6S+Zf1lgF3IKX9tqVbNlX1qMmRZUQ4
# 4Xfg7Ml8SuB1UqFU0DvuoIoQndg51sX2yA4Tahq0xD/ilyga1NHQ3jgbw5QKBZDt
# VbGGwfUgcNCwbZGDZCQtrRTcRmr4B7/tK4xFFOfvrKLEA8MByHdzmAYlIMsPM/fX
# YCl95S60wLsCjC5aJHBzM5NN2ZRQSLYLg0RKtrBoR5Zq+RlIBXKsS0hKgRbk9WAj
# uB5MvW3T/tl40KdnBlT9lCGh677+NM8EykDf1UIwggTsMIID1KADAgECAhMzAAAB
# Cix5rtd5e6asAAEAAAEKMA0GCSqGSIb3DQEBBQUAMHkxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBMB4XDTE1MDYwNDE3NDI0NVoXDTE2MDkwNDE3NDI0NVowgYMxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIx
# HjAcBgNVBAMTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJL8bza74QO5KNZG0aJhuqVG+2MWPi75R9LH7O3HmbEm
# UXW92swPBhQRpGwZnsBfTVSJ5E1Q2I3NoWGldxOaHKftDXT3p1Z56Cj3U9KxemPg
# 9ZSXt+zZR/hsPfMliLO8CsUEp458hUh2HGFGqhnEemKLwcI1qvtYb8VjC5NJMIEb
# e99/fE+0R21feByvtveWE1LvudFNOeVz3khOPBSqlw05zItR4VzRO/COZ+owYKlN
# Wp1DvdsjusAP10sQnZxN8FGihKrknKc91qPvChhIqPqxTqWYDku/8BTzAMiwSNZb
# /jjXiREtBbpDAk8iAJYlrX01boRoqyAYOCj+HKIQsaUCAwEAAaOCAWAwggFcMBMG
# A1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSJ/gox6ibN5m3HkZG5lIyiGGE3
# NDBRBgNVHREESjBIpEYwRDENMAsGA1UECxMETU9QUjEzMDEGA1UEBRMqMzE1OTUr
# MDQwNzkzNTAtMTZmYS00YzYwLWI2YmYtOWQyYjFjZDA1OTg0MB8GA1UdIwQYMBaA
# FMsR6MrStBZYAck3LjMWFrlMmgofMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY0NvZFNpZ1BDQV8w
# OC0zMS0yMDEwLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljQ29kU2lnUENBXzA4LTMx
# LTIwMTAuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCmqFOR3zsB/mFdBlrrZvAM2PfZ
# hNMAUQ4Q0aTRFyjnjDM4K9hDxgOLdeszkvSp4mf9AtulHU5DRV0bSePgTxbwfo/w
# iBHKgq2k+6apX/WXYMh7xL98m2ntH4LB8c2OeEti9dcNHNdTEtaWUu81vRmOoECT
# oQqlLRacwkZ0COvb9NilSTZUEhFVA7N7FvtH/vto/MBFXOI/Enkzou+Cxd5AGQfu
# FcUKm1kFQanQl56BngNb/ErjGi4FrFBHL4z6edgeIPgF+ylrGBT6cgS3C6eaZOwR
# XU9FSY0pGi370LYJU180lOAWxLnqczXoV+/h6xbDGMcGszvPYYTitkSJlKOGMIIF
# vDCCA6SgAwIBAgIKYTMmGgAAAAAAMTANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMTAwODMx
# MjIxOTMyWhcNMjAwODMxMjIyOTMyWjB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALJyWVwZMGS/HZpgICBC
# mXZTbD4b1m/My/Hqa/6XFhDg3zp0gxq3L6Ay7P/ewkJOI9VyANs1VwqJyq4gSfTw
# aKxNS42lvXlLcZtHB9r9Jd+ddYjPqnNEf9eB2/O98jakyVxF3K+tPeAoaJcap6Vy
# c1bxF5Tk/TWUcqDWdl8ed0WDhTgW0HNbBbpnUo2lsmkv2hkL/pJ0KeJ2L1TdFDBZ
# +NKNYv3LyV9GMVC5JxPkQDDPcikQKCLHN049oDI9kM2hOAaFXE5WgigqBTK3S9dP
# Y+fSLWLxRT3nrAgA9kahntFbjCZT6HqqSvJGzzc8OJ60d1ylF56NyxGPVjzBrAlf
# A9MCAwEAAaOCAV4wggFaMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFMsR6MrS
# tBZYAck3LjMWFrlMmgofMAsGA1UdDwQEAwIBhjASBgkrBgEEAYI3FQEEBQIDAQAB
# MCMGCSsGAQQBgjcVAgQWBBT90TFO0yaKleGYYDuoMW+mPLzYLTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTAfBgNVHSMEGDAWgBQOrIJgQFYnl+UlE/wq4QpTlVnk
# pDBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEE
# SDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2Nl
# cnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDANBgkqhkiG9w0BAQUFAAOCAgEAWTk+
# fyZGr+tvQLEytWrrDi9uqEn361917Uw7LddDrQv+y+ktMaMjzHxQmIAhXaw9L0y6
# oqhWnONwu7i0+Hm1SXL3PupBf8rhDBdpy6WcIC36C1DEVs0t40rSvHDnqA2iA6VW
# 4LiKS1fylUKc8fPv7uOGHzQ8uFaa8FMjhSqkghyT4pQHHfLiTviMocroE6WRTsgb
# 0o9ylSpxbZsa+BzwU9ZnzCL/XB3Nooy9J7J5Y1ZEolHN+emjWFbdmwJFRC9f9Nqu
# 1IIybvyklRPk62nnqaIsvsgrEA5ljpnb9aL6EiYJZTiU8XofSrvR4Vbo0HiWGFzJ
# NRZf3ZMdSY4tvq00RBzuEBUaAF3dNVshzpjHCe6FDoxPbQ4TTj18KUicctHzbMrB
# 7HCjV5JXfZSNoBtIA1r3z6NnCnSlNu0tLxfI5nI3EvRvsTxngvlSso0zFmUeDord
# EN5k9G/ORtTTF+l5xAS00/ss3x+KnqwK+xMnQK3k+eGpf0a7B2BHZWBATrBC7E7t
# s3Z52Ao0CW0cgDEf4g5U3eWh++VHEK1kmP9QFi58vwUheuKVQSdpw5OPlcmN2Jsh
# rg1cnPCiroZogwxqLbt2awAdlq3yFnv2FoMkuYjPaqhHMS+a3ONxPdcAfmJH0c6I
# ybgY+g5yjcGjPa8CQGr/aZuW4hCoELQ3UAjWwz0wggYHMIID76ADAgECAgphFmg0
# AAAAAAAcMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAX
# BgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290
# IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0wNzA0MDMxMjUzMDlaFw0yMTA0MDMx
# MzAzMDlaMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAf
# BgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQTCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJ+hbLHf20iSKnxrLhnhveLjxZlRI1Ctzt0YTiQP7tGn
# 0UytdDAgEesH1VSVFUmUG0KSrphcMCbaAGvoe73siQcP9w4EmPCJzB/LMySHnfL0
# Zxws/HvniB3q506jocEjU8qN+kXPCdBer9CwQgSi+aZsk2fXKNxGU7CG0OUoRi4n
# rIZPVVIM5AMs+2qQkDBuh/NZMJ36ftaXs+ghl3740hPzCLdTbVK0RZCfSABKR2YR
# JylmqJfk0waBSqL5hKcRRxQJgp+E7VV4/gGaHVAIhQAQMEbtt94jRrvELVSfrx54
# QTF3zJvfO4OToWECtR0Nsfz3m7IBziJLVP/5BcPCIAsCAwEAAaOCAaswggGnMA8G
# A1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFCM0+NlSRnAK7UD7dvuzK7DDNbMPMAsG
# A1UdDwQEAwIBhjAQBgkrBgEEAYI3FQEEAwIBADCBmAYDVR0jBIGQMIGNgBQOrIJg
# QFYnl+UlE/wq4QpTlVnkpKFjpGEwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5ghB5rRahSqClrUxzWPQHEy5lMFAGA1UdHwRJ
# MEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1
# Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRIMEYwRAYIKwYB
# BQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0Um9vdENlcnQuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEB
# BQUAA4ICAQAQl4rDXANENt3ptK132855UU0BsS50cVttDBOrzr57j7gu1BKijG1i
# uFcCy04gE1CZ3XpA4le7r1iaHOEdAYasu3jyi9DsOwHu4r6PCgXIjUji8FMV3U+r
# kuTnjWrVgMHmlPIGL4UD6ZEqJCJw+/b85HiZLg33B+JwvBhOnY5rCnKVuKE5nGct
# xVEO6mJcPxaYiyA/4gcaMvnMMUp2MT0rcgvI6nA9/4UKE9/CCmGO8Ne4F+tOi3/F
# NSteo7/rvH0LQnvUU3Ih7jDKu3hlXFsBFwoUDtLaFJj1PLlmWLMtL+f5hYbMUVbo
# nXCUbKw5TNT2eb+qGHpiKe+imyk0BncaYsk9Hm0fgvALxyy7z0Oz5fnsfbXjpKh0
# NbhOxXEjEiZ2CzxSjHFaRkMUvLOzsE1nyJ9C/4B5IYCeFTBm6EISXhrIniIh0EPp
# K+m79EjMLNTYMoBMJipIJF9a6lbvpt6Znco6b72BJ3QGEe52Ib+bgsEnVLaxaj2J
# oXZhtG6hE6a/qkfwEm/9ijJssv7fUciMI8lmvZ0dhxJkAj0tr1mPuOQh5bWwymO0
# eFQF1EEuUKyUsKV4q7OglnUa2ZKHE3UiLzKoCG6gW4wlv6DvhMoh1useT8ma7kng
# 9wFlb4kLfchpyOZu6qeXzjEp/w7FW1zYTRuh2Povnj8uVRZryROj/TGCBK8wggSr
# AgEBMIGQMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAh
# BgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBAhMzAAABCix5rtd5e6as
# AAEAAAEKMAkGBSsOAwIaBQCggcgwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFJGC
# AFkEzB5jVHpv8ldVeGLB+WgoMGgGCisGAQQBgjcCAQwxWjBYoDaANABNAGkAYwBy
# AG8AcwBvAGYAdAAgAEgAUABDACAAUABhAGMAawAgADIAMAAxADIAIABSADKhHoAc
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL0hQQzANBgkqhkiG9w0BAQEFAASCAQBk
# hP/1mQnOgTGN5Zabzlrz8R/NAJgBVSkpJfpC4KMEhu066/vHOOC8gnaKKoD2IqJB
# c4BjvZsCiQzJJPmwNxDdqCt8OdeZd0HtSMgn8tRR/px1Han2RohTQkNo/RfwGVaj
# lrG0MIwY3357a7XjifLK8o38yTfWicGqgvDSqq6ns7u5ZhrXNTJVSTSOxiWzMsZp
# vbpDy5IHUiOpE1MT2sGyVeKFTwjzaOikOb9xgMvrj+ppBL9AnTuAtWZlwUIhnYU9
# KbJ9ihxU5E/nv/irWLaNlReq5656mShmPUao/CWLgNlLZ/JcBZJW790ZzSfT3arg
# BEk1oFgSoLyBqQpePv3yoYICKDCCAiQGCSqGSIb3DQEJBjGCAhUwggIRAgEBMIGO
# MHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAfBgNVBAMT
# GE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQQITMwAAAHJp9pDCFDtJoQAAAAAAcjAJ
# BgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkDMQsGCSqGSIb3DQEHATAcBgkqhkiG9w0B
# CQUxDxcNMTUwNzA2MTcyMzExWjAjBgkqhkiG9w0BCQQxFgQUp0wYC49uA2JKS/LF
# R7KK8zErccowDQYJKoZIhvcNAQEFBQAEggEAY/UjbJAOF8GSXTE5rf/871zxxrP6
# QEyZXSZPBwSs1IrxtFCB6fsPf+lPMVvGX6LoIQs5X3RXNVTTnbHB1RmQvlMfCf1z
# 6HlqtyQlOcdMeRdfxRUs5uWUKnM6pTNJckpLzLkNTWkZ44TABpPMOzu3C00vTbbZ
# mto3sUM2kofUNw6FcSHM9xqbz/dZXoim/LbgmosVFsrC1pRO+ks5LsYtPnCCLRPJ
# Aypj/zBDQwfq0gu/fym7cSj8oxPQLJGtFe5yQe/vGgq2/x3qOwB2PJEJFQBgIGm8
# iusjM8IZXw6tiz8L40EfKYODXW1/iZ+V/JXZ1Uk2JUZ68M8g1LZt4VAZaA==
# SIG # End signature block
