<#
.SYNOPSIS
  Create a VNET
.DESCRIPTION
  Create a VNET
.EXAMPLE
  The following example will create an affinity group wide virtual network "contosovnet" in affinity group "contoso-afgroup" with address space 192.168.0.0/20, 
  and a subnet named "Subnet-1"  with address space 192.168.0.0/23
    CreateNewVNet -VNetName "contosovnet" -AffinityGroupName "contoso-afgroup"

  The following example will create an affinity group wide virtual network "contosovnet" in affinity group "contoso-afgroup" with address space 10.0.0.0/20, 
  and a subnet named "Subnet22"  with address space 10.0.0.0/23
    CreateNewVNet -VNetName "contosovnet" -AffinityGroupName "contoso-afgroup" -AddressSpace "10.0.0.0/20" -SubnetName "Subnet22" -SubnetAddressSpace "10.0.0.0/23"

  The following example will create a region wide virtual network "contosoRegional" in the region "East Asia" with address space 10.0.0.0/20, 
  and a subnet named "Subnet-1"  with address space 10.0.0.0/23
    CreateNewVNet -VNetName "contosoRegional" -Location "East Asia" -AddressSpace "10.0.0.0/20" -SubnetAddressSpace "10.0.0.0/23"
#>
function CreateNewVNet
{
    Param
    (
        # The name of VNet
        [Parameter(Mandatory=$true)]
        [String] $VNetName,
		
        # The affinity group name for a affinity group wide virtual network
        [Parameter(Mandatory=$true, ParameterSetName="AffinityGroupVNet")]
        [String] $AffinityGroupName,

        # The location for a region wide virtual network
        [Parameter(Mandatory=$true, ParameterSetName="RegionalVNet")]
        [String] $Location,

        # The address space for the virtual network
        [Parameter(Mandatory=$false)]
        [String] $AddressSpace = "192.168.0.0/20",

        # Subnet name
        [Parameter(Mandatory=$false)]
        [String] $SubnetName = "Subnet-1",

        # The address space for the subnet
        [Parameter(Mandatory=$false)]
        [String] $SubnetAddressSpace = "192.168.0.0/23"
    )


    if($PsCmdlet.ParameterSetName -eq "AffinityGroupVNet")
    {
        TraceInfo "Creating virtual network $VNetName with Subnet $SubnetName in AffinityGroup $AffinityGroupName"
    }
    else
    {
        TraceInfo "Creating virtual network $VNetName with Subnet $SubnetName in Location $Location"
    }
            
    $maxRetries = 10
    $retry = 0
    while($true)
    {
        $vnetMutex = New-Object System.Threading.Mutex $false,'HPCIaaSUpdateVNetMutex'
        $mutexAcquired = $false
        try
        {
            try
            {
                $mutexAcquired = $vnetMutex.WaitOne()
            }
            catch [System.Threading.AbandonedMutexException]
            {
                $mutexAcquired = $true
            }

            $emptyVNetCfg = '<NetworkConfiguration xmlns="http://schemas.microsoft.com/ServiceHosting/2011/07/NetworkConfiguration">
      <VirtualNetworkConfiguration>
        <Dns />
      </VirtualNetworkConfiguration>
    </NetworkConfiguration>'

            $uniqueID = [Guid]::NewGuid().ToString()
            $vnetCfgFile = "$env:temp\VNetCfg_$uniqueID.xml"

            InvokeAzureCmdWithRetry -Command "Get-AzureVNetConfig -ExportToFile `"$vnetCfgFile`"" | Out-Null
            if(Test-Path -Path $vnetCfgFile)
            {
                $xmlDoc = [xml](Get-Content $vnetCfgFile)
            }
            else
            {
                $xmlDoc = [xml] $emptyVNetCfg
            }
            $namespaceUri = $xmlDoc.NetworkConfiguration.NamespaceURI
            $vnets = $xmlDoc.NetworkConfiguration.VirtualNetworkConfiguration.GetElementsByTagName("VirtualNetworkSites") | Select-Object -First(1)
            if($null -eq $vnets)
            {
               $vnets = $xmlDoc.CreateElement("VirtualNetworkSites", $namespaceUri)
               $xmlDoc.NetworkConfiguration.VirtualNetworkConfiguration.AppendChild($vnets) | Out-Null
            }

            $vnetSite = $vnets.GetElementsByTagName("VirtualNetworkSite") | Where-Object {$_.name -eq $VNetName}
            if($null -ne $vnetSite)
            {
                if($vnetSite.HasAttribute("Location"))
                {
                    if(($PsCmdlet.ParameterSetName -ne "RegionalVNet") -or ($vnetSite.Location -ne $Location))
                    {
                        throw "The virtual network $VNetName already exists but not in the region $Location."
                    }
                }
                elseif($vnetSite.HasAttribute("AffinityGroup"))
                {
                    if(($PsCmdlet.ParameterSetName -ne "AffinityGroupVNet") -or ($vnetSite.AffinityGroup -ne $AffinityGroupName))
                    {
                        throw "The virtual network $VNetName already exists but not in the affinity group $AffinityGroupName."
                    }
                }

                if($null -eq ($vnetSite.GetElementsByTagName("Subnet") | Where-Object {$_.name -eq $SubnetName}))
                {
                    throw "The virtual network $VNetName already exists but not subnet $SubnetName in it."
                }
                else
                {
                    TraceInfo "The virtual network $VNetName already exists."
                    return
                }
            }

            # Create the new VNet
            $vnetSite = $xmlDoc.CreateElement("VirtualNetworkSite", $namespaceUri)
            $vnetSite.SetAttribute("name", $VNetName)
            if($PsCmdlet.ParameterSetName -eq "AffinityGroupVNet")
            {
                $vnetSite.SetAttribute("AffinityGroup", $AffinityGroupName)
            }
            else
            {
                $vnetSite.SetAttribute("Location", $Location)
            }

            $vnets.AppendChild($vnetSite) | Out-Null
            $addrspace = $xmlDoc.CreateElement("AddressSpace", $namespaceUri)
            $vnetSite.AppendChild($addrspace) | Out-Null
            $addrPrefix = $xmlDoc.CreateElement("AddressPrefix", $namespaceUri)
            $addrPrefix.InnerText = $AddressSpace
            $addrspace.AppendChild($addrPrefix) | Out-Null
            $subnets = $xmlDoc.CreateElement("Subnets", $namespaceUri)
            $vnetSite.AppendChild($subnets) | Out-Null
            $subnet1 = $xmlDoc.CreateElement("Subnet", $namespaceUri)
            $subnet1.SetAttribute("name", $SubnetName)
            $subnets.AppendChild($subnet1) | Out-Null
            $addrPrefix1 = $xmlDoc.CreateElement("AddressPrefix", $namespaceUri);
            $addrPrefix1.InnerText = $SubnetAddressSpace
            $subnet1.AppendChild($addrPrefix1) | Out-Null
            $xmlDoc.Save($vnetCfgFile)

            try
            {
                Set-AzureVNetConfig -ConfigurationPath $vnetCfgFile -ErrorAction Stop
                TraceInfo "Successfully created the virtual network $VNetName with subnet $SubnetName"
                return
            }
            catch
            {
                $errMsg = "Failed to Create virtual network $VNetName : " + ($_ | Out-String)
                if($retry -lt $maxRetries)
                {
                    TraceWarning $errMsg
                }
                else
                {
                    throw $errMsg
                }
            }
        }
        finally
        {
            if($mutexAcquired)
            {
                $vnetMutex.ReleaseMutex()
            }
        }
                    
        $interval = Get-Random -minimum 20 -maximum 40
        TraceInfo "Retry to create virtual network $VNetName after $interval seconds"
        Start-Sleep -Seconds $interval
        $retry++
    }
}

<#
.SYNOPSIS
   Modifies the virtual network configuration xml file to include a DNS service reference.
.DESCRIPTION
   This a small utility that programmatically modifies the vnet configuration file to add a DNS server
   then adds the DNS server's reference to the specified virtual network site.
.EXAMPLE
    ConfigVNetDnsServer -DnsServerName "contoso" -IpAddress "10.0.0.4" -VNetName "dcvnet"
.INPUTS
   None
.OUTPUTS
   None
#>
function ConfigVNetDnsServer
{
   param
    (
        [Parameter(Mandatory = $true)]
        [String] $DnsServerName,

        [Parameter(Mandatory = $true)]
        [String] $IpAddress,

        [Parameter(Mandatory = $true)]
        [String] $VNetName
    )
    
    TraceInfo "Configuring a DNS server for virtual network $VNetName : Name=$DnsServerName, IP=$IpAddress"
    $maxRetries = 10
    $retry = 0
    while($true)
    {
        $vnetMutex = New-Object System.Threading.Mutex $false,'HPCIaaSUpdateVNetMutex'
        $mutexAcquired = $false
        try
        {
            try
            {
                $mutexAcquired = $vnetMutex.WaitOne()
            }
            catch [System.Threading.AbandonedMutexException]
            {
                $mutexAcquired = $true
            }

            $vNet = Get-AzureVNetSiteWithRetry -VNetName $VNetName
            if ($null -eq $vNet)
            {
                throw "The virtual network $VNetName does not exist. Cannot add an DNS server for it."
            }
            $uniqueID = [Guid]::NewGuid().ToString()
            $vnetFilePath = "$env:temp\VNetCfg_$uniqueID.xml"
            InvokeAzureCmdWithRetry -Command "Get-AzureVNetConfig -ExportToFile `"$vnetFilePath`"" | Out-Null
            if (-not (Test-Path $vNetFilePath))
            {
                throw "Cannot retrieve the virtual network configuration file."
            }

            $xml = [Xml] (Get-Content $vnetFilePath)
            $nsUri = $xml.NetworkConfiguration.NamespaceURI

            $dnsElement = $xml.NetworkConfiguration.VirtualNetworkConfiguration.GetElementsByTagName("Dns") | Select-Object -First(1)
            if ($null -eq $dnsElement)
            {
                $dnsElement = $xml.CreateElement("Dns", $nsUri)
                $xml.NetworkConfiguration.VirtualNetworkConfiguration.AppendChild($dnsElement) | Out-Null
            }

            $dnsServersElement = $dnsElement.GetElementsByTagName("DnsServers") | Select-Object -First(1)
            if ($null -eq $dnsServersElement)
            {
                $dnsServersElement = $xml.CreateElement("DnsServers", $nsUri)
                $dnsElement.AppendChild($dnsServersElement) | Out-Null
            }

            $dnsSrvIPUpdated = $true
            $dnsServerElement = $dnsServersElement.GetElementsByTagName("DnsServer") | Where-Object {$_.name -eq $DnsServerName}
            if ($null -ne $dnsServerElement)
            {
                if($dnsServerElement.IpAddress -eq $IpAddress)
                {
                    $dnsSrvIPUpdated = $false
                }
                else
                {
                    $dnsServerElement.IpAddress = $IpAddress
                }
            }
            else
            {
                $dnsServerElement = $xml.CreateElement("DnsServer", $nsUri)
                $dnsServerElement.SetAttribute("name", $DnsServerName) | Out-Null
                $dnsServerElement.SetAttribute("IPAddress", $IpAddress) | Out-Null
                $dnsServersElement.AppendChild($dnsServerElement) | Out-Null
            }

            # Now set the DnsReference for the network site
            $vnetSiteElement = $xml.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites.GetElementsByTagName("VirtualNetworkSite") | Where-Object {$_.name -eq $VNetName}

            if ($null -eq $vnetSiteElement)
            {
                throw "Cannot find the virtual network $VNetName"
            }

            $dnsServersRefElement = $vnetSiteElement.GetElementsByTagName("DnsServersRef") | Select-Object -First(1)
            if ($null -eq $dnsServersRefElement)
            {
                $dnsServersRefElement = $xml.CreateElement("DnsServersRef", $nsUri)
                $vnetSiteElement.AppendChild($dnsServersRefElement) | Out-Null
            }
    
            $dnsServerRefElement = $dnsServersRefElement.GetElementsByTagName("DnsServerRef") | Where-Object {$_.name -eq $DnsServerName}
            if($null -eq $dnsServerRefElement)
            {
                $dnsServerRefElement = $xml.CreateElement("DnsServerRef", $nsUri)        
                $dnsServerRefElement.SetAttribute("name", $DnsServerName) | Out-Null
                $dnsServersRefElement.AppendChild($dnsServerRefElement) | Out-Null
            }
            else
            {
                if(-not $dnsSrvIPUpdated)
                {
                    TraceInfo "No update needed for virtual network $VNetName"
                    return
                }
            }

            $xml.Save($vnetFilePath)
            try
            {
                Set-AzureVNetConfig -ConfigurationPath $vnetFilePath -ErrorAction Stop
                TraceInfo "Successfully configured the DNS server for virtual network $VNetName"
                return
            }
            catch
            {
                $errMsg = "Failed to configure a DNS server for virtual network $VNetName :" + ($_ | Out-String)
                if($retry -lt $maxRetries)
                {
                    TraceWarning $errMsg
                }
                else
                {
                    throw $errMsg
                }
            }
        }
        finally
        {
            if($mutexAcquired)
            {
                $vnetMutex.ReleaseMutex()
            }
        }
        
        $interval = Get-Random -minimum 20 -maximum 40
        TraceInfo "Retry to configure the DNS server for virtual network $VNetName after $interval seconds"
        Start-Sleep -Seconds $interval
        $retry++
    }
}

function GetAvailableVNetIP
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName       
    )

    $vnet = Get-AzureVNetSiteWithRetry -VNetName $VNetName
    if($null -eq $vnet)
    {
        throw "The virtual network $VNetName doesn't exist"
    }
    $subnet = $vnet.Subnets | Where-Object {$_.Name -eq $SubnetName}
    if($null -eq $subnet)
    {
        throw "The subnet $SubnetName doesn't exist in virtual network $VNetName"
    }
    
    $testIP = [String]$subnet.AddressPrefix
    $slashIndex = $testIP.IndexOf('/')
    $testIP = $testIP.Substring(0, $slashIndex)
    $testRet = InvokeAzureCmdWithRetry -Command "Test-AzureStaticVNetIP -VNetName '$VNetName' -IPAddress $testIP"
    if($null -eq $testRet)
    {
        throw "Failed to get available IP address in virtual network $VNetName"
    }
    if($testRet.IsAvailable)
    {
        return $testIP
    }
    else
    {
        return $testRet.AvailableAddresses[0]
    }
}


function CreateVMForDC
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String] $VMName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName,         

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel,  

        [Parameter(Mandatory=$true)]
        [PSCredential] $AdminCredential
    )

    try
    {

        TraceInfo "Creating VM $VMName in cloud service $ServiceName for domain controller. Size:$VMSize Image:$ImageName VNet:$VNetName. Subnet:$SubnetName"

        if (VMExists -Name $VMName -ServiceName $ServiceName)
        {
            TraceWarning "The VM $VMName already exists in cloud service $ServiceName"
            return
        }

        $staticIP = GetAvailableVNetIP -VNetName $VNetName -SubnetName $SubnetName
        $dcVm = NewVMConfig  -Windows -VMName $VMName -VMSize $VMSize -HostCaching ReadOnly -ImageName $ImageName -SubnetName $SubnetName `
            -LocalAdminCred $AdminCredential | Set-AzureVMBGInfoExtension | Set-AzureStaticVNetIP -IPAddress $staticIP

        if(IsVMSizeWithRDMA $VMSize)
        {
            TraceInfo "CreateVMForDC: Set VM extension Microsoft.HpcCompute.HpcVmDrivers with version 1.* for the virtual machine $VMName."
            $dcVm = $dcVm | Set-AzureVMExtension -ExtensionName "HpcVmDrivers" -Publisher "Microsoft.HpcCompute" -Version "1.*"
        }

        New-AzureVMWithRetry -VM $dcVm -ServiceName $ServiceName -VNetName $VNetName -DeploymentLabel $DeploymentLabel

        WaitForVMReady -VMName $VMName -ServiceName $ServiceName
    }
    catch
    {
        TraceError "Failed to create virtual machine for domain controller"
        throw
    }
    
    ConfigVNetDnsServer -DnsServerName $VMName -IpAddress $staticIP -VNetName $VNetName
}

function PromoteDC
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String] $VMName, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $AdminCredential,

        [Parameter(Mandatory=$false)]
        [String] $DNSForwarder="",
        
        [Parameter(Mandatory=$false)]
        [Switch] $NoWindowsAutoUpdate
    )
    TraceInfo "Promoting VM $VMName as a domain controller, the domain FQDN is $DomainFQDN."
    $scriptBlock = { 
        param ([String]$Fqdn, [String]$Psw)
        
        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -eq 5)
        {
            WriteLog "$env:COMPUTERNAME was already a domain controller."
            return
        }
        
        WriteLog "Installing windows features AD-Domain-Services and GPMC."
        Install-WindowsFeature -Name AD-Domain-Services,GPMC -IncludeManagementTools
        
        WriteLog "Start to install domain forest $Fqdn"
        # During the domain forest installation, the PSSession may be broken occasionally, which makes the current powershell process gets killed,
        # We create a runaway process to install the domain forest by "Invoke-WmiMethod" cmdlet, so that the installation process can survive in this case.
        # Because the password can contain special characters, which may happen to be escape character (e.g. '\') or meaningful character (e.g. '"'),
        # it will result in cmd line parsing error, so we will convert the password into ascii codes, then convert back in the powershell command line.
        $pswAscii = $Psw.ToCharArray() | %{[Convert]::ToUInt32($_)} | %{$_.ToString()}
        $pshCmd = 'Import-Module ADDSDeployment; $pswAscii = ''{0}''.Split('' '');' -f ($pswAscii -join ' ')
        $pshCmd += '$psw = -join ($pswAscii | %{[Convert]::ToChar([Uint32]$_)});$secpsw = ConvertTo-SecureString -String $psw -AsPlainText -Force;'
        $pshCmd += 'Install-ADDSForest -DomainName {0} -InstallDNS -SafeModeAdministratorPassword $secpsw -Force' -f $Fqdn

        $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "Powershell.exe -Command ""$pshCmd"""
        $p = $null
        if($pobj.ReturnValue -eq 0)
        {
            $p = Get-Process -Id $pobj.ProcessId -ErrorAction SilentlyContinue
        }

        if($null -ne $p)
        {
            WriteLog "Installing domain forest $Fqdn"
            $p.WaitForExit()
        }
        else
        {
            throw "Failed to promote $env:COMPUTERNAME to a domain controller"
        }
    }

    # Error code 995 is expected here: The I/O operation has been aborted because of either a thread exit or an application request. 
    # Error code 5 is sometimes thrown when the remote call execution succeeded: Processing data for a remote command failed with the following error message: Access is denied. 
    # No particular action is taken for Error code 5, just retry as we do for other error codes, the script block already handles re-entry.
    RemotePshCall -VMName $VMName `
                  -ServiceName $ServiceName `
                  -Credential $AdminCredential `
                  -ScriptBlock  $scriptBlock `
                  -ArgumentList @($DomainFQDN, $AdminCredential.GetNetworkCredential().Password) `
                  -IgnoreErrorCodes 995 | Out-Null

    # Sleep 60 seconds to wait DC restart
    Start-Sleep -Seconds 60
    WaitForVMReady -VMName $VMName -ServiceName $ServiceName

    RemotePshCall -VMName $VMName -ServiceName $ServiceName -Credential $AdminCredential -RetryTimes 10 -ScriptBlock {
        param ([String] $dnsForwarder, $noWindowsAutoUpdate, $domainFQDN)

        if($null -eq (Get-DnsServerForwarder).IPAddress)
        {
            # Cannot use @((Get-DnsServerForwarder).IPAddress) in this case, because it will get an array with a $null element
            $orgFwdIPs = @()
        }
        else
        {
            $orgFwdIPs = @((Get-DnsServerForwarder).IPAddress)
        }

        $newFwdIPs = @()
        foreach($fwdIP in $orgFwdIPs)
        {
            if(($fwdIP -eq "fec0:0:0:ffff::1") -or ($fwdIP -eq "fec0:0:0:ffff::2") -or ($fwdIP -eq "fec0:0:0:ffff::3"))
            {
                WriteLog "Removing DNS forwarder from the domain controller: $fwdIP"
                Remove-DnsServerForwarder -IPAddress $fwdIP -Force
            }
            else
            {
                $newFwdIPs += $fwdIP
            }
        }

        if(-not [string]::IsNullOrEmpty($dnsForwarder))
        {
            $fwdIP = [IPAddress]$dnsForwarder
            if($newFwdIPs -notcontains $fwdIP)
            {
                WriteLog "Adding DNS forwarder to the domain controller: $fwdIP"
                $newFwdIPs += $fwdIP
                Set-DnsServerForwarder -IPAddress $newFwdIPs -Confirm:$false
            }
        }

        if($noWindowsAutoUpdate)
        {
            $dn = @($domainFQDN.split(".") | %{"DC=$_"}) -join ","
            WriteLog "Adding a group policy to disable windows auto update on the target $dn"
            Import-Module GroupPolicy
            New-GPO -Name "WindowsAUPolicy" | New-GPLink -Target $dn -LinkEnabled Yes
            Set-GPRegistryValue -Name "WindowsAUPolicy" -Key "HKEY_LOCAL_MACHINE\Software\Policies\Microsoft\Windows\WindowsUpdate\AU" -ValueName AUOptions -Value 2 -Type DWord
        }
    } -ArgumentList @($DNSForwarder, $NoWindowsAutoUpdate.IsPresent, $DomainFQDN) | Out-Null
}

function VMJoinToDomain
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $VMName,
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred,

        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN
    )
 
    $scriptblock = {
        param($cred, $domainFQDN)
        
        $maxRetries = 10
        $retry = 0
        while($true)
        {
            # -Force is added from PowerShell 3.0 on
            if($PSVersionTable["PSVersion"] -lt "3.0")
            {
                Add-Computer -DomainName $domainFQDN -Credential $cred -ErrorAction SilentlyContinue
            }
            else
            {
                Add-Computer -DomainName $domainFQDN -Credential $cred -Force -ErrorAction SilentlyContinue
            }
            
            $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
            if($domainRole -ge 3)
            {
                WriteLog "$env:COMPUTERNAME successfully joined domain $DomainFQDN and will restart."
                break
            }
            
            if($retry -lt $maxRetries)
            {
                WriteLog ("$env:COMPUTERNAME failed to join domain $DomainFQDN : " + $Error[0])
                Start-Sleep -Seconds 10
                $retry++
            }
            else
            {
                throw "$env:COMPUTERNAME failed to join domain $DomainFQDN after $maxRetries retries."
            }
        }

        try
        {
            Write-VolumeCache C -ErrorAction SilentlyContinue
        }
        catch
        {
            # Ignore the error
        }
    }
    
    $tryRestart = $false
    while($true)
    {
        try
        {
            RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $LocalAdminCred -ScriptBlock $scriptblock -ArgumentList @($DomainUserCred, $DomainFQDN) | Out-Null
            InvokeAzureCmdWithRetry -Command "Restart-AzureVM -ServiceName $ServiceName -Name $VMName"
            break
        }
        catch [System.Management.Automation.RemoteException]
        {
            if(($_.Exception.Message.Contains("Failed to join domain $DomainFQDN")) -and (-not $tryRestart))
            {
                $tryRestart = $true
                TraceWarning "$VMName failed to join domain $DomainFQDN, will restart and try again."
                InvokeAzureCmdWithRetry -Command "Restart-AzureVM -ServiceName $ServiceName -Name $VMName"
                Start-Sleep -Seconds 20
                WaitForVMReady -VMName $VMName -ServiceName $ServiceName                
            }
            else
            {
                throw
            }
        }
    }
    
    Start-Sleep -Seconds 20
    WaitForVMReady -VMName $VMName -ServiceName $ServiceName
}

function AddADUserToLocalAdmin
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred,

        # Full user name, for example "consto\john"
        [Parameter(Mandatory=$true)]
        [String] $ADUserName
    )
    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $LocalAdminCred -ScriptBlock {
        param([String]$aduser)

        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -lt 3)
        {
            throw "$env:COMPUTERNAME is not domain joined"
        }
        
        $maxRetries = 3
        $retry = 0
        while($true)
        {
            try
            {
                WriteLog "Adding user $aduser to local administrators group on $env:COMPUTERNAME"
                $adminGroup = [ADSI]("WinNT://$env:COMPUTERNAME/administrators, group")
                $usr = $aduser.Replace("\","/")
                $adminGroup.Add("WinNT://$usr, user")
                WriteLog "The user $aduser successfully added to local administrators group on $env:COMPUTERNAME"
                break
            }
            catch [System.Management.Automation.MethodInvocationException]
            {
                if($retry -lt $maxRetries)
                {
                    WriteLog ("Failed to add user $aduser to local administrators group, retrying after 5 seconds: "+ $_)
                    Start-Sleep -Seconds 5
                    $retry++
                }
                else
                {
                    throw ("Failed to add user $aduser to local administrators group after $maxRetries retries: " + $_)
                }
            }
        }

    } -ArgumentList $ADUserName | Out-Null
}

function ConfigRemoteDBServer
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $DBAdminCred,

        # Full user name, for example "consto\john"
        [Parameter(Mandatory=$true)]
        [String] $HpcSetupUserName,

        # Full headnode machine account name, for example "consto\myheadnode$"
        [Parameter(Mandatory=$true)]
        [String] $HNMachineAccount
    )

    TraceInfo "Configuring remote DB server $VMName : SetupUser=$HpcSetupUserName HeadNode=$HNMachineAccount"
    try
    {
        $dbcreateSql = [IO.File]::ReadAllBytes("$env:HPCPSScriptRoot\HPCIaaSCreateDatabase.sql")
        $dbpermissionSql = [IO.File]::ReadAllBytes("$env:HPCPSScriptRoot\AddDbUserForHpcService.sql")
    }
    catch
    {
        throw ("Failed to load SQL scripts to configure remote DB server:" + $_)
    }

    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $DBAdminCred -ScriptBlock {
        param([string]$aduser, [String]$hnMachine, [Byte[]]$dbCreateScript, [Byte[]]$dbPermissionScript)

        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -lt 3)
        {
            throw "The DB Server $env:COMPUTERNAME is not domain joined"
        }

        # Try to update IP address to DNS server
        Register-DnsClient -ErrorAction SilentlyContinue

        # Get the data disk ready
        $maxRetries = 10
        $retry = 0
        while($true)
        {
            try
            {
                $disk = Get-Disk -Number 2 -ErrorAction Stop
                if($disk.PartitionStyle -eq "RAW")
                {
                    WriteLog "Initializing the data disk."
                    Initialize-Disk -Number 2 -PartitionStyle MBR -ErrorAction Stop | Out-Null
                }

                $partition = Get-Partition -DiskNumber 2 -ErrorAction SilentlyContinue
                if($null -eq $partition)
                {
                    WriteLog "Creating new partition with drive letter F."
                    New-Partition -DiskNumber 2 -UseMaximumSize -IsActive -DriveLetter F -ErrorAction Stop | Out-Null
                }

                WriteLog "Formating volume F to NTFS file system."
                Format-Volume -DriveLetter F -FileSystem NTFS -NewFileSystemLabel "HPC Data" -Force:$true -confirm:$false -ErrorAction Stop | Out-Null

                WriteLog "Creating HPCData directory."
                New-Item -ItemType directory -Path F:\ -Name HPCData -Force -ErrorAction Stop | Out-Null
                break
            }
            catch
            {
                if($retry -lt $maxRetries)
                {
                    WriteLog "Failed to get data disk ready, will retry after 10 seconds: $_"
                    Start-Sleep -Seconds 10
                    $retry++
                }
                else
                {
                    throw
                }
            }
        }

        #Configure firewall to allow incoming SQL connections
        WriteLog "Configuring firewall for SQL Server"
        New-NetFirewallRule -DisplayName "Allow SQL Server TCP In" -Direction Inbound -Action Allow -LocalPort 1433 -Protocol TCP -Profile Domain | Out-Null

        # Add domain user as 'sysadmin' of SQL server
        Set-ExecutionPolicy -ExecutionPolicy RemoteSigned -Force
        WriteLog "Connecting to SQL server instance"
        $sqlConnection = New-Object -TypeName System.Data.SqlClient.SqlConnection
        $sqlConnection.ConnectionString = "Data Source=localhost;Initial Catalog=master;Integrated Security=True;Connection Timeout=90"

        $maxRetries = 50
        $retry = 0
        while($true)
        {
            try 
            {
               $sqlConnection.Open()
               WriteLog "Connected to SQL server instance"
               break
            }
            catch
            {
                if($retry -lt $maxRetries)
                {
                    WriteLog "Failed to connect SQL server instance, retry after 20 seconds: $_"
                    Start-Sleep -Seconds 20
                    $retry++
                }
                else
                {
                    throw "Failed to connect SQL Server instance after $maxRetries retries: $_"
                }
            }
            finally
            {
                $sqlConnection.Close()
            }
        }


        WriteLog "Adding user $aduser as a sysadmin of SQL server."
        $maxRetries = 3
        $retry = 0
        while($true)
        {
            $Command = $null
            try
            {
                $sqlConnection.Open()
                $Command = $sqlConnection.CreateCommand()
                $Command.CommandTimeout = 120
                $Command.CommandText = "if not exists (select * from sys.server_principals where name = '$aduser') create login [$aduser] from windows"
                $Command.ExecuteNonQuery() | Out-Null
                $Command.CommandText = "if is_srvrolemember('sysadmin', '$aduser')=0 exec sp_addsrvrolemember '$aduser', 'sysadmin'"
                $Command.ExecuteNonQuery() | Out-Null

                WriteLog "Successfully added user $aduser as a sysadmin of SQL server"
                break
            }
            catch
            {
                WriteLog "Failed to add user $aduser as a sysadmin of SQL server, retrying: $_"
                if($retry -lt $maxRetries)
                {
                    $retry++
                    Start-Sleep -Seconds 2
                }
                else
                {
                    throw
                }
            }
            finally
            {
                if($null -ne $Command)
                {
                    $Command.Dispose()
                    $Command = $null
                }

                $sqlConnection.Close()
            }
        }

        [IO.File]::WriteAllBytes("$env:temp\HPCIaaSCreateDatabase.sql", $dbCreateScript)
        [IO.File]::WriteAllBytes("$env:temp\AddDbUserForHpcService.sql", $dbPermissionScript)
        try
        {
            WriteLog "Creating HPC databases"
            Import-Module "sqlps" -DisableNameChecking -Force
            Invoke-Sqlcmd -ServerInstance localhost -InputFile "$env:temp\HPCIaaSCreateDatabase.sql" -QueryTimeout 0 -Variable "HPCDataPath='F:\HPCData'"
            WriteLog "HPC Databases created"
                        
            WriteLog "Configuring the permissions for HPC databases"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCManagement" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCScheduler" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCMonitoring" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCReporting" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCDiagnostics" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            WriteLog "The permissions for HPC databases configured"
        }
        catch
        {
            WriteLog ("Failed to Create HPC databases: " + $_)
            throw
        }

    } -ArgumentList @($HpcSetupUserName,$HNMachineAccount,$dbcreateSql,$dbpermissionSql) | Out-Null
}

function PrepareHeadNode
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $SetupUserCred,

        [Parameter(Mandatory=$true, ParameterSetName="localdb")]
        [Switch] $LocalDB, 

        [Parameter(Mandatory=$true, ParameterSetName="remotedb")]
        [Switch] $RemoteDB, 

        [Parameter(Mandatory=$true, ParameterSetName="remotedb")]
        [String] $DBServerInstance,

        [Parameter(Mandatory=$true)]
        [String] $SubscriptionId,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName,
        
        [Parameter(Mandatory=$true)]
        [String] $Location,  

        [Parameter(Mandatory=$false)]
        [String] $AffinityGroupName = "",

        [Parameter(Mandatory=$false)]
        [String] $PreConfigScirptFile = "",
        
        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $ManagementDB = "HpcManagement",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $SchedulerDB = "HpcScheduler",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $MonitoringDB = "HpcMonitoring",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $ReportingDB = "HpcReporting",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $DiagnosticsDB = "HpcDiagnostics"
    )


    $deployment = InvokeAzureCmdWithRetry -Command "Get-AzureDeployment -ServiceName $ServiceName -Slot Production"
    if($null -eq $deployment)
    {
        throw "No deployment found in cloud service $ServiceName"
    }

    if($LocalDB.IsPresent)
    {
        $dbArgs = "-DBServerInstance .\COMPUTECLUSTER"
    }
    else
    {
        $dbArgs = "-RemoteDB -DBServerInstance $DBServerInstance -ManagementDB $ManagementDB -SchedulerDB $SchedulerDB -MonitoringDB $MonitoringDB -ReportingDB $ReportingDB -DiagnosticsDB $DiagnosticsDB"
    }

    if((-not [String]::IsNullOrEmpty($PreConfigScirptFile)) -and (Test-Path -Path $PreConfigScirptFile -PathType Leaf))
    {
        $preConfig = [IO.File]::ReadAllBytes($PreConfigScirptFile)
    }
    else
    {
        $preConfig = $null
    }

    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $SetupUserCred -ScriptBlock {
        Param ([String]$depid, [String]$subsid, [String]$vnet, [String]$subnet, [String]$agname, [String]$location, [String]$user, [String]$psw, [String]$dbArgs, [Byte[]]$PreConfigContent)

        # Try to clear local dns cache and update local IP to dns server
        Clear-DnsClientCache -ErrorAction SilentlyContinue
        Register-DnsClient -ErrorAction SilentlyContinue

        # Try to execute HPCHNPreConfig.ps1, but will continue whatever the execution result is
        if(($null -ne $PreConfigContent) -and ($PreConfigContent.Count -gt 0))
        {
            $preConfigScriptFile = "$env:temp\HPCHNPreConfig.ps1"
            try
            {
                WriteLog "Creating $preConfigScriptFile"
                [IO.File]::WriteAllBytes($preConfigScriptFile, $PreConfigContent)
                WriteLog "$preConfigScriptFile created"
                WriteLog "Executing $preConfigScriptFile"
                Start-Process -FilePath "PowerShell.exe" -ArgumentList "-ExecutionPolicy Unrestricted -File `"$preConfigScriptFile`"" -NoNewWindow -Wait
                WriteLog "$preConfigScriptFile executed"
            }
            catch
            {
                WriteLog "Failed to run $preConfigScriptFile : $_"
            }
        }

        New-Item -Path HKLM:\SOFTWARE\Microsoft\HPC -Name IaaSInfo -Force | Out-Null
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name SubscriptionId -Value $subsid
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name DeploymentId -Value $depid
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name VNet -Value $vnet
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name Subnet -Value $subnet
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name AffinityGroup -Value $agname
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name Location -Value $location
        WriteLog "The information needed for in-box management scripts succcessfully configured."

        Import-Module ScheduledTasks
        $HNPreparePsFile = [System.IO.Path]::Combine($env:CCP_HOME, "Bin\HPCHNPrepare.ps1")
        $action = New-ScheduledTaskAction -Execute "PowerShell.exe" -Argument "-ExecutionPolicy Unrestricted -Command `"& '$HNPreparePsFile' $dbArgs`""
        Register-ScheduledTask -TaskName "HPCPrepare" -Action $action -User $user -Password $psw -RunLevel Highest | Out-Null
        if(-not $?)
        {
            throw "Failed to schedule HPC head node preparation task"
        }
        WriteLog "HPC head node preparation task scheduled"

        #Manually trigger the task
        Start-ScheduledTask -TaskName "HPCPrepare"
        if(-not $?)
        {
            throw "Failed to start HPC head node preparation task"
        }
        WriteLog "Running HPC head node preparation task"

        Start-Sleep -Milliseconds 500
        $taskSucceeded = $false
        do
        {
            $taskState = (Get-ScheduledTask -TaskName "HPCPrepare").State
            if($taskState -eq "Ready")
            {
                $taskInfo = Get-ScheduledTaskInfo -TaskName "HPCPrepare"
                if($null -eq $taskInfo.LastRunTime)
                {
                    Start-ScheduledTask -TaskName "HPCPrepare"
                }
                else
                {
                    if($taskInfo.LastTaskResult -eq 0)
                    {
                        $taskSucceeded = $true
                        break
                    }
                    else
                    {
                        WriteLog ("The HPC head node preparation task failed:" + $taskInfo.LastTaskResult)
                        break
                    }
                }
            }
            elseif($taskState -ne "Queued" -and $taskState -ne "Running")
            {
                WriteLog "The HPC head node preparation task entered into unexpected state: $taskState"
                break
            }

            Start-Sleep -Seconds 2
        } while ($true)
        
        if($taskSucceeded)
        {
            WriteLog "Checking the HPC services status in HPC head node ..."
            $HNServiceList = @("HpcSdm", "HpcManagement", "HpcReporting", "HpcMonitoringClient", "HpcNodeManager", "msmpi", "HpcBroker", `
                "HpcDiagnostics", "HpcScheduler", "HpcMonitoringServer", "HpcSession", "HpcSoaDiagMon")
            foreach($svcname in $HNServiceList)
            {
                $service = Get-Service -Name $svcname -ErrorAction SilentlyContinue
                if($null -eq $service)
                {
                    WriteLog "Service $svcname not found"
                    $taskSucceeded = $false
                }
                elseif($service.Status -eq "Running")
                {
                    WriteLog "Service $svcname is running"
                }
                else
                {
                    WriteLog "Service $svcname is in $($service.Status) status"
                    $taskSucceeded = $false
                }
            }
        }

        Unregister-ScheduledTask -TaskName "HPCPrepare" -Confirm:$false

        if($taskSucceeded)
        {
            WriteLog "Succeeded to prepare HPC head node"
        }
        else
        {
            WriteLog "Failed to prepare HPC head node"
            if(Test-Path -Path "$env:windir\Temp\HPCHeadNodePrepare.log" -PathType Leaf)
            {
                WriteLog "The HPC head node preparation logs as below:"
                Get-Content -Path "$env:windir\Temp\HPCHeadNodePrepare.log" | Write-Verbose -Verbose
            }

            throw "Failed to prepare HPC head node"
        }
    } -ArgumentList @($deployment.DeploymentId, $SubscriptionId, $VNetName, $SubnetName, $AffinityGroupName, $Location, `
        $SetupUserCred.UserName, $SetupUserCred.GetNetworkCredential().Password, $dbArgs, $preConfig) | Out-Null
}


function ConfigHeadNode
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [PSCredential] $SetupUserCred,

        [Parameter(Mandatory=$true)]
        [String] $NodeNamePattern,

        [Parameter(Mandatory=$false)]
        [Switch] $EnableRESTAPI,

        [Parameter(Mandatory=$false)]
        [Switch] $EnableWebPortal,

        [Parameter(Mandatory=$false)]
        [String] $PostConfigScirptFile = "",

        [Parameter(Mandatory=$false)]
        [String] $MPINetMask = "",

        [Parameter(Mandatory=$false)]
        [Object[]] $Certificates = @(),

        [Parameter(Mandatory=$false)]
        [Object] $AutoGrowShrinkConfig = $null
    )

    TraceInfo "Configuring the HeadNode $VMName"
    $connUri = InvokeAzureCmdWithRetry -Command "Get-AzureWinRMUri -ServiceName $ServiceName -Name $VMName" -RetryTimes 5
    if($null -eq $connUri)
    {
        throw "Failed to retrieve the WinRM https listener URI of the virtual machine $VMName in cloud service $ServiceName."
    }

    $postConfig = $null
    if((-not [String]::IsNullOrEmpty($PostConfigScirptFile)) -and (Test-Path -Path $PostConfigScirptFile -PathType Leaf))
    {
        $postConfig = [IO.File]::ReadAllBytes($PostConfigScirptFile)
    }

    $ret = RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $SetupUserCred -ScriptBlock {
        param($creds, $nodenaming, $dnsHostName, $restAPI, $webPortal, $mpiNetMask, $certificates, $autoGrowShrinkConfig, [Byte[]]$postConfigContent)

        Add-PSSnapin Microsoft.HPC
        # setting network topology to 5 (enterprise)
        WriteLog "Setting network topologogy..."
        $nics = @(Get-WmiObject win32_networkadapterconfiguration -filter "IPEnabled='true' AND DHCPEnabled='true'")
        if ($nics.Count -ne 1)
        {
            throw "Cannot find a suitable network adapter for enterprise topology"
        }

        try
        {
            $startTime = Get-Date
            $retry = 0
            while($true)
            {
                try
                {
                    Set-HpcNetwork -Topology "Enterprise" -Enterprise $nics[0].Description -EnterpriseFirewall $true -ErrorAction Stop | Out-Null
                    break
                }
                catch
                {
                    if($retry -lt 5)
                    {
                        WriteLog "Failed to set Hpc network topology, maybe the head node is still on initialization"
                        Start-Sleep -Seconds 10
                        $retry++
                    }
                    else
                    {
                        throw
                    }
                }
            }
            $topo = Get-HpcNetworkTopology -ErrorAction SilentlyContinue
            WriteLog "Network topology is set to $topo"

            # Set installation credentials
            Set-HpcClusterProperty -InstallCredential $creds | Out-Null
            $hpccred = Get-HpcClusterProperty -InstallCredential
            WriteLog ("Installation Credentials set to " + $hpccred.Value)

		    # set node naming series
            Set-HpcClusterProperty -NodeNamingSeries $nodenaming | Out-Null
            WriteLog "Node naming series set to $nodenaming."
        
            # Create a default compute node template
            New-HpcNodeTemplate -Name "Default ComputeNode Template" -Description "This is the default compute node template" | Out-Null
            WriteLog "'Default ComputeNode Template' created."

            $cerContent = $null
            if($restAPI -or $webPortal)
            {
                $hpcBinPath = [System.IO.Path]::Combine($env:CCP_HOME, "Bin")
                $restWebCert = Get-ChildItem -Path Cert:\LocalMachine\My | ?{($_.Subject -eq "CN=$dnsHostName") -and $_.HasPrivateKey} | select -First(1)
                if($null -eq $restWebCert)
                {
                    WriteLog "Generating a self-signed certificate for the HPC web service ..."
                    $thumbprint = . $hpcBinPath\New-HpcCert.ps1 -MachineName $dnsHostName -SelfSigned
                    WriteLog "A self-signed certificate $thumbprint was created and installed"
                    $restWebCert = Get-ChildItem -Path Cert:\LocalMachine\My\$thumbprint
                }
                else
                {
                    WriteLog "Use the existing certificate $thumbprint (CN=$dnsHostName) for the HPC web service."
                    $thumbprint = $restWebCert.Thumbprint
                }

                if($webPortal)
                {
                    WriteLog "Enabling HPC Pack web portal ..."
                    . $hpcBinPath\Set-HPCWebComponents.ps1 -Service Portal -enable -Certificate $thumbprint | Out-Null
                    WriteLog "HPC Pack web portal enabled."
                }

                if($restAPI)
                {
                    WriteLog "Starting HPC web service ..."
                    Set-Service -Name "HpcWebService" -StartupType Automatic | Out-Null
                    Start-Service -Name "HpcWebService" | Out-Null
                    WriteLog "HPC web service started."

                    WriteLog "Enabling HPC Pack REST API ..."
                    . $hpcBinPath\Set-HPCWebComponents.ps1 -Service REST -enable -Certificate $thumbprint | Out-Null
                    WriteLog "HPC Pack REST API enabled."

                    WriteLog "Restarting HPCScheduler service ..."
                    Restart-Service -Name "HpcScheduler" -Force | Out-Null
                    WriteLog "HPCScheduler service restarted."
                }

                $cerFile = "$env:TEMP\hpcwebcomponent_{0}.cer" -f (Get-Random)
                Export-Certificate -Cert $restWebCert -FilePath $cerFile | Out-Null
                $cerContent = [IO.File]::ReadAllBytes($cerFile)
                WriteLog "The certificate file with public key was exported: $thumbprint"
                Remove-Item $cerFile -Force -ErrorAction SilentlyContinue  | Out-Null
            }

            # Run HPCHNPreConfig.ps1 in a separate process if it exists, will not wait for its completion.
            if(($null -ne $postConfigContent) -and ($postConfigContent.Count -gt 0))
            {
                $postConfigScriptFile = "$env:temp\HPCHNPostConfig.ps1"
                try
                {
                    WriteLog "Creating $postConfigScriptFile"
                    [IO.File]::WriteAllBytes($postConfigScriptFile, $postConfigContent)
                    WriteLog "$postConfigScriptFile created"
                    $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "PowerShell.exe -ExecutionPolicy Unrestricted -File `"$postConfigScriptFile`""
                    if($pobj.ReturnValue -eq 0)
                    {
                        WriteLog "Start to run $postConfigScriptFile."
                    }
                    else
                    {
                        WriteLog "Failed to run $postConfigScriptFile."
                    }
                }
                catch
                {
                    WriteLog "Failed to run $postConfigScriptFile : $_"
                }
            }

            if(-not [string]::IsNullOrEmpty($mpiNetMask))
            {
                ## Wait for the completion of the "Updating cluster configuration" operation after setting network topology,
                ## because in the operation, the CCP_MPI_NETMASK may be set.
                $waitLoop = 0
                while ($null -eq (Get-HpcOperation -StartTime $startTime -State Committed | ?{$_.Name -eq "Updating cluster configuration"}))
                {
                    if($waitLoop++ -ge 10)
                    {
                        break
                    }

                    Start-Sleep -Seconds 10
                }

                Set-HpcClusterProperty -Environment "CCP_MPI_NETMASK=$mpiNetMask"  | Out-Null
                WriteLog "Set cluster environment CCP_MPI_NETMASK to $mpiNetMask"
            }

            if($null -ne $certificates -and $certificates.Count -gt 0)
            {
                $rootDir = "$env:TEMP\Certs" + (Get-Random)
                New-Item -ItemType directory -Path $rootDir -Force | Out-Null

                foreach($certItem in $certificates)
                {
                    $CertName = $certItem.CertName
                    $pfxFileName = [IO.Path]::Combine($rootDir, "$CertName.pfx")
                    WriteLog "Importing the certificate from '$CertName.pfx'"
                    [IO.File]::WriteAllBytes($pfxFileName, $certItem.PfxContent)
                    try
                    {
                        if(-not [string]::IsNullOrEmpty($certItem.Password))
                        {
                            $secPassword = ConvertTo-SecureString -String $certItem.Password -AsPlainText -Force
                            $importedCertificate = Import-PfxCertificate -FilePath $pfxFileName -CertStoreLocation Cert:\LocalMachine\My -Password $secPassword
                        }
                        else
                        {
                            $importedCertificate = Import-PfxCertificate -FilePath $pfxFileName -CertStoreLocation Cert:\LocalMachine\My
                        }
                
                        if($importedCertificate.Subject -eq $importedCertificate.Issuer)
                        {
                            # Self-signed certificate, export a cer file and import to cert:\localMachine\root
                            WriteLog "Exporting $CertName.cer from $CertName.pfx"
                            $cerFileName = [IO.Path]::Combine($rootDir, "$CertName.cer")
                            $certFullPath = "Cert:\LocalMachine\My\" + $importedCertificate.Thumbprint
                            Export-Certificate -Cert $certFullPath -FilePath $cerFileName -Force  | Out-Null

                            WriteLog "Importing the certificate from $CertName.cer"
                            Import-Certificate -FilePath $cerFileName -CertStoreLocation Cert:\LocalMachine\Root  | Out-Null
                        }
                    }
                    catch
                    {
                        throw ("Failed to import certificate $CertName.pfx : " + ($_ | Out-String)) 
                    }
                }
            }

            if($null -ne $autoGrowShrinkConfig)
            {
                try
                {
                    WriteLog "Configuring auto grow and shrink."
                    foreach($certItem in $certificates)
                    {
                        if($certItem.Id -eq $autoGrowShrinkConfig.CertificateId)
                        {
                            Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name ThumbPrint -Value $certItem.ThumbPrint | Out-Null
                            break
                        }
                    }

                    if($autoGrowShrinkConfig.GrowInterval -ne 0)
                    {
                        WriteLog "Set GrowInterval to $($autoGrowShrinkConfig.GrowInterval)."
                        Set-HpcClusterProperty -GrowInterval $autoGrowShrinkConfig.GrowInterval  | Out-Null
                    }
                    if($autoGrowShrinkConfig.ShrinkInterval -ne 0)
                    {
                        WriteLog "Set ShrinkInterval to $($autoGrowShrinkConfig.ShrinkInterval)."
                        Set-HpcClusterProperty -ShrinkInterval $autoGrowShrinkConfig.ShrinkInterval  | Out-Null
                    }
                }
                catch
                {
                    WriteLog "Failed to configure auto grow and shrink: $_"
                }
            }

            return @{"Certificate"=$cerContent}
        }
        catch
        {
            throw ("Failed to configure head node $env:COMPUTERNAME : " + ($_ | Out-String))
        }
    } -ArgumentList @($setupUserCred, $NodeNamePattern, $connUri.Host, $EnableRESTAPI.IsPresent, $EnableWebPortal.IsPresent, $MPINetMask, $Certificates, $AutoGrowShrinkConfig, $postConfig)

    $cerFileName = ""
    if($EnableRESTAPI.IsPresent -or $EnableWebPortal.IsPresent)
    {
        try
        {
            if($null -ne $ret["Certificate"])
            {
                $cerFileName = "HPCWebComponent_{0}_{1}.cer" -f $VMName, (Get-Date).ToString("yyyyMMddHHmmss")
                try
                {
                    $certdir = [Environment]::GetFolderPath("MyDocuments")
                }
                catch
                {
                }

                if(-not (Test-Path -Path $certdir -PathType Container))
                {
                    $certdir = $env:TEMP
                }

                $cerFileName = [IO.Path]::Combine($certdir, $cerFileName)
                [IO.File]::WriteAllBytes($cerFileName, $ret["Certificate"])
                    
                # Validate the certificate file
                GetCertificate $cerFileName | Out-Null
                TraceInfo "The certificate file for the HPC Pack web service was saved at: $cerFileName"
            }

            TraceInfo "Adding endpoint 'HTTPS'(Protocol=tcp, PublicPort=443, LocalPort=443) for virtual machine $VMName ..."
            $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -Name $VMName -ServiceName $ServiceName" | Add-AzureEndpoint -Name 'HTTPS' -Protocol 'tcp' -PublicPort 443 -LocalPort 443
            Update-AzureVMWithRetry -Name $VMName -ServiceName $ServiceName -VM $vm.VM | Out-Null
            TraceInfo "Endpoint 'HTTPS' was added for virtual machine $VMName on cloud service $ServiceName" 
        }
        catch
        {
            throw ("Failed to configure head node $VMName : " + ($_ | Out-String))
        }
    }

    TraceInfo "Head node $VMName configured."
    return $cerFileName
}

function RemoveCorpAzureDisk
{
    param
    (
        #The name of Azure disk
        [Parameter(Mandatory=$true)]
        [String] $DiskName,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $retry = 0
    while($true)
    {
        try
        {
            Remove-AzureDisk -DiskName $DiskName -DeleteVHD -ErrorAction Stop
            break
        }
        catch
        {
            $caughtException = $_.Exception
            if(IsCloudException $caughtException)
            {
                $errCode = ErrorCodeOfCloudException $caughtException
                if($errCode -eq "ResourceNotFound")
                {
                    TraceDetail "Azure Disk $DiskName removed."
                    break
                }
            }

            TraceWarning "Failed to remove Azure Disk $DiskName : $_"
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove Azure Disk $DiskName after 20 seconds"
            Start-Sleep -Seconds 20
            $retry++
        }
        else
        {
            throw "Failed to remove the Azure disk $DiskName, you can remove it manually."            
        }
    }
}

function CreateHPCNodeBaseImage
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $Name, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize, 
        
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 
        
        [Parameter(Mandatory=$true)]
        [String] $ClusterName, 
        
        [Parameter(Mandatory=$true)]
        [String] $HeadNodeServiceName, 
                
        [Parameter(Mandatory=$true)]
        [ValidateSet("ComputeNode","BrokerNode")]
        [String] $NodeType,
        
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $SourceImageName
    )

    # The source image is not a valid HPC CN/BN image for this cluster, create a base image
    # Use a non-HPC deployment label for base node
    TraceInfo "Start to create HPC $NodeType image"
    CreateDomainJoinedVM -VMName $Name `
                         -ServiceName $ServiceName `
                         -VMSize $VMSize `
                         -ImageName $SourceImageName `
                         -VNetName $VNetName `
                         -SubnetName $SubnetName `
                         -LocalAdminCred $LocalAdminCred `
                         -DomainUserCred $DomainUserCred `
                         -DomainFQDN $DomainFQDN `
                         -DeploymentLabel "HPCBaseImage-$NodeType" `
                         -HostCaching ReadOnly `
                         -DisableGuestAgent

    # Wait for the base node to be ready
    WaitForVMReady -VMName $Name -ServiceName $ServiceName

    # Wait for the Head Node VM to be ready
    WaitForVMReady -VMName $ClusterName -ServiceName $HeadNodeServiceName

    #Install Compute/Broker Node
    $rmPshBlock = {
        param($clustername, $nodetype)

        if(Test-Path -Path "C:\HPCPatches")
        {
            Remove-Item -Path "C:\HPCPatches" -Recurse -Force -ErrorAction SilentlyContinue | Out-Null
        }

        New-Item -ItemType directory -Path "C:\HPCPatches" -Force | Out-Null
    
        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -lt 3)
        {
            throw "$nodetype $env:COMPUTERNAME is not domain joined"
        }
    
        # Test the connection to head node
        WriteLog "Testing the connection to $ClusterName ..."
        $maxRetries = 50
        $retry = 0
        while($true)
        {
            # Flush the DNS cache in case the cached head node ip is wrong.
            # Do not use Clear-DnsClientCache because it is not supported in Windows Server 2008 R2
            Start-Process -FilePath ipconfig -ArgumentList "/flushdns" -Wait -NoNewWindow | Out-Null
            if(Test-Connection -ComputerName $ClusterName -Quiet)
            {
                WriteLog "Head node $ClusterName is now reachable."
                break
            }
            else
            {
                if($retry -lt 20)
                {
                    Start-Sleep -Seconds 20
                    $retry++
                }
                else
                {
                    throw "Head node $ClusterName is unreachable"
                }
            }
        }

        # Because ScheduledTasks PowerShell module not available in Windows Server 2008 R2,
        # We use ComObject Schedule.Service to schedule task
        try
        {
            $schdService = new-object -ComObject "Schedule.Service"
            $schdService.Connect("localhost") | Out-Null
            $rootFolder = $schdService.GetFolder("\")
            $taskDefinition = $schdService.NewTask(0)
            $testPathAction = $taskDefinition.Actions.Create(0)
            $testPathAction.Path = "PowerShell.exe"
            $testPathPshCmd = '$retry=0; while($true){if(Test-Path \\{ClusterName}\REMINST\setup.exe){Copy-Item \\{ClusterName}\REMINST\Patches\KB*.exe C:\HPCPatches -ErrorAction SilentlyContinue; return} elseif($retry -lt 30){$retry++; start-sleep -seconds 20} else{throw ''not available''}}'
            $testPathPshCmd = $testPathPshCmd.Replace("{ClusterName}", $ClusterName)
            $testPathAction.Arguments = "-Command `"$testPathPshCmd`""
            $Action = $taskDefinition.Actions.Create(0)
            $Action.Path = "\\$clustername\REMINST\setup.exe"
            $Action.Arguments = "-unattend -{0}:{1}" -f $nodetype.ToLower(), $clustername
            $hpcSetupTask = $Rootfolder.RegisterTaskDefinition("hpcsetup", $taskDefinition, 2, "system", $null, 5)
            WriteLog  "HPC $nodetype installation task scheduled"
        }
        catch
        {
            throw "Failed to schedule HPC $nodetype installation task" 
        }
    

        try
        {
            $hpcSetupTask.Run($null) | Out-Null
            WriteLog  "HPC $nodetype installation task started"
            WriteLog "Waiting for HPC $nodetype installation ..."
            Start-Sleep -Seconds 1
            $retry = 0
            while($true)
            {
                $taskState = $hpcSetupTask.State
                # 2:Queued, 3:Ready, 4:Running
                if($taskState -eq 3)
                {
                    #If the task never run, the lastRunTime is December 30, 1899 12:00:00 AM
                    if($hpcSetupTask.LastRunTime -lt "1900-01-01")
                    {
                        $hpcSetupTask.Run($null) | Out-Null
                    }
                    else
                    {
                        if($hpcSetupTask.LastTaskResult -eq 0)
                        {
                            WriteLog  "HPC $nodetype installation completed"
                            break
                        }
                        else
                        {
                            # The setup task failed last time, re-run it
                            if($retry -lt 3)
                            {
                                $hpcSetupTask.Run($null) | Out-Null
                                $retry++
                            }
                            else
                            {
                                throw ("HPC $nodetype installation Failed: " + $hpcSetupTask.LastTaskResult)
                            }
                        }
                    }
                }
                elseif($taskState -eq 4)
                {
                    # if the setup hangs, kill the process to make the task fail, and wait for retry
                    $p = Get-Process -Name "HpcSetupChainer" -ErrorAction SilentlyContinue
                    if($null -ne $p)
                    {
                        $elapsedMins = ((Get-Date) - $p.StartTime).TotalMinutes
                        if($elapsedMins -gt 20)
                        {
                            WriteLog "HPC $nodetype installation hung, kill the HPC setup process."
                            $p.Kill()  | Out-Null
                        }
                    }
                }
                elseif($taskState -ne 2)
                {
                    throw "The HPC $nodetype installation task entered into unexpected state: $taskState"
                }

                Start-Sleep -Seconds 2
            }
        }
        catch
        {
            WriteLog "HPC $nodetype installation Failed"
            throw
        }
        finally
        {
            $Rootfolder.DeleteTask("hpcsetup",0)  | Out-Null
        }

        # Apply the patches if any
        $patchfiles = @(Get-ChildItem "C:\HPCPatches" -Filter "KB*.exe" | select -ExpandProperty FullName)
        if($patchfiles.Count -gt 0)
        {
            $patchTable = @{}
            foreach($pfile in $patchfiles)
            {
                $versionStr = [System.Diagnostics.FileVersionInfo]::GetVersionInfo($pfile).ProductVersion
                $version = New-Object System.Version $versionStr
                $patchTable[$version] = $pfile
            }

            $versions = @($patchTable.Keys | Sort-Object)
            foreach($ver in $versions)
            {
                $pfile = $patchTable[$ver]
                WriteLog "Applying QFE hotfix $pfile"
                $p = Start-Process -FilePath $pfile -ArgumentList "/unattend" -PassThru -Wait
                if(($p.ExitCode -ne 0) -and ($p.ExitCode -ne 3010))
                {
                    throw "Failed to apply QFE hotfix $pfile : $($p.ExitCode)"
                }

                Start-Sleep -Seconds 10
            }
        }
    }

    try
    {
        $setupScriptBytes = [IO.File]::ReadAllBytes("$env:HPCPSScriptRoot\SetupComplete2.cmd")
    }
    catch
    {
        throw ("Failed to load SetupComplete2.cmd :" + $_)
    }

    $joinDomainRetried = $false
    while($true)
    {
        try
        {
            RemotePshCall -ServiceName $ServiceName -VMName $Name -Credential $LocalAdminCred -ScriptBlock $rmPshBlock -ArgumentList @($ClusterName, $NodeType) | Out-Null
            break
        }
        catch [System.Management.Automation.RemoteException]
        {
            if($_.Exception.Message.Contains("is not domain joined"))
            {
                if($joinDomainRetried)
                {
                    throw
                }
                else
                {
                    TraceWarning "$Name is not domain joined, trying to join domain $DomainFQDN"
                    VMJoinToDomain -ServiceName $ServiceName -VMName $Name -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN
                    $joinDomainRetried = $true
                }
            }
            else
            {
                throw
            }
        }
    }
    
    $connUri = InvokeAzureCmdWithRetry -Command "Get-AzureWinRMUri -ServiceName $ServiceName -Name $Name" -RetryTimes 5
    if($null -eq $connUri)
    {
        throw "Failed to retrieve the WinRM https listener URI of the virtual machine $Name in cloud service $ServiceName."
    }

    $rmPshBlock = {
        param($nodetype, $dnsHostName, $setupscript)

        WriteLog "Generating $env:windir\OEM\SetupComplete2.cmd"
        [IO.File]::WriteAllBytes("$env:windir\OEM\SetupComplete2.cmd", $setupscript)

        Remove-Item -Path "C:\HPCPatches" -Recurse -Force -ErrorAction SilentlyContinue

        WriteLog "Removing HPC setup logs"
        Remove-Item -Path "$env:windir\Temp\HPCSetupLogs" -Recurse -Force -ErrorAction SilentlyContinue

        # Stop the HPC services so that we can remove the log files
        $hpcServices = @("HpcManagement", "HpcNodeManager", "msmpi", "HpcMonitoringClient", "HpcSoaDiagMon")
        if($nodetype -eq "BrokerNode")
        {
            $hpcServices += "HpcBroker"
        }

        $hpcServices | Stop-Service -Force -ErrorAction SilentlyContinue

        WriteLog "Removing HPC operation log files"
        $datadir = [System.Environment]::GetEnvironmentVariable("CCP_DATA", [System.EnvironmentVariableTarget]::Machine)
        $logFilesPath = [System.IO.Path]::Combine($datadir, "LogFiles")
        if(Test-Path $logFilesPath -PathType Container)
        {
            Get-ChildItem -Path $logFilesPath | select -ExpandProperty FullName | Remove-Item -Force -Recurse -ErrorAction SilentlyContinue
        }

        Remove-Item "$env:SYSTEMDRIVE\AzureData\CustomData.bin" -Force -ErrorAction SilentlyContinue

        WriteLog "Clear registry value ClusterName"
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC -Name ClusterName -Value ""
        if(Test-Path HKLM:\SOFTWARE\Wow6432Node\Microsoft\HPC)
        {
            Set-ItemProperty -Path HKLM:\SOFTWARE\Wow6432Node\Microsoft\HPC -Name ClusterName -Value ""
        }

        WriteLog "Removing cloud service certificate"
        $myStore = New-Object System.Security.Cryptography.x509Certificates.x509Store("My","LocalMachine")
        try
        {
            $myStore.Open("ReadWrite")
            $svcCerts = @($myStore.Certificates | ?{$_.Subject -eq "CN=$dnsHostName"})
            foreach($cert in $svcCerts)
            {
                $myStore.Remove($cert)
            }
        }
        catch
        {
            WriteLog "Failed to remove cloud service certificate"
        }
        finally
        {
            # Close doesn't throw even when Open not successfully called
            $myStore.Close()
        }

        # Sysprep the VM
        WriteLog "Syspreping HPC $nodetype $env:COMPUTERNAME"
        $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "$env:windir\System32\Sysprep\sysprep.exe /oobe /quiet /generalize /shutdown"
        $p = $null
        if($pobj.ReturnValue -eq 0)
        {
            $p = Get-Process -Id $pobj.ProcessId -ErrorAction SilentlyContinue
        }

        if($null -ne $p)
        {
            $p.WaitForExit()
        }
        else
        {
            throw "Failed to sysprep $env:COMPUTERNAME"
        }
    }

    $loopRetry = 0
    # Set the remote powershell call retry times to 30 for the first loop because the VM may not be ready to accept remote powershell call
    while($true)
    {
        try
        {
            # Error code 995 is expected here: The I/O operation has been aborted because of either a thread exit or an application request. 
            # Error code 5 is sometimes thrown when the remote call execution succeeded: Processing data for a remote command failed with the following error message: Access is denied. 
            RemotePshCall -ServiceName $ServiceName `
                          -VMName $Name `
                          -Credential $LocalAdminCred `
                          -ScriptBlock $rmPshBlock `
                          -ArgumentList @($NodeType, $connUri.Host, $setupScriptBytes) `
                          -IgnoreErrorCodes 995 `
                          -NoRetryErrorCodes 5 `
                          -RetryTimes 3 | Out-Null

            break
        }
        catch [System.Management.Automation.Remoting.PSRemotingTransportException]
        {
            if($loopRetry -lt 3)
            {
                try
                {
                    WaitForVMReady -VMName $Name -ServiceName $ServiceName -TargetStatus "StoppedVM" -MaxWaitSeconds 300
                    break
                }
                catch
                {
                    TraceWarning "$Name didn't enter into StoppedVM state"
                }
                
                $loopRetry++
            }
            else
            {
                throw
            }
        }
    }
    
    $baseVM = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $Name"
    $baseVMDiskName = $baseVM.VM.OSVirtualHardDisk.DiskName
    WaitForVMReady -VMName $Name -ServiceName $ServiceName -TargetStatus "StoppedVM"
    $TargetImageName = NewBaseHPCNodeImageName -NodeType $NodeType
    TraceInfo "Creating base image $TargetImageName from $NodeType $Name"
    Save-AzureVMImageWithRetry -Name $Name -ServiceName $ServiceName -ImageName $TargetImageName -ImageLabel "HPC Pack $NodeType Image"
    return @{"ImageName"=$TargetImageName;"CorpDiskName"=$baseVMDiskName}
}

function StartHPCNodesForService
{

    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $VMNames, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize, 
        
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 
        
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName, 

        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$false)]
        [PSCredential] $DomainUserCred, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$true)]
        [String] $CustomDataFile,

        [Parameter(Mandatory=$false)]
        [Object[]] $VMExtensions = @(),

        [Parameter(Mandatory=$false)]
        [Switch] $Linux,

        [Parameter(Mandatory=$false)]
        [Object] $SSHKeyPairForRoot = $null
    )

    TraceInfo ("Starting a job to deploy the following nodes in cloud service ${ServiceName}: " + ($VMNames -join ","))
    $sub = GetAzureSubscription -Current
    if($null -eq $sub)
    {
        throw "StartHPCNodesForService: current Azure subscription is not available."
    }
    $job = Start-Job -Name $ServiceName -ScriptBlock {
        param(
            $subscriptionId, $StorageAccountName, $VMNames, $VMSize, $ServiceName, $VNetName, $SubnetName, $DeploymentLabel, $LocalAdminCred, 
            $DomainUserCred, $DomainFQDN, $ImageName, $CustomDataFile, $VMExtensions, $IsLinux, $SSHKeyPairForRoot
        )

        ### Sleep a 0-10 secinds ramdom time to mitigate concurrent access to Azure profile
        Start-Sleep -Seconds (Get-Random -Minimum 0 -Maximum 10)
        try
        {
            $initRetry = 0
            while($true)
            {
                try
                {
                    [System.Net.ServicePointManager]::DefaultConnectionLimit = 40
                    Import-Module -Name Azure
                    break
                }
                catch
                {
                    ### Exception may be thrown when Import-Module
                    ### Simply retry several times in this case
                    if($initRetry -lt 10)
                    {
                        Start-Sleep -Seconds (Get-Random -Minimum 2 -Maximum 10)
                        $initRetry++
                    }
                    else
                    {
                        throw
                    }
                }
            }

            . "$env:HPCPSScriptRoot\HPCIaaSClusterUtils.ps1"
            . "$env:HPCPSScriptRoot\IaaSAzureCmdLib.ps1"
            . "$env:HPCPSScriptRoot\HPCIaaSClusterConfigLib.ps1"
            . "$env:HPCPSScriptRoot\HPCIaaSClusterLib.ps1"

            Register-EngineEvent -SourceIdentifier "DeployHPCNodeInService" -Forward

            SelectAzureSubscription -SubscriptionId $subscriptionId
            SetSubscriptionStorageAccount -SubscriptionId $subscriptionId -StorageAccountName $StorageAccountName

            if(-not $IsLinux)
            {
                $netbios = $DomainFQDN.Split(".")[0].ToUpper()
                $domainUserName = $DomainUserCred.UserName
                if($DomainUserCred.UserName.Contains("\"))
                {
                    $domainUserName = $domainUserName.Split("\")[1]
                }
            }
            else
            {
                ### for Linux node
                if($SSHKeyPairForRoot -eq $null)
                {
                    $dnsName = "HPCLinuxComputeNodes.$DomainFQDN"
                    $cert = Get-ChildItem Cert:\CurrentUser\My | ?{$_.HasPrivateKey -and $_.Subject -eq "CN=$dnsName"} | select -First(1)
                    if($null -eq $cert)
                    {
                        throw "Failed to create Linux compute nodes in cloud service $ServiceName : no SSH key pair found for Linux compute nodes."
                    }

                    $thumbprint = $cert.Thumbprint
                }
                New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "InfoLevel" -MessageData "Adding certificate (SSH key pair) to cloud service $ServiceName for Linux compute nodes." | Out-Null
                $retry = 0
                while($true)
                {
                    try
                    {
                        if($SSHKeyPairForRoot -eq $null)
                        {
                            Add-AzureCertificate -ServiceName $ServiceName -CertToDeploy (Get-Item "Cert:\CurrentUser\My\$thumbprint") -ErrorAction Stop
                        }
                        else
                        {
                            $sshCert = GetCertificate $SSHKeyPairForRoot.PfxFile $SSHKeyPairForRoot.Password
                            $thumbprint = $sshCert.Thumbprint
                            Add-AzureCertificate -ServiceName $ServiceName -CertToDeploy $SSHKeyPairForRoot.PfxFile -Password $SSHKeyPairForRoot.Password -ErrorAction Stop
                        }
                        break
                    }
                    catch
                    {
                        New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "WarningLevel" -MessageData "Failed to add certificate (SSH key pair) to cloud service $ServiceName for Linux compute nodes: $_" | Out-Null
                        if($retry -lt 5)
                        {
                            $retry++
                        }
                        else
                        {
                            throw
                        }
                    }
                }

                $sshKeyPair = New-AzureSSHKey -KeyPair -Fingerprint $thumbprint -Path '/root/.ssh/id_rsa'
                $sshPublicKey = New-AzureSSHKey -PublicKey -Fingerprint $thumbprint -Path '/root/.ssh/authorized_keys'
            }
            $vmConfigs = @{}
            foreach($vmName in $VMNames)
            {
                if($IsLinux)
                {
                    $vmConfigs[$vmName] = NewVMConfig -Linux -VMName $vmName -VMSize $VMSize -ImageName $ImageName -HostCaching ReadOnly -SubnetName $SubnetName `
                        -LocalAdminCred $LocalAdminCred -CustomDataFile $CustomDataFile -VMExtensions $VMExtensions -SSHKeyPairs $sshKeyPair -SSHPublicKeys $sshPublicKey
                }
                else
                {
                    $vmConfigs[$vmName] = NewVMConfig -WindowsDomain -VMName $vmName -VMSize $VMSize -ImageName $ImageName -HostCaching ReadOnly `
                        -SubnetName $SubnetName -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN `
                        -CustomDataFile $CustomDataFile -VMExtensions $VMExtensions
                }
            }                

            $maxRetries = 3
            $retry = 0
            while(($vmConfigs.Count -gt 0) -and ($retry -lt $maxRetries))
            {
                $leftVMNames = @($vmConfigs.Keys | sort)
                foreach($vmName in $leftVMNames)
                {
                    try
                    {
                        New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "InfoLevel" -MessageData "Creating the virtual machine $vmName" | Out-Null
                        New-AzureVMWithRetry -VM $vmConfigs[$vmName] -ServiceName $ServiceName -VNetName $VNetName -DeploymentLabel $DeploymentLabel -RetryTimes 4
                        $vmConfigs.Remove($vmName)
                    }
                    catch
                    {
                        New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "WarningLevel" -MessageData "Failed to create VM $vmName : $_" | Out-Null
                    }
                }

                $retry++
            }

            if($vmConfigs.Count -gt 0)
            {
                $leftVMNames = @($vmConfigs.Keys | sort)
                New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "ErrorLevel" -MessageData "Failed to create $($vmConfigs.Count) node(s): $($leftVMNames -join ',')" | Out-Null
            }
    
            $startingVMs = @($VMNames | ?{$null -eq $vmConfigs[$_]})
            if($startingVMs.Count -gt 0)
            {
                New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "InfoLevel" -MessageData "$($startingVMs.Count) nodes created in cloud service $ServiceName, waiting for them to become ready." | Out-Null
                try
                {
                    WaitForVMReady -VMName $startingVMs -ServiceName $ServiceName
                    New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "InfoLevel" -MessageData "The virtual machines in cloud service $ServiceName are now ready for use: $($startingVMs -join ',')"  | Out-Null
                }
                catch
                {
                    New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "ErrorLevel" -MessageData "$_" | Out-Null
                }
            }
        }
        catch
        {
            # If we don't add try-catch here, the script will continue running even when some exceptions happens
            New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "ErrorLevel" -MessageData ($_ | Out-String)
            throw
        }
    } -ArgumentList @($sub.SubscriptionId, $StorageAccountName, $VMNames, $VMSize, $ServiceName, $VNetName, $SubnetName, `
                $DeploymentLabel, $LocalAdminCred, $DomainUserCred, $DomainFQDN, $ImageName, $CustomDataFile, $VMExtensions, $Linux, $SSHKeyPairForRoot)
    return $job
}

function CreateHPCNodeGroup
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceVMTable, 

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceStorageTable, 
        
        [Parameter(Mandatory=$true)]
        [String] $ClusterName, 
        
        [Parameter(Mandatory=$true)]
        [String] $HeadNodeServiceName, 
                
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$false)]
        [PSCredential] $DomainUserCred = $null, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [psobject] $NodeGroupConfig
    )
    
    TraceInfo "Start to create HPC node group: $($NodeGroupConfig.NodeGroupName)"
    if(($NodeGroupConfig.NodeType -ne "LinuxNode") -and ($null -eq $DomainUserCred))
    {
        throw "Parameter DomainUserCred not provided in function CreateHPCNodeGroup for NodeType $($NodeGroupConfig.NodeType)."
    }

    $baseVMDiskName = ""
    if($NodeGroupConfig.ImageHPCPackInstalled -or ($NodeGroupConfig.NodeType -eq "LinuxNode"))
    {
        # If the source image is a valid HPC CN/BN image or a Linux image, directly use it to create HPC node
        $TargetImageName = $NodeGroupConfig.ImageName
    }
    else
    {
        $srcImg = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $($NodeGroupConfig.ImageName)"
        if($null -eq $srcImg)
        {
            throw "The source image doesn't exist: $($NodeGroupConfig.ImageName)"
        }

        $baseServiceName = $ServiceVMTable.Keys | sort | select -First 1
        $baseVMName = $ServiceVMTable[$baseServiceName][0]
        $sub = GetAzureSubscription -Current
        if($null -eq $sub)
        {
            throw "CreateHPCNodeGroup: current Azure subscription is not available"
        }

        if($srcImg.Category -eq "user")
        {
            # if the src image is a user private image, we just create the HPCNodeBaseImage in the same storage account, so that no image copy needed.
            $srcImgMediaLink = GetVMImageMediaLink -Image $srcImg
            $srcImgStg = GetStorageAccountFromMediaLink $srcImgMediaLink
            SetSubscriptionStorageAccount -SubscriptionId $sub.SubscriptionId -StorageAccountName $srcImgStg
        }
        else
        {
            SetSubscriptionStorageAccount -SubscriptionId $sub.SubscriptionId -StorageAccountName $ServiceStorageTable[$baseServiceName]
        }

        $ret = CreateHPCNodeBaseImage -Name $baseVMName -VMSize $NodeGroupConfig.VMSize -ServiceName $baseServiceName -ClusterName $ClusterName -HeadNodeServiceName $HeadNodeServiceName `
                    -NodeType $NodeGroupConfig.NodeType -VNetName $VNetName -SubnetName $SubnetName -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred `
                    -DomainFQDN $DomainFQDN -SourceImageName $NodeGroupConfig.ImageName
        $TargetImageName = $ret["ImageName"]
        $baseVMDiskName = $ret["CorpDiskName"]
    }

    $vmExts = @()
    if($null -ne $NodeGroupConfig.VMExtensions -and $NodeGroupConfig.VMExtensions.Count -gt 0)
    {
        $vmExts = $NodeGroupConfig.VMExtensions
    }
    if($NodeGroupConfig.NodeType -eq "LinuxNode")
    {
        $IsLinux = $true
        $ClusterFQDN = "{0}.{1}" -f $ClusterName, $DomainFQDN
        $privateConfig = '{"ClusterName": "{CLUSTERFQDN}"}'
        $privateConfig = $privateConfig.Replace('{CLUSTERFQDN}', $ClusterFQDN)
        $nodeAgentExt = $vmExts | ?{($_.ExtensionName -eq "LinuxNodeAgent") -and ($_.Publisher -eq "Microsoft.HpcPack")} | select -First(1)
        if($null -eq $nodeAgentExt)
        {
            TraceInfo "VM extension Microsoft.HpcPack.LinuxNodeAgent was not specified in the config file, set it with 1.* version"
            # if the LinuxNodeAgent isn't specified, specify it with Version 1.*
            $nodeAgentExt = NewVMExtensionConfig -ExtensionName "LinuxNodeAgent" -Publisher "Microsoft.HpcPack" -Version "1.*" -PrivateConfiguration $privateConfig
            $vmExts += $nodeAgentExt
        }
        else
        {
            TraceInfo "VM extension Microsoft.HpcPack.LinuxNodeAgent was already specified with version $($nodeAgentExt.Version)"
            # if the LinuxNodeAgent is already specified, just set the PrivateConfiguration in case it was not set or set incorrectly.
            $nodeAgentExt.PrivateConfiguration = $privateConfig
        }
    }
    else
    {
        $IsLinux = $false
        if(IsVMSizeWithRDMA -VMSize $NodeGroupConfig.VMSize)
        {
            $RDMAExt = $vmExts | ?{($_.ExtensionName -eq "HpcVmDrivers") -and ($_.Publisher -eq "Microsoft.HpcCompute")} | select -First(1)
            if($null -eq $RDMAExt)
            {
                TraceInfo "VM extension Microsoft.HpcCompute.HpcVmDrivers was not specified in the config file, set it with 1.* version"
                # for VM Sizes with RDMA, if HpcVmDrivers not specified, we will specify it with Version 1.*
                $RDMAExt = NewVMExtensionConfig -ExtensionName "HpcVmDrivers" -Publisher "Microsoft.HpcCompute" -Version "1.*"
                $vmExts += $RDMAExt
            }
            else
            {
                TraceInfo "VM extension Microsoft.HpcCompute.HpcVmDrivers was already specified with version $($RDMAExt.Version)"
            }
        }
    }

    DeployHPCNodesWithImage -ServiceVMTable $ServiceVMTable -ServiceStorageTable $ServiceStorageTable -ClusterName $ClusterName -VMSize $NodeGroupConfig.VMSize -VNetName $VNetName `
            -SubnetName $SubnetName -DeploymentLabel $DeploymentLabel -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN `
            -ImageName $TargetImageName -VMExtensions $vmExts -Linux:$IsLinux -SSHKeyPairForRoot $NodeGroupConfig.SSHKeyPairForRoot

    # Remove the corpse disk of the base node
    if(-not [String]::IsNullOrEmpty($baseVMDiskName))
    {
        try
        {
            RemoveCorpAzureDisk -DiskName $baseVMDiskName
        }
        catch
        {
            # Just give a warning message
            TraceWarning "Failed to remove the Azure disk $baseVMDiskName, you can remove it manually."
        }
    }

    Write-Output $TargetImageName
}

function GetTemplateParameters
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $Template
    )

    $params = @{}
    $params["TemplateName"] = $Template.TemplateName.Trim()
    $params["SubscriptionId"] = $Template.SubscriptionId.Trim()
    $params["CertificateThumbprint"] = $Template.CertificateId.Trim()
    $params["ServiceName"] = $Template.ServiceName.Trim()
    $params["StorageServiceName"] = $Template.StorageAccount.Trim()
    if($null -ne $Template["NodeCount"])
    {
        $nodeCount = [Int]$Template.NodeCount.Trim()
        if($nodeCount -gt 0)
        {
            $params["NodeCount"] = $nodeCount
            $params["RoleSize"] = $Template.RoleSize.Trim()
        }
    }

    if($null -eq $Template["Proxy"])
    {
        $params["UsesStaticProxyCount"] = "True"
        $params["ProxyMultiplicity"] = "2"
    }
    else
    {
        if($Template.Proxy.UsesStaticProxyCount.Trim() -eq "true")
        {
            $params["UsesStaticProxyCount"] = "True"
            if($null -ne $Template.Proxy["ProxyMultiplicity"])
            {
                $params["ProxyMultiplicity"] = $Template.Proxy.ProxyMultiplicity.Trim()
            }
        }
        else
        {
            $params["UsesStaticProxyCount"] = "False"
            if($null -ne $Template.Proxy["ProxyRatio"])
            {
                $params["ProxyRatio"] = $Template.Proxy.ProxyRatio.Trim()
            }

            if($null -ne $Template.Proxy["ProxyRatioBase"])
            {
                $params["ProxyRatioBase"] = $Template.Proxy.ProxyRatioBase.Trim()
            }
        }
    }

    if($null -ne $Template["OSVersion"])
    {
        $osversion = $Template.OSVersion.Trim()
        if($osversion -eq "WindowsServer2008R2")
        {
            $params["WorkerOSFamily"] = "Win2008R2"
        }
        elseif($osversion -eq "WindowsServer2012")
        {
            $params["WorkerOSFamily"] = "Win8"
        }
    }

    if($null -ne $Template["VhdDriveUrl"])
    {
        $params["VhdDriveUrl"] = $Template.VhdDriveUrl.Trim()
    }

    if($null -ne $Template["StartupScript"])
    {
        $params["StartupScript"] = $Template.StartupScript.Trim()
    }

    return $params
}

function GetAzureBustConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $AzureBustConfigXml,

        [Parameter(Mandatory=$true)]
        [Object[]] $Certificates
    )

    $xmlBurst = ([xml]$AzureBustConfigXml).AzureBurst
    $certIdThumbprintMap = @{}
    foreach($cert in $Certificates)
    {
        $certIdThumbprintMap[$cert.Id] = $cert.Thumbprint
    }

    $templateList = @()
    foreach($temp in $xmlBurst.AzureNodeTemplate)
    {
        $tempParams = GetTemplateParameters $temp
        $certId = [Int]$tempParams["CertificateThumbprint"]
        $tempParams["CertificateThumbprint"] = $certIdThumbprintMap[$certId]
        $templateList += $tempParams
    }

    return $templateList
}

function CreateAzureNodeTemplates
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [PSCredential] $SetupUserCred,

        [Parameter(Mandatory=$true)]
        [Object[]] $NodeTemplates,

        [Parameter(Mandatory=$false)]
        [String] $PostConfigScirptFile = ""
    )

    TraceInfo "Creating Azure node templates on the HPC head node $VMName"

    $DefaultAzureNodeTemplateXml = @'
<?xml version="1.0" encoding="utf-8"?>
<Template xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Description="" Version="6">
  <Item Name="DeployAzureNode">
    <Parameter Name="CertificateThumbprint" Value="" />
    <Parameter Name="SubscriptionId" Value="" />
    <Parameter Name="StorageServiceName" Value="" />
    <Parameter Name="ServiceName" Value="" />
    <Parameter Name="UsesStaticProxyCount" Value="True" />
    <Parameter Name="ProxyMultiplicity" Value="2" />
    <Parameter Name="ProxyRatio" Value="200" />
    <Parameter Name="ProxyRatioBase" Value="300" />
    <Parameter Name="WorkerOSFamily" Value="Win8R2" />
    <Parameter Name="VhdDriveUrl" Value="" />
    <Parameter Name="UsesVMImage" Value="False" />
    <Parameter Name="AzureAdminUserName" Value="" />
    <Parameter Name="AzureAdminPassword" Value="" />
    <Parameter Name="StartupScript" Value="" />
    <ParameterList Name="SubNets" />
    <Parameter Name="Description" Value="Deploys an Windows Azure node with a given subscription, service and specified role type." />
  </Item>
  <Item Name="DeploymentSchedule">
    <Parameter Name="ManualAvailability" Value="True" />
    <Parameter Name="SerializedDeploymentSchedule" Value="000000000000000000000000000000000000000000000000000000000000000000000000000000000000" />
    <Parameter Name="GracePeriod" Value="0" />
    <Parameter Name="Description" Value="Determines when a given node is started and stopped." />
  </Item>
</Template>
'@

    $baseNodeTemp = [xml]$DefaultAzureNodeTemplateXml

    # Set the user name and password used for RDP to the Azure Nodes
    $usrName = $SetupUserCred.UserName.Split('\')[-1]
    $usrPsw = $SetupUserCred.GetNetworkCredential().Password
    $node = $baseNodeTemp.SelectSingleNode("//Parameter[@Name='AzureAdminUserName']")
    if($null -ne $node)
    {
        $node.Value = $usrName
    }
    $node = $baseNodeTemp.SelectSingleNode("//Parameter[@Name='AzureAdminPassword']")
    if($null -ne $node)
    {
        $node.Value = $usrPsw
    }

    $postConfig = $null
    if((-not [String]::IsNullOrEmpty($PostConfigScirptFile)) -and (Test-Path -Path $PostConfigScirptFile -PathType Leaf))
    {
        $postConfig = [IO.File]::ReadAllBytes($PostConfigScirptFile)
    }

    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $SetupUserCred -ScriptBlock  {
        param([String]$baseTemplate, [Int]$templateNum, [Object[]]$NodeTemplates, [Byte[]]$postConfigContent)

        Add-PSSnapin Microsoft.HPC
        $rootDir = "$env:TEMP\HPCAzureBurst" + (Get-Random)
        try
        {
            New-Item -ItemType directory -Path $rootDir -Force | Out-Null
            foreach($tempParams in $NodeTemplates)
            {
                try
                {
                    $templateName = $tempParams["TemplateName"]
                    WriteLog "Creating Azure node template $templateName."
                    $tempXml = [xml]$baseTemplate
                    foreach($param in $tempParams.Keys)
                    {
                        if(($param -eq "TemplateName") -or ($param -eq "NodeCount") -or ($param -eq "RoleSize"))
                        {
                            continue
                        }

                        $node = $tempXml.SelectSingleNode("//Parameter[@Name='$param']")
                        if($null -ne $node)
                        {
                            $node.Value = $tempParams[$param]
                        }
                    }

                    $templateFile = [IO.Path]::Combine($rootDir,  "$templateName.xml")
                    $tempXml.Save($templateFile)
                    Import-HpcNodeTemplate $templateFile -Force
                }
                catch
                {
                    throw ("Failed to create Azure node template $templateName :" + ($_ | Out-String)) 
                }

                try
                {
                    if($null -ne $tempParams["NodeCount"])
                    {
                        WriteLog "Adding Azure nodes for Azure node template $templateName."
                        Get-HpcNodeTemplate -Name $templateName | Add-HpcNodeSet -Quantity $tempParams["NodeCount"] -Size $tempParams["RoleSize"]
                    }
                }
                catch
                {
                    throw ("Failed to create Azure nodes for Azure node template $templateName :" + ($_ | Out-String)) 
                }
            }

            if(($null -ne $postConfigContent) -and ($postConfigContent.Count -gt 0))
            {
                $postConfigScriptFile = "$env:temp\HPCHNPostConfig.ps1"
                try
                {
                    WriteLog "Creating $postConfigScriptFile"
                    [IO.File]::WriteAllBytes($postConfigScriptFile, $postConfigContent)
                    WriteLog "$postConfigScriptFile created"
                    $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "PowerShell.exe -ExecutionPolicy Unrestricted -File `"$postConfigScriptFile`""
                    if($pobj.ReturnValue -eq 0)
                    {
                        WriteLog "Start to run $postConfigScriptFile."
                    }
                    else
                    {
                        WriteLog "Failed to run $postConfigScriptFile."
                    }
                }
                catch
                {
                    WriteLog "Failed to run $postConfigScriptFile : $_"
                }
            }
        }
        finally
        {
            Remove-Item -Path $rootDir -Recurse -Force -ErrorAction SilentlyContinue
        }

    } -ArgumentList @($baseNodeTemp.OuterXml, $NodeTemplates.Count, $NodeTemplates, $postConfig) | Out-Null

    TraceInfo "Azure node templates created"
}


function SetSubscriptionStorageAccount
{
    param
    (
        [Parameter(Mandatory=$true, ParameterSetName="SetByName")]
        [String] $SubscriptionName, 

        [Parameter(Mandatory=$true, ParameterSetName="SetById")]
        [String] $SubscriptionId, 

        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName
    )

    $curSubscription = GetAzureSubscription -Current
    if($PsCmdlet.ParameterSetName -eq "SetByName")
    {
        if(($null -ne $curSubscription) -and ($curSubscription.SubscriptionName -eq $SubscriptionName))
        {
            $subscription = $curSubscription
        }
        else
        {
            $subscription = GetAzureSubscription -SubscriptionName $SubscriptionName
            if($null -eq $subscription)
            {
                throw "The Azure subscription $SubscriptionName was not configured."
            }
        }

        $SubscriptionId = $subscription.SubscriptionId
    }
    else
    {
        if(($null -ne $curSubscription) -and ($curSubscription.SubscriptionId -eq $SubscriptionId))
        {
            $subscription = $curSubscription
        }
        else
        {
            $subscription = GetAzureSubscription -SubscriptionId $SubscriptionId
            if($null -eq $subscription)
            {
                throw "The Azure subscription $SubscriptionId was not configured."
            }
        }

        $SubscriptionName = $subscription.SubscriptionName
    }

    if($subscription.CurrentStorageAccountName -ne $StorageAccountName)
    {
        # use cross-process mutex to prevent from calling Set-AzureSubscription concurrently
        $setSubMutex = New-Object System.Threading.Mutex $false,'HPCIaaSSetAzureSubscriptionMutex'
        $mutexAcquired = $false
        try
        {
            try
            {
                $mutexAcquired = $setSubMutex.WaitOne()
            }
            catch [System.Threading.AbandonedMutexException]
            {
                $mutexAcquired = $true
            }

            $setSubRetry = 0
            while($true)
            {
                try
                {
                    Set-AzureSubscription -SubscriptionId $SubscriptionId -CurrentStorageAccountName $StorageAccountName -ErrorAction Stop | Out-Null
                    TraceInfo "Set current storage account to $StorageAccountName for subscription $SubscriptionName ($SubscriptionId)"
                    break
                }
                catch
                {
                    $exMsg = "Failed to set current storage account: " + ($_ | Out-String)
                    if($setSubRetry -lt 3)
                    {
                        TraceWarning $exMsg
                        Start-Sleep -Seconds (Get-Random -Minimum 2 -Maximum 5)
                        $setSubRetry++
                    }
                    else
                    {
                        throw $exMsg
                    }
                }
            }
        }
        finally
        {
            if($mutexAcquired)
            {
                $setSubMutex.ReleaseMutex()
            }
        }
    }
}

function GetStorageBlobEndpoint
{
    param
    (
        [Parameter(Mandatory=$true)]
        $StorageAccount
    )

    $storageBlobEndPoint = $StorageAccount.Endpoints | ?{$_ -match ".blob."}
    $storageBlobSuffix = $storageBlobEndPoint.Substring($storageBlobEndPoint.IndexOf(".blob.")).TrimEnd("/")
    return ($StorageAccount.StorageAccountName + $storageBlobSuffix)
}

function ResolveStorageEndpointIP
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Endpoint
    )
    
    $retry = 0
    while($retry -lt 3)
    {
        try
        {
            $ips = [System.Net.Dns]::GetHostAddresses($Endpoint)
            return $ips[0].IPAddressToString
        }
        catch
        {
            $retry++
            Clear-DnsClientCache
        }
    }
    
    throw "The Endpoint cannot be resolved: $Endpoint"  
}

function DeployHPCNodesWithImage
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceVMTable, 

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceStorageTable, 

        [Parameter(Mandatory=$true)]
        [String] $ClusterName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 

        [Parameter(Mandatory=$false)]
        [PSCredential] $DomainUserCred = $null, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$false)]
        [Object[]] $VMExtensions = @(),

        [Parameter(Mandatory=$false)]
        [Switch] $Linux,

        [Parameter(Mandatory=$false)]
        [Object] $SSHKeyPairForRoot = $null
    )

    if($Linux.IsPresent -and ($null -eq $SSHKeyPairForRoot))
    {
        ### For Linux node, if SSH key pair for root is not specified, we will use a self-signed certificate
        TraceInfo "'SSHKeyPairForRoot' is not specified for Linux compute nodes, use an auto-generated SSH key pair for test purpose."
        $dnsName = "HPCLinuxComputeNodes.$DomainFQDN"
        $certFound = $false
        $linuxCerts = @(Get-ChildItem Cert:\CurrentUser\My | ?{$_.HasPrivateKey -and $_.Subject -eq "CN=$dnsName"})
        $targetTime = (Get-Date).AddMonths(3)
        foreach($cert in $linuxCerts)
        {
            if($targetTime -gt $cert.NotAfter)
            {
                # Remove the certificate which will expire in 3 months
                Remove-Item -Path "Cert:\CurrentUser\My\$($cert.Thumbprint)" -Force -ErrorAction SilentlyContinue
            }
            else
            {
                TraceInfo "Self-signed certificate found for SSH key pair: $($cert.Thumbprint)"
                $certFound = $true
            }
        }

        if(-not $certFound)
        {
            $thumbprint = CreateSelfSignedCertificate -DnsName $dnsName -ExpiryDate (Get-Date).AddYears(5) -Exportable
            TraceInfo "Generated a self-signed certificate for SSH key pair: $thumbprint"
        }
    }

    Register-EngineEvent -SourceIdentifier "DeployHPCNodeInService" -Action {
        try
        {
            if($Sender -eq "ErrorLevel")
            {
                TraceError $event.MessageData
            }
            elseif($Sender -eq "WarningLevel")
            {
                TraceWarning $event.MessageData
            }
            else
            {
                TraceInfo $event.MessageData
            }
        }
        catch
        {
            #Ignore the error
        }
    }

    $provsionJobs = @()

    $hnNameFile = "$env:TEMP\HPCHN_{0}_{1}.txt" -f $ClusterName, (Get-Random)
    "HPCClusterName=$ClusterName" | Out-File $hnNameFile

    $allImageList = @(InvokeAzureCmdWithRetry -Command "Get-AzureVMImage")
    $srcImage = $allImageList | ?{$_.ImageName -eq $ImageName}
    if($null -eq $srcImage)
    {
        throw "The VM image $ImageName does not exist!"
    }

    if($srcImage.Category -eq "Public")
    {
        TraceInfo "No replication needed for public image $ImageName"
        foreach($svc in @($ServiceVMTable.Keys | sort))
        {
            $provsionJobs += StartHPCNodesForService -VMNames $ServiceVMTable[$svc] -VMSize $VMSize -ServiceName $svc -StorageAccountName $ServiceStorageTable[$svc] `
                     -VNetName $VNetName -SubnetName $SubnetName -DeploymentLabel $DeploymentLabel -LocalAdminCred $LocalAdminCred `
                     -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN -ImageName $ImageName -CustomDataFile $hnNameFile `
                     -VMExtensions $VMExtensions -Linux:$Linux -SSHKeyPairForRoot $SSHKeyPairForRoot
        }        
    }
    else
    {
        $curTime = Get-Date
        $userImageList = @($allImageList | ?{$_.Category -eq "User"})
        $existingStorages = @(InvokeAzureCmdWithRetry -Command "Get-AzureStorageAccount")
        $DestStorageAccounts = @($ServiceVMTable.Keys | %{$ServiceStorageTable[$_]} | select -Unique)
        $sub = GetAzureSubscription -Current
        if($null -eq $sub)
        {
            throw "DeployHPCNodesWithImage: Current Azure subscription is not available."
        }
        $environment = $sub.Environment


        $srcImageTimeStamp = GetVMImageTimeStamp $srcImage
        $srcMediaLink = GetVMImageMediaLink $srcImage
        $srcStorageAccount = GetStorageAccountFromMediaLink $srcMediaLink
        $srcLocation = $srcImage.Location
        if([String]::IsNullOrEmpty($srcLocation))
        {
            # in case the storage account is in an affinity group
            $srcLocation = $existingStorages | ?{$_.StorageAccountName -eq $srcStorageAccount} | %{$_.GeoPrimaryLocation}
        }
        if($ImageName -match "_$srcStorageAccount$")
        {
            $imageNameRoot = $ImageName.Substring(0, $ImageName.Length - $srcStorageAccount.Length - 1)
        }
        else
        {
            $imageNameRoot = $ImageName
        }

        $srcUris = @{}
        $srcUris[$srcStorageAccount] = $srcMediaLink
        foreach($img in $userImageList)
        {
            if(-not $img.ImageName.StartsWith($imageNameRoot))
            {
                continue
            }

            $mediaLink = GetVMImageMediaLink $img
            $stg = GetStorageAccountFromMediaLink $mediaLink
            $imgLocation = $img.Location
            if([String]::IsNullOrEmpty($imgLocation))
            {
                # in case the storage account is in an affinity group
                $imgLocation = $existingStorages | ?{$_.StorageAccountName -eq $stg} | %{$_.GeoPrimaryLocation}
            }
        
            if($imgLocation -ne $srcLocation)
            {
                continue
            }

            if(($img.ImageName -eq $imageNameRoot) -or ($img.ImageName -eq "${imageNameRoot}_${stg}"))
            {
                $imgTimeStamp = GetVMImageTimeStamp $img
                $imgName = $img.ImageName
                if(($ImageName -eq $imgName) -or ($imgTimeStamp -eq $srcImageTimeStamp))
                {
                    $srcUris[$stg] = $mediaLink
                    # It is possible that some source images are not in the destination storage accounts
                    if($DestStorageAccounts -contains $stg)
                    {
                        TraceInfo "The image $imgName already exists in the storage account $stg"
                        $DestStorageAccounts = @($DestStorageAccounts | ?{$_ -ne $stg})
                        foreach($svc in @($ServiceVMTable.Keys | ?{$ServiceStorageTable[$_] -eq $stg}))
                        {
                            $provsionJobs += StartHPCNodesForService -VMNames $ServiceVMTable[$svc] -VMSize $VMSize -ServiceName $svc -StorageAccountName $stg `
                                                        -VNetName $VNetName -SubnetName $SubnetName -DeploymentLabel $DeploymentLabel -LocalAdminCred $LocalAdminCred `
                                                        -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN -ImageName $imgName -CustomDataFile $hnNameFile `
                                                        -VMExtensions $VMExtensions -Linux:$Linux -SSHKeyPairForRoot $SSHKeyPairForRoot
                        }
                    }
                }
                else
                {
                    # It is possible that some source images are not in the destination storage accounts
                    if($DestStorageAccounts -contains $stg)
                    {
                        TraceInfo "Removing the image $imgName from storage account $stg because its time stamp is different from that of the source image $ImageName"
                        Remove-AzureVMImageWithRetry -ImageName $imgName -DeleteVHD
                    }
                }
            }
        }

        $containerName = "hpcimages"
        $srcStorageTenants = @{}
        $destImageNames = @{}
        $destStorageTenants = @{}
        $storageContext = @{}
        $storageAccounts = @($DestStorageAccounts + $srcUris.Keys | select -Unique)
        foreach($stg in $storageAccounts)
        {
            $storage = $existingStorages | ?{$_.StorageAccountName -eq $stg}
            if($null -eq $storage)
            {
                throw "The storage account $stg doesn't exist"
            }

            $endPoint = GetStorageBlobEndpoint -StorageAccount $storage
            $ipStr = ResolveStorageEndpointIP -Endpoint $endPoint
            if($null -eq $srcUris[$stg])
            {
                $destImageNames[$stg] = "${imageNameRoot}_${stg}"
                $destStorageTenants[$ipStr] += @($stg)
            }
            else
            {
                # If multiple source storage accounts in one tenant, we just copy from one of them
                if($null -eq $srcStorageTenants[$ipStr])
                {
                    $srcStorageTenants[$ipStr] = $stg
                }
                else
                {
                    continue
                }
            }

            $storageKey = (InvokeAzureCmdWithRetry -Command "Get-AzureStorageKey -StorageAccountName $stg").Primary
            if([string]::IsNullOrEmpty($environment))
            {
                $storageContext[$stg] = New-AzureStorageContext -StorageAccountName $stg -StorageAccountKey $storageKey
            }
            else
            {
                $storageContext[$stg] = New-AzureStorageContext -StorageAccountName $stg -StorageAccountKey $storageKey -Environment $environment
            }
            ## Create the container on the destination ### 
            if($null -eq $srcUris[$stg])
            {
                $cont = Get-AzureStorageContainerWithRetry -Name $containerName -Context $storageContext[$stg]
                if($null -eq $cont)
                {
                    $null = New-AzureStorageContainerWithRetry -Name $containerName -Context $storageContext[$stg]
                }
            }
        }

        $destVHDName =  "$imageNameRoot.vhd"
        $copyMaps = @{}
        # Copy from storage accounts in the same tenants
        foreach($ipStr in $srcStorageTenants.Keys)
        {
            $srcStg = $srcStorageTenants[$ipStr]
            if($null -ne $destStorageTenants[$ipStr])
            {
                $copyMaps[$srcStg] = $destStorageTenants[$ipStr]
                $destStorageTenants.Remove($ipStr)
            }
        }

        # Copy from storage accounts in different tenants
        if($destStorageTenants.Count -gt 0)
        {
            $srcStgs = @($srcStorageTenants.Values)
            $destStorageCount = $destStorageTenants.Count
            $index = 0
            foreach($ipStr in $destStorageTenants.Keys)
            {
                $srcStgIndex = $index % $srcStgs.Count
                $srcStg = $srcStgs[$srcStgIndex]
                $copyMaps[$srcStg] += @($destStorageTenants[$ipStr][0])
                $index++
            }
        }

        $blobCopies = @{}
        foreach($stg in $DestStorageAccounts)
        {
            $blobCopies[$stg] = $null
        }

        $queueTime = @{}
        $startTime = @{}
        $doneTime = @{}
        try
        {
            foreach($srcStg in $copyMaps.Keys)
            {
                foreach($destStg in $copyMaps[$srcStg])
                {
                    TraceInfo "Start to copy $destVHDName to the storage account $destStg"
                    ### Start the asynchronous copy - specify the source authentication with -SrcContext ### 
                    $blobCopies[$destStg] += Start-AzureStorageBlobCopyWithRetry -AbsoluteUri $srcUris[$srcStg] `
                                                                                 -Context $storageContext[$srcStg] `
                                                                                 -DestContainer $containerName `
                                                                                 -DestBlob $destVHDName `
                                                                                 -DestContext $storageContext[$destStg]

                    $queueTime[$destStg] = Get-Date
                }
            }

            while($blobCopies.Count -gt $doneTime.Count)
            {
                foreach($stg in $DestStorageAccounts)
                {
                    if($null -eq $blobCopies[$stg])
                    {
                        # Not Started yet
                        continue                
                    }

                    if($null -ne $doneTime[$stg])
                    {
                        # Already completed
                        continue
                    }

                    try
                    {
                        $status = $blobCopies[$stg] | Get-AzureStorageBlobCopyState -ErrorAction Stop
                    }
                    catch
                    {
                        TraceWarning "Failed to get blob copy state for $stg : $_"
                        continue
                    }

                    if($status.Status -eq "Pending")
                    {
                        $copied = 0
                        if($null -ne (Get-Member -InputObject $status -Name "BytesCopied"))
                        {
                            $copied = $status.BytesCopied
                        }
                        if(($null -eq $startTime[$stg]) -and $copied -gt 0)
                        {
                            $startTime[$stg] = Get-Date
                            TraceInfo "Copying image to the storage account $stg"
                        }
                    }
                    elseif($status.Status -eq "Success")
                    {
                        # Completed
                        if($null -eq $startTime[$stg])
                        {
                            $startTime[$stg] = Get-Date
                        }
                        $doneTime[$stg] = Get-Date
                        TraceInfo "The VHD of $ImageName was copied to the storage account $stg successfully"

                        # Copy to other storage accounts in the same tenant
                        $tenantIP = $destStorageTenants.Keys | ?{$destStorageTenants[$_] -contains $stg}
                        if($null -ne $tenantIP)
                        {                        
                            foreach($otherstg in @($destStorageTenants[$tenantIP] | ?{$_ -ne $stg}))
                            {
                                if($null -eq $blobCopies[$otherstg])
                                {
                                    TraceInfo "Start to copy $destVHDName to the storage account $otherstg"
                                    $blobCopies[$otherstg] = Start-AzureStorageBlobCopyWithRetry -SrcBlob $destVHDName `
                                                                        -SrcContainer $containerName `
                                                                        -Context $storageContext[$stg] `
                                                                        -DestContainer $containerName `
                                                                        -DestBlob $destVHDName `
                                                                        -DestContext $storageContext[$otherstg]

                                    $queueTime[$otherstg] = Get-Date
                                }
                            }
                        }

                        $imgName = $destImageNames[$stg]
                        if($srcImageTimeStamp.Year -le 2000)
                        {
                            # If the src image time stamp is too old, use the current time.
                            $tmpImgTimeStamp = $curTime
                        }
                        else
                        {
                            $tmpImgTimeStamp = $srcImageTimeStamp
                        }

                        $OSType = "Windows"
                        if($Linux.IsPresent)
                        {
                            $OSType = "Linux"
                        }

                        Add-AzureVMImageWithRetry -ImageName $imgName -MediaLocation $blobCopies[$stg].ICloudBlob.Uri.ToString() -OS $OSType `
                                -Label $srcImage.Label -Description $srcImage.Description -PublishedDate $tmpImgTimeStamp | Out-Null
                        foreach($svc in @($ServiceVMTable.Keys | ?{$ServiceStorageTable[$_] -eq $stg}))
                        {
                            $provsionJobs += StartHPCNodesForService -VMNames $ServiceVMTable[$svc] -VMSize $VMSize -ServiceName $svc `
                                                    -StorageAccountName $stg -VNetName $VNetName -SubnetName $SubnetName -DeploymentLabel $DeploymentLabel `
                                                    -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN `
                                                    -ImageName $imgName -CustomDataFile $hnNameFile -VMExtensions $VMExtensions -Linux:$Linux `
                                                    -SSHKeyPairForRoot $SSHKeyPairForRoot
                        }
                    }
                    else
                    {                    
                        throw "Failed to copy the VHD of $ImageName to the storage account $stg, the status is $($status.Status)"
                    }
                }

                Start-Sleep -Seconds 20
            }
        }
        finally
        {
            foreach($stg in $blobCopies.Keys)
            {
                if($null -eq $doneTime[$stg])
                {
                    try
                    {
                        $blobCopies[$stg] | Stop-AzureStorageBlobCopy -ErrorAction Continue
                    }
                    catch
                    {
                        TraceWarning "Failed to stop Azure storage blob copy: $_"
                    }
                }
            }
        }
    }
    if($provsionJobs.Count -gt 0)
    {
        while($true)
        {
            $leftJobs = @()
            foreach($job in $provsionJobs)
            {
                $isLeftJob = $false
                if($job.State -eq "Failed")
                {
                    $errMsg = "The node provision job failed in cloud service {0}:{1}`n{2}`n{3}" -f `
                        $job.Name, `
                        $job.ChildJobs[0].JobStateInfo.Reason.Message, `
                        (GetExceptionTypeInfo $job.ChildJobs[0].JobStateInfo.Reason), `
                        ($job.ChildJobs[0].JobStateInfo.Reason.SerializedRemoteInvocationInfo | Out-String)
                    TraceError $errMsg
                }
                elseif($job.State -eq "Blocked")
                {
                    TraceError "The node provision job was blocked in cloud service $($job.Name)" 
                }
                elseif($job.State -eq "Completed")
                {
                    TraceInfo "The node provision job completed in cloud service $($job.Name)" 
                }
                else
                {
                    $isLeftJob = $true
                    $leftJobs += $job
                }

                if(-not $isLeftJob)
                {
                    TraceDetail "The detailed provision logs for cloud service $($job.Name) as below: "
                    TraceDetail "------ Start of VM provisioning in cloud service $($job.Name) ------" 
                    $job.ChildJobs[0].Verbose | Write-Verbose -Verbose
                    TraceDetail "------ End of VM provisioning in cloud service $($job.Name) ------" 
                }
            }
            if($leftJobs.Count -eq 0)
            {
                break
            }
            $provsionJobs = $leftJobs

            Start-Sleep -Seconds 10
        }
    }

    Remove-Item -Path $hnNameFile -Force
}

function UpdateDNSRecords
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $DCName, 
        
        [Parameter(Mandatory=$true)]
        [String] $DCServiceName,

        [Parameter(Mandatory=$true)]
        [PSCredential] $Credential, 

        [Parameter(Mandatory=$true)]
        [String] $ZoneName, 

        [Parameter(Mandatory=$true)]
        [HashTable] $HostIPTable
    )

    try
    {
        TraceInfo "Trying to update DNS entries for Linux compute nodes ..."
        RemotePshCall -VMName $DCName -ServiceName $DCServiceName -Credential $Credential -RetryTimes 10 -ScriptBlock {
            param ($zoneName, $hosts)

            $curTime = Get-Date
            $curEntries = @(Get-DnsServerResourceRecord -ZoneName $ZoneName -RRType A)
            foreach($entry in $curEntries)
            {
                $recordName = $entry.HostName
                ### Remove the entries conflicted with this update
                if(($hosts.Values -contains $entry.RecordData.IPv4Address) -or ($hosts.Keys -contains $recordName))
                {
                    WriteLog "Remove a stale DNS record: $recordName"
                    Remove-DnsServerResourceRecord -ZoneName $ZoneName -Name $recordName -RRType A -Force -ErrorAction SilentlyContinue
                }
            }

            foreach($hostname in $hosts.Keys)
            {
                $ipaddr = $hosts[$hostname]
                WriteLog "Adding DNS entry for $hostname, IP=$ipaddr"
                try
                {
                    Add-DnsServerResourceRecordA -ZoneName $zoneName -Name $hostname -AllowUpdateAny -IPv4Address $ipaddr
                    WriteLog "DNS entry added for $hostname"
                }
                catch
                {
                    WriteLog "Failed to add DNS entry for $hostname : $_"
                }
            }                    
        } -ArgumentList @($ZoneName, $HostIPTable) | Out-Null
    }
    catch
    {
        TraceWarning "Failed to add DNS entries for Linux compute nodes: $_"
    }
}

function SetAutoGrowShrinkState
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $HNVMName, 
        
        [Parameter(Mandatory=$true)]
        [String] $HNServiceName,

        [Parameter(Mandatory=$true)]
        [PSCredential] $Credential, 

        [Parameter(Mandatory=$true)]
        [ValidateSet("enable","disable")]
        [String] $State
    )

    try
    {
        TraceInfo "Trying to $State auto grow shrink ..."
        RemotePshCall -VMName $HNVMName -ServiceName $HNServiceName -Credential $Credential -RetryTimes 10 -ScriptBlock {
            param ($State)
            try
            {
                # Try to clear local dns cache in case IP addresses of some nodes have been changed.
                WriteLog "Flush the DNS cache on the head node."
                Clear-DnsClientCache -ErrorAction SilentlyContinue

                Add-PSSnapin Microsoft.HPC
                WriteLog "Try to $State auto grow shrink ..."
                if($State -eq "enable")
                {
                    Set-HpcClusterProperty -EnableGrowShrink 1  | Out-Null
                }
                else
                {
                    Set-HpcClusterProperty -EnableGrowShrink 0  | Out-Null
                }
                WriteLog "Succeeded to $State auto grow shrink."
            }
            catch
            {
                throw "Failed to $State auto grow shrink: $_"
            }
        } -ArgumentList $State | Out-Null
    }
    catch
    {
        TraceWarning "$_"
    }
}
# SIG # Begin signature block
# MIIaxwYJKoZIhvcNAQcCoIIauDCCGrQCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUdlo8nUoZzfHMkciwVLbhq7CG
# hHOgghWCMIIEwzCCA6ugAwIBAgITMwAAAHJp9pDCFDtJoQAAAAAAcjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUwMzIwMTczMjA0
# WhcNMTYwNjIwMTczMjA0WjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OjMxQzUtMzBCQS03QzkxMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoxPVJ2iBlQVE
# sJuqKGMDs9En3ra2ZyP2mieV2SfpWI7AXCdUfRHZ7B+712Ff5z/jWtu5iNakH+yw
# OsPZy911xkYMU2T9mm1IXZcHoKEMvj6SL5H3q2dq0SX5n3iqXJcxf4wwjDu9cKit
# 5v8t+rrQxeVIuseUK34MXOH/NYpAv1H0FU0vm0MjOh7ltNRzua86yAgrQBSj7gML
# NybNLvq+M9b3JBVF6cFLmII8syS9LSf7drlwWzPRJ71Q/6+TVq59UtTQO217Juvk
# 9BdtStzFmMyR+7VdVxybCwxPr6dW/2U7oouf4UxtszHDMtHw7XEnA2jg2iq6ODkZ
# frNXUQCMhQIDAQABo4IBCTCCAQUwHQYDVR0OBBYEFND9obr0IhS4jv6FL059LwBC
# pwFTMB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZI
# hvcNAQEFBQADggEBACJEd3U6Z3Bw2w+djrTKkrwYGl8Ys2vdoryZDDm9ybp6jjCk
# QCKLBGssgmRVJcEaYjUTkqGomLk5Mt4r5X6S+Zf1lgF3IKX9tqVbNlX1qMmRZUQ4
# 4Xfg7Ml8SuB1UqFU0DvuoIoQndg51sX2yA4Tahq0xD/ilyga1NHQ3jgbw5QKBZDt
# VbGGwfUgcNCwbZGDZCQtrRTcRmr4B7/tK4xFFOfvrKLEA8MByHdzmAYlIMsPM/fX
# YCl95S60wLsCjC5aJHBzM5NN2ZRQSLYLg0RKtrBoR5Zq+RlIBXKsS0hKgRbk9WAj
# uB5MvW3T/tl40KdnBlT9lCGh677+NM8EykDf1UIwggTsMIID1KADAgECAhMzAAAB
# Cix5rtd5e6asAAEAAAEKMA0GCSqGSIb3DQEBBQUAMHkxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBMB4XDTE1MDYwNDE3NDI0NVoXDTE2MDkwNDE3NDI0NVowgYMxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIx
# HjAcBgNVBAMTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJL8bza74QO5KNZG0aJhuqVG+2MWPi75R9LH7O3HmbEm
# UXW92swPBhQRpGwZnsBfTVSJ5E1Q2I3NoWGldxOaHKftDXT3p1Z56Cj3U9KxemPg
# 9ZSXt+zZR/hsPfMliLO8CsUEp458hUh2HGFGqhnEemKLwcI1qvtYb8VjC5NJMIEb
# e99/fE+0R21feByvtveWE1LvudFNOeVz3khOPBSqlw05zItR4VzRO/COZ+owYKlN
# Wp1DvdsjusAP10sQnZxN8FGihKrknKc91qPvChhIqPqxTqWYDku/8BTzAMiwSNZb
# /jjXiREtBbpDAk8iAJYlrX01boRoqyAYOCj+HKIQsaUCAwEAAaOCAWAwggFcMBMG
# A1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSJ/gox6ibN5m3HkZG5lIyiGGE3
# NDBRBgNVHREESjBIpEYwRDENMAsGA1UECxMETU9QUjEzMDEGA1UEBRMqMzE1OTUr
# MDQwNzkzNTAtMTZmYS00YzYwLWI2YmYtOWQyYjFjZDA1OTg0MB8GA1UdIwQYMBaA
# FMsR6MrStBZYAck3LjMWFrlMmgofMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY0NvZFNpZ1BDQV8w
# OC0zMS0yMDEwLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljQ29kU2lnUENBXzA4LTMx
# LTIwMTAuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCmqFOR3zsB/mFdBlrrZvAM2PfZ
# hNMAUQ4Q0aTRFyjnjDM4K9hDxgOLdeszkvSp4mf9AtulHU5DRV0bSePgTxbwfo/w
# iBHKgq2k+6apX/WXYMh7xL98m2ntH4LB8c2OeEti9dcNHNdTEtaWUu81vRmOoECT
# oQqlLRacwkZ0COvb9NilSTZUEhFVA7N7FvtH/vto/MBFXOI/Enkzou+Cxd5AGQfu
# FcUKm1kFQanQl56BngNb/ErjGi4FrFBHL4z6edgeIPgF+ylrGBT6cgS3C6eaZOwR
# XU9FSY0pGi370LYJU180lOAWxLnqczXoV+/h6xbDGMcGszvPYYTitkSJlKOGMIIF
# vDCCA6SgAwIBAgIKYTMmGgAAAAAAMTANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMTAwODMx
# MjIxOTMyWhcNMjAwODMxMjIyOTMyWjB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALJyWVwZMGS/HZpgICBC
# mXZTbD4b1m/My/Hqa/6XFhDg3zp0gxq3L6Ay7P/ewkJOI9VyANs1VwqJyq4gSfTw
# aKxNS42lvXlLcZtHB9r9Jd+ddYjPqnNEf9eB2/O98jakyVxF3K+tPeAoaJcap6Vy
# c1bxF5Tk/TWUcqDWdl8ed0WDhTgW0HNbBbpnUo2lsmkv2hkL/pJ0KeJ2L1TdFDBZ
# +NKNYv3LyV9GMVC5JxPkQDDPcikQKCLHN049oDI9kM2hOAaFXE5WgigqBTK3S9dP
# Y+fSLWLxRT3nrAgA9kahntFbjCZT6HqqSvJGzzc8OJ60d1ylF56NyxGPVjzBrAlf
# A9MCAwEAAaOCAV4wggFaMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFMsR6MrS
# tBZYAck3LjMWFrlMmgofMAsGA1UdDwQEAwIBhjASBgkrBgEEAYI3FQEEBQIDAQAB
# MCMGCSsGAQQBgjcVAgQWBBT90TFO0yaKleGYYDuoMW+mPLzYLTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTAfBgNVHSMEGDAWgBQOrIJgQFYnl+UlE/wq4QpTlVnk
# pDBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEE
# SDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2Nl
# cnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDANBgkqhkiG9w0BAQUFAAOCAgEAWTk+
# fyZGr+tvQLEytWrrDi9uqEn361917Uw7LddDrQv+y+ktMaMjzHxQmIAhXaw9L0y6
# oqhWnONwu7i0+Hm1SXL3PupBf8rhDBdpy6WcIC36C1DEVs0t40rSvHDnqA2iA6VW
# 4LiKS1fylUKc8fPv7uOGHzQ8uFaa8FMjhSqkghyT4pQHHfLiTviMocroE6WRTsgb
# 0o9ylSpxbZsa+BzwU9ZnzCL/XB3Nooy9J7J5Y1ZEolHN+emjWFbdmwJFRC9f9Nqu
# 1IIybvyklRPk62nnqaIsvsgrEA5ljpnb9aL6EiYJZTiU8XofSrvR4Vbo0HiWGFzJ
# NRZf3ZMdSY4tvq00RBzuEBUaAF3dNVshzpjHCe6FDoxPbQ4TTj18KUicctHzbMrB
# 7HCjV5JXfZSNoBtIA1r3z6NnCnSlNu0tLxfI5nI3EvRvsTxngvlSso0zFmUeDord
# EN5k9G/ORtTTF+l5xAS00/ss3x+KnqwK+xMnQK3k+eGpf0a7B2BHZWBATrBC7E7t
# s3Z52Ao0CW0cgDEf4g5U3eWh++VHEK1kmP9QFi58vwUheuKVQSdpw5OPlcmN2Jsh
# rg1cnPCiroZogwxqLbt2awAdlq3yFnv2FoMkuYjPaqhHMS+a3ONxPdcAfmJH0c6I
# ybgY+g5yjcGjPa8CQGr/aZuW4hCoELQ3UAjWwz0wggYHMIID76ADAgECAgphFmg0
# AAAAAAAcMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAX
# BgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290
# IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0wNzA0MDMxMjUzMDlaFw0yMTA0MDMx
# MzAzMDlaMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAf
# BgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQTCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJ+hbLHf20iSKnxrLhnhveLjxZlRI1Ctzt0YTiQP7tGn
# 0UytdDAgEesH1VSVFUmUG0KSrphcMCbaAGvoe73siQcP9w4EmPCJzB/LMySHnfL0
# Zxws/HvniB3q506jocEjU8qN+kXPCdBer9CwQgSi+aZsk2fXKNxGU7CG0OUoRi4n
# rIZPVVIM5AMs+2qQkDBuh/NZMJ36ftaXs+ghl3740hPzCLdTbVK0RZCfSABKR2YR
# JylmqJfk0waBSqL5hKcRRxQJgp+E7VV4/gGaHVAIhQAQMEbtt94jRrvELVSfrx54
# QTF3zJvfO4OToWECtR0Nsfz3m7IBziJLVP/5BcPCIAsCAwEAAaOCAaswggGnMA8G
# A1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFCM0+NlSRnAK7UD7dvuzK7DDNbMPMAsG
# A1UdDwQEAwIBhjAQBgkrBgEEAYI3FQEEAwIBADCBmAYDVR0jBIGQMIGNgBQOrIJg
# QFYnl+UlE/wq4QpTlVnkpKFjpGEwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5ghB5rRahSqClrUxzWPQHEy5lMFAGA1UdHwRJ
# MEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1
# Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRIMEYwRAYIKwYB
# BQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0Um9vdENlcnQuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEB
# BQUAA4ICAQAQl4rDXANENt3ptK132855UU0BsS50cVttDBOrzr57j7gu1BKijG1i
# uFcCy04gE1CZ3XpA4le7r1iaHOEdAYasu3jyi9DsOwHu4r6PCgXIjUji8FMV3U+r
# kuTnjWrVgMHmlPIGL4UD6ZEqJCJw+/b85HiZLg33B+JwvBhOnY5rCnKVuKE5nGct
# xVEO6mJcPxaYiyA/4gcaMvnMMUp2MT0rcgvI6nA9/4UKE9/CCmGO8Ne4F+tOi3/F
# NSteo7/rvH0LQnvUU3Ih7jDKu3hlXFsBFwoUDtLaFJj1PLlmWLMtL+f5hYbMUVbo
# nXCUbKw5TNT2eb+qGHpiKe+imyk0BncaYsk9Hm0fgvALxyy7z0Oz5fnsfbXjpKh0
# NbhOxXEjEiZ2CzxSjHFaRkMUvLOzsE1nyJ9C/4B5IYCeFTBm6EISXhrIniIh0EPp
# K+m79EjMLNTYMoBMJipIJF9a6lbvpt6Znco6b72BJ3QGEe52Ib+bgsEnVLaxaj2J
# oXZhtG6hE6a/qkfwEm/9ijJssv7fUciMI8lmvZ0dhxJkAj0tr1mPuOQh5bWwymO0
# eFQF1EEuUKyUsKV4q7OglnUa2ZKHE3UiLzKoCG6gW4wlv6DvhMoh1useT8ma7kng
# 9wFlb4kLfchpyOZu6qeXzjEp/w7FW1zYTRuh2Povnj8uVRZryROj/TGCBK8wggSr
# AgEBMIGQMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAh
# BgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBAhMzAAABCix5rtd5e6as
# AAEAAAEKMAkGBSsOAwIaBQCggcgwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFDra
# 5WsLtyseLjCPa2CAHNjcy8ZiMGgGCisGAQQBgjcCAQwxWjBYoDaANABNAGkAYwBy
# AG8AcwBvAGYAdAAgAEgAUABDACAAUABhAGMAawAgADIAMAAxADIAIABSADKhHoAc
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL0hQQzANBgkqhkiG9w0BAQEFAASCAQCG
# cLXNO1ofaV9BMAQXmWURIBHlJJAsej2c2IkXtbXZlTdmldAiV5xGcKuWSyoIjPb8
# I6H740hI9n3lTxvM9/FnaWhf7WTF+6TwmqIhssO8cKaoxK4swhQ0d7S/70JiLv1I
# BB0yjz77dJG/htIqAK9aEQpZzWD5DT5XrgV8FmxzinIZa9eHOWQ+D1GFucOtdLo5
# 9l5qeA/9J8bvq+rzdDjyI0uWgOSOGZwyrY7uzBnjaVP/SfJbI7U5HUbJ6rgbe4bt
# zDxTqiQ6ApEc5wZlAlThtUjNYIIbVnMPDoBShTMXTV/I8A3symlv9b+QZfSbfJXh
# pkIX5ffmndvaCuy+bZeBoYICKDCCAiQGCSqGSIb3DQEJBjGCAhUwggIRAgEBMIGO
# MHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAfBgNVBAMT
# GE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQQITMwAAAHJp9pDCFDtJoQAAAAAAcjAJ
# BgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkDMQsGCSqGSIb3DQEHATAcBgkqhkiG9w0B
# CQUxDxcNMTUwNzA2MTcyMzA4WjAjBgkqhkiG9w0BCQQxFgQUp/TbCrEnjAA2cBLA
# 8CuKkAoav8MwDQYJKoZIhvcNAQEFBQAEggEAcp2baRG4oDjMySFc1HSxFdpLBMKd
# uyhoZjehw8NM89yPgWa2Kdgwo7qJ9xEU33r7PQRSwdYiAfHWArnQu0hoH5wti+aZ
# wVv9N2wLlrB78WuH2ZkkTMqmzral4VZcCRufbekXqEksdYs2c/gVdFIRRyKci8TG
# AqNFe4Tdw9arqgq58yuchUdnAR/KBn6hONTVx9w3kkVUz5RRloXs52UlaMJlkVnT
# OHGJPVU7SKl1Zmsz4ovjoZIS54yhvshN0Z1zkH9rUvoEiiidEKZ+TuNwjnln/Vrp
# 4LHeClKEZ+6ZPwhnXPHa5SbmSDaEx39Jbysk3MHjtOoHW9QDFbM3STSNWA==
# SIG # End signature block
