﻿<#
.SYNOPSIS
  Check whether an exception is cloud exception. 
.DESCRIPTION
  The cloud exception class used to be "Microsoft.WindowsAzure.CloudException", from Azure PowerShell 0.8.14 on, it becomes "Hyak.Common.CloudException"
#>
function IsCloudException
{
    param
    (
        [Parameter(Mandatory=$true, Position=0)]
        $Exception
    )

    $exFullName = $Exception.GetType().FullName
    if($exFullName -eq "Microsoft.WindowsAzure.CloudException" -or $exFullName -eq "Hyak.Common.CloudException")
    {
        return $true      
    }

    return $false
}

<#
.SYNOPSIS
  Get the error code of cloud exception
.DESCRIPTION
  Get the error code of cloud exception. The cloud exception class used to be "Microsoft.WindowsAzure.CloudException", from Azure PowerShell 0.8.14 on, it becomes "Hyak.Common.CloudException". 
#>
function ErrorCodeOfCloudException
{
    param
    (
        [Parameter(Mandatory=$true, Position=0)]
        $Exception
    )

    $exFullName = $Exception.GetType().FullName
    if($exFullName -eq "Microsoft.WindowsAzure.CloudException")
    {
        return $Exception.ErrorCode
    }
    elseif($exFullName -eq "Hyak.Common.CloudException")
    {
        return $Exception.Error.Code
    }

    return ""
}

<#
.SYNOPSIS
  Invoke an Azure cmdlet with retries
.DESCRIPTION
  Invoke an Azure cmdlet with retries. The Azure cmdlet must be a "query" command (typically, Get-Azurexxx or Test-Azurexxx), must not modify any Azure resources. 
.EXAMPLE
  InvokeAzureCmdWithRetry -Command "Get-AzureVMImage"
  InvokeAzureCmdWithRetry -Command "Get-AzureDeployment -ServiceName testservice" -RetryTimes 10
#>
function InvokeAzureCmdWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true, Position=0)]
        [String] $Command,

        [Parameter(Mandatory=$false, Position=1)]
        [Int] $RetryTimes = 20
    )

    # if ErrorAction is not specified, specify it as Stop so that the error can be thrown by "Invoke-Expression" command.
    if(($Command -notmatch " -ErrorAction ") -and ($Command -notmatch " -EA "))
    {
        $newCommand = $Command + " -ErrorAction Stop"
    }
    else
    {
        $newCommand = $Command
    }
    $retry = 0
    while($true)
    {
        try
        {
            $ret = Invoke-Expression -Command $newCommand
            return $ret
        }
        catch
        {
            $caughtException = $_.Exception
            $errMsg = "Failed to execute the command '$Command': " + ($_ | Out-String)
            $shouldThrow = $true
            if(IsCloudException $caughtException)
            {
                $errCode = ErrorCodeOfCloudException $caughtException
                if($errCode -eq "ResourceNotFound")
                {
                    TraceDetail "The command '$Command' executed: the resource not found."
                    return
                }

                $shouldThrow = $false
            }
            elseif(($caughtException -is [System.Net.Http.HttpRequestException]) -or ($caughtException -is [System.Threading.Tasks.TaskCanceledException]))
            {
                $shouldThrow = $false
            }
            elseif($caughtException -is [System.NullReferenceException])
            {
                # NullReferenceException may be thrown in some race conditions
                $shouldThrow = $false
            }
        }

        TraceWarning $errMsg
        if($shouldThrow -or ($retry -ge $RetryTimes))
        {
            throw $caughtException
        }
        else
        {
            TraceInfo "Retry to execute the command '$Command' after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
    }
}

function New-AzureServiceWithRetry
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true, ParameterSetName="InAffinityGroup")]
        [String] $AffinityGroupName,

        [Parameter(Mandatory=$true, ParameterSetName="InRegion")]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            if($PsCmdlet.ParameterSetName -eq "InAffinityGroup")
            {
                TraceInfo "Creating the cloud service $ServiceName in the affinity group $AffinityGroupName"
                New-AzureService -ServiceName $ServiceName -AffinityGroup $AffinityGroupName -ErrorAction Stop | Out-Null
                TraceInfo "The cloud service $ServiceName created in the affinity group $AffinityGroupName"
            }
            else
            {
                TraceInfo "Creating the cloud service $ServiceName in the region $Location"
                New-AzureService -ServiceName $ServiceName -Location $Location -ErrorAction Stop | Out-Null
                TraceInfo "The cloud service $ServiceName created in the region $Location"
            }
            return
        }
        catch
        {
            TraceWarning ("Failed to create cloud service $ServiceName :" + ($_ | Out-String))
        }
        
        $service = InvokeAzureCmdWithRetry -Command "Get-AzureService -ServiceName $ServiceName"
        if($null -ne $service)
        {
            TraceInfo "Cloud service $ServiceName Created"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create cloud service $ServiceName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to create cloud service $ServiceName after $RetryTimes retries"
        }
  
    }
}

function New-AzureVMWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true, ValueFromPipeline = $true)]
        [Microsoft.WindowsAzure.Commands.ServiceManagement.Model.PersistentVM] $VM,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel,

        [Parameter(Mandatory=$false)]
        [Switch] $WaitForBoot,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    $vmname = $VM.RoleName
    while($true)
    {
        try
        {
            TraceInfo "Creating the virtual machine $vmname in cloud service $ServiceName"
            if($WaitForBoot.IsPresent)
            {
                New-AzureVM -VMs $VM -ServiceName $ServiceName -DeploymentLabel $DeploymentLabel -VNetName $VNetName -WaitForBoot -ErrorAction Stop -WarningAction SilentlyContinue | Out-Null
            }
            else
            {
                New-AzureVM -VMs $VM -ServiceName $ServiceName -DeploymentLabel $DeploymentLabel -VNetName $VNetName -ErrorAction Stop -WarningAction SilentlyContinue | Out-Null
            }

            TraceInfo "The virtual machine $vmname Created in cloud service $ServiceName"           
            return
        }
        catch
        {
            TraceWarning ("Failed to create the virtual machine $vmname : " + ($_ | Out-String))
        }
        
        $newvm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $vmname"
        if($null -ne $newvm)
        {
            TraceInfo "The virtual machine $vmname Created in cloud service $ServiceName"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the virtual machine $vmname after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to create the virtual machine $vmname after $RetryTimes retries"
        }
    }
}

function Update-AzureVMWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true, ValueFromPipeline = $true)]
        [Microsoft.WindowsAzure.Commands.ServiceManagement.Model.PersistentVM] $VM,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Updating the virtual machine $Name in cloud service $ServiceName ..."
            Update-AzureVM -Name $Name -ServiceName $ServiceName -VM $VM -ErrorAction Stop -WarningAction SilentlyContinue | Out-Null
            TraceInfo "The virtual machine $Name updated."           
            return
        }
        catch
        {
            TraceWarning ("Failed to update the virtual machine $Name : " + ($_ | Out-String))
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to update the virtual machine $Name after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to update the virtual machine $Name after $RetryTimes retries"
        }
    }
}

function New-AzureAffinityGroupWithRetry
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Creating the affinity group $Name at region $Location"
            New-AzureAffinityGroup -Name $Name -Location $Location -Label $Name -ErrorAction Stop | Out-Null
            TraceInfo "The affinity group $Name created at region $Location"
            return
        }
        catch
        {
            TraceWarning ("Failed to create affinity group $Name : " + ($_ | Out-String))
        }
        
        $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $Name"
        if($null -ne $ag)
        {
            TraceInfo "The affinity group $Name Created"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the affinity group $Name after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to create the affinity group $Name after $RetryTimes retries"
        }
    }
}

function New-AzureStorageAccountWithRetry
{
    Param
    (
        # The storage account Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName,

        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Creating the storage account $StorageAccountName at region $Location"
            New-AzureStorageAccount -StorageAccountName $StorageAccountName -Location $Location -ErrorAction Stop | Out-Null
            TraceInfo "Storage account $StorageAccountName created at region $Location"
            return
        }
        catch
        {
            TraceWarning ("Failed to create the storage account $StorageAccountName : " + ($_ | Out-String))
        }
        
        $storage = InvokeAzureCmdWithRetry -Command "Get-AzureStorageAccount -StorageAccountName $StorageAccountName"
        if($null -ne $storage)
        {
            TraceInfo "Storage account $StorageAccountName Created"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the storage account $StorageAccountName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++  
        }
        else
        {
            throw "Failed to create the storage account $StorageAccountName  after $RetryTimes retries"
        }
    }
}

function Save-AzureVMImageWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $ImageName,
        
        [Parameter(Mandatory=$false)]
        [String] $ImageLabel = "HPC Image",
        
        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Saving Azure VM Image $ImageName"
            Save-AzureVMImage -Name $Name -ServiceName $ServiceName -ImageName $ImageName -ImageLabel $ImageLabel -OSState Generalized -ErrorAction Stop | Out-Null
            TraceInfo "The VM image $ImageName was saved"
            return
        }
        catch
        {
            TraceWarning ("Failed to save the VM image $ImageName :" + ($_ | Out-String))
        }

        $image = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $ImageName"
        if($null -ne $image)
        {
            TraceInfo "The VM image $ImageName already saved"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to save the VM image $ImageName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to save the VM image $ImageName after $RetryTimes retries"
        }
    }
}

function Add-AzureVMImageWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$true)]
        [String] $MediaLocation,

        [Parameter(Mandatory=$true)]
        [ValidateSet("Linux","Windows")]
        [String] $OS,
        
        [Parameter(Mandatory=$false)]
        [String] $Label,

        [Parameter(Mandatory=$false)]
        [String] $Description,

        [Parameter(Mandatory=$false)]
        [System.DateTime] $PublishedDate,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    if([String]::IsNullOrEmpty($Label))
    {
        $Label = $ImageName
    }
    if([String]::IsNullOrEmpty($Description))
    {
        $Description = $Label
    }
    
    if($null -eq $PublishedDate)
    {
        $PublishedDate = Get-Date
    }

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Adding Azure VM Image $ImageName"
            Add-AzureVMImage -ImageName $ImageName -MediaLocation $MediaLocation -OS $OS -Label $Label -Description $Description -PublishedDate $PublishedDate -ErrorAction Stop | Out-Null
            TraceInfo "The VM image $ImageName was successfully added"
            return
        }
        catch
        {
            TraceWarning ("Failed to add the VM image $ImageName :" + ($_ | Out-String))
        }

        $image = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $ImageName"
        if($null -ne $image)
        {
            TraceInfo "The VM image $ImageName was already added"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to add the VM image $ImageName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to add the VM image $ImageName after $RetryTimes retries"
        }
    }
}

function Remove-AzureVMImageWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$false)]
        [Switch] $DeleteVHD,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Removing Azure VM Image $ImageName"
            Remove-AzureVMImage -ImageName $ImageName -DeleteVHD:$DeleteVHD -ErrorAction Stop | Out-Null
            TraceInfo "The VM image $ImageName was successfully removed"
            return
        }
        catch
        {
            if(IsCloudException $_.Exception)
            {
                $errCode = ErrorCodeOfCloudException $_.Exception
                if($errCode -eq "ResourceNotFound")
                {
                    TraceInfo "The VM image $ImageName was already removed"
                    return
                }
            }

            TraceWarning ("Failed to remove the VM image $ImageName :" + ($_ | Out-String))
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove the VM image $ImageName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to remove the VM image $ImageName after $RetryTimes retries"
        }
    }
}

function Get-AzureVNetSiteWithRetry
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName
    )
    
    $vnetSite = $null
    try
    {
        #The VNetName may contain space
        $vnetSite = InvokeAzureCmdWithRetry -Command "Get-AzureVNetSite -VNetName '$VNetName'"
    }
    catch
    {
    }

    return $vnetSite
}

   
function Remove-AzureVMWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 3
    )

    $retry = 0
    TraceInfo "Trying to remove the virtual machine $VMName from the cloud service $ServiceName"
    while($true)
    {
        try
        {
            TraceInfo "Removing the virtual machine $VMName from the cloud service $ServiceName"
            Remove-AzureVM -Name $VMName -ServiceName $ServiceName -DeleteVHD -ErrorAction Stop -WarningAction SilentlyContinue
            TraceInfo "The virtual machine $VMName removed from the cloud service $ServiceName"
            return
        }
        catch
        {
            TraceWarning ("Failed to remove the virtual machine $VMName : " + ($_ | Out-String))
        }

        $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $VMName -WarningAction SilentlyContinue"
        if($null -eq $vm)
        {
            TraceInfo "The virtual machine $VMName already removed from the cloud service $ServiceName"
            return
        }
        
        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove the virtual machine $VMName after 5 seconds"
            Start-Sleep -Seconds 5
            $retry++
        }
        else
        {
            throw "Failed to remove the virtual machine $VMName after $RetryTimes retries"
        }
    }
}

function Remove-AzureServiceWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 3
    )

    $retry = 0
    TraceInfo "Trying to remove the cloud service $ServiceName"
    while($true)
    {
        try
        {
            TraceInfo "Removing the cloud service $ServiceName"
            Remove-AzureService -ServiceName $ServiceName -DeleteAll -Force -ErrorAction Stop -WarningAction SilentlyContinue
            TraceInfo "The cloud service $ServiceName removed"
            return
        }
        catch
        {
            TraceWarning ("Failed to remove the cloud service $ServiceName : " + ($_ | Out-String))
        }

        $svc = InvokeAzureCmdWithRetry -Command "Get-AzureService -ServiceName $ServiceName"
        if($null -eq $svc)
        {
            TraceInfo "The cloud service $ServiceName already removed"
            return
        }
                
        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove the cloud service $ServiceName after 5 seconds"
            Start-Sleep -Seconds 5
            $retry++
        }
        else
        {
            throw "Failed to remove the cloud service $ServiceName after $RetryTimes retries"
        }
    }
}

function GetStorageAccountFromMediaLink
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $MediaLink
    )

    if(-not [string]::IsNullOrEmpty($MediaLink))
    {
        $startIndex = $MediaLink.IndexOf("//") + 2
        $endIndex = $MediaLink.IndexOf(".")
        if(($endIndex -gt ($startIndex + 2)) -and ($startIndex -gt 0))
        {
            return $MediaLink.Substring($startIndex, $endIndex - $startIndex)
        }
    }

    return ""
}


function DecodeLabelIfBase64Encoded
{
    param($orgLabel)

    try
    {
        $bytes = [System.Convert]::FromBase64String($orgLabel)
        return [System.Text.Encoding]::UTF8.GetString($bytes)
    }
    catch
    {
        return $orgLabel
    }
}


<#
.SYNOPSIS
  Check whether a VM exists in the given cloud service
.DESCRIPTION
  Check whether a VM exists in the given cloud service
.EXAMPLE
  VMExists -Name "vmname" -ServiceName "servicename"
#>
function VMExists 
{
    param
    (
        #The name of VM
        [Parameter(Mandatory=$true)]
        [String] $Name, 

        #The name of the host cloud service
        [Parameter(Mandatory=$true)]
        [String] $ServiceName
    )

    $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -Name $Name -ServiceName $ServiceName"
    if ($null -ne $vm) 
    {
        return $true
    }
    
    return $false
}

<#
.SYNOPSIS
  Wait the specified VMs in a cloud service to become the target status
.DESCRIPTION
  Wait the specified VMs to become the target status in a given maximum waiting time, throw if the timer expires
.EXAMPLE
  The following example will wait the VM "mytestvm" in cloud service "mytestservice" to become 'ReadyRole' for a maximum time 900 seconds
    WaitForVMReady -VMName @("testvm1", "testvm2") -ServiceName "mytestservice" -MaxWaitSeconds 900
#>
function WaitForVMReady 
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$false)]
        [String] $TargetStatus = "ReadyRole", 

        [Parameter(Mandatory=$false)]
        [ValidateRange(0, 3600)]
        [int] $MaxWaitSeconds = 1800
    )

    TraceInfo "Waiting for the following VM(s) to become $TargetStatus status: $VMName"
    $noReadyVMNames = $VMName
    $beginTime = Get-Date
    while ($noReadyVMNames.Count -gt 0)
    {
        $vms = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -WarningAction SilentlyContinue" | ?{($noReadyVMNames -contains $_.InstanceName) -and ($_.InstanceStatus -eq $TargetStatus)})
        if($vms.Count -gt 0)
        {
            TraceInfo "The following VM(s) are now in $TargetStatus status: $($vms.InstanceName)"
            if($vms.Count -eq $noReadyVMNames.Count)
            {
                break
            }

            $noReadyVMNames = @($noReadyVMNames | ?{$vms.InstanceName -notcontains $_})
        }

        $timeSpan = (Get-Date) - $beginTime
        $elapsedSeconds = [int]$timeSpan.TotalSeconds
        if ($elapsedSeconds -ge $MaxWaitSeconds)
        {
            throw "Timed out waiting for the following virtual machine(s) to become $TargetStatus : $noReadyVMNames"
        }

        Start-Sleep -Seconds 10
    }
}


function WaitForVMCreated 
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$false)]
        [ValidateRange(0, 3600)]
        [int] $MaxWaitSeconds = 1800
    )

    TraceInfo "Waiting for the following VM(s) created: $VMName"
    $beginTime = Get-Date
    $CreatedVMNames = @()
    while ($true)
    {
        $lastCreatedNum = $CreatedVMNames.Count
        $CreatedVMNames = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -WarningAction SilentlyContinue" | %{$_.InstanceName} | ?{$VMName -contains $_})
        if($CreatedVMNames.Count -ne $lastCreatedNum)
        {
            TraceInfo "The following VM(s) are now created: $CreatedVMNames"
            if($CreatedVMNames.Count -eq $VMName.Count)
            {
                break
            }
        }
        
        $timeSpan = (Get-Date) - $beginTime
        $elapsedSeconds = [int]$timeSpan.TotalSeconds
        if ($elapsedSeconds -ge $MaxWaitSeconds)
        {
            $notCreatedVMNames = @($VMName | ?{$CreatedVMNames -notcontains $_})
            throw "Timed out waiting for the following virtual machine(s) created: $notCreatedVMNames"
        }

        Start-Sleep -Seconds 10
    }
}

function GetVMSizesWithRDMA
{
    return @("A8", "A9")
}

function IsVMSizeWithRDMA
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMSize
    )
    
    $sizes = @(GetVMSizesWithRDMA)
    return ($sizes -contains $VMSize)
}

function NewVMExtensionConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ExtensionName, 

        [Parameter(Mandatory=$true)]
        [String] $Publisher, 

        [Parameter(Mandatory=$true)]
        [String] $Version, 
        
        [Parameter(Mandatory=$false)]
        [String] $PublicConfiguration = $null, 
        
        [Parameter(Mandatory=$false)]
        [String] $PrivateConfiguration = $null
    )

    $extConfig = @{}
    $extConfig['ExtensionName'] = $ExtensionName
    $extConfig['Publisher'] = $Publisher
    $extConfig['Version'] = $Version
    $extConfig['PublicConfiguration'] = $PublicConfiguration
    $extConfig['PrivateConfiguration'] = $PrivateConfiguration
    $ret = New-Object psobject -Property $extConfig
    return $ret
}

function SetAzureVMExtensions
{
    Param
    (
        [Parameter(Mandatory=$true, ValueFromPipeline = $true)]
        [Microsoft.WindowsAzure.Commands.ServiceManagement.Model.PersistentVM] $VM,

        [Parameter(Mandatory=$true)]
        [Object[]] $VMExtensions,

        [Parameter(Mandatory=$false)]
        [Switch] $SkipIfExists
    )

    TraceInfo "SetAzureVMExtensions: Setting VM extensions for the virtual machine $($VM.RoleName)"
    $tmpVM = $VM
    $curExtensions = @($tmpVM | Get-AzureVMExtension)
    foreach($ext in $VMExtensions)
    {
        $dupExt = $curExtensions | ?{($_.ExtensionName -eq $ext.ExtensionName) -and ($_.Publisher -eq $ext.Publisher)} | select -First(1)
        
        if($null -ne $dupExt)
        {
            if($SkipIfExists)
            {
                TraceInfo "Skip to set VM extension $($dupExt.Publisher).$($dupExt.ExtensionName) because it already exists"
                continue
            }
            else
            {
                TraceInfo "Remove the existing VM extension $($dupExt.Publisher).$($dupExt.ExtensionName) with version $($dupExt.Version)"
                $tmpVM = $tmpVM | Remove-AzureVMExtension -ExtensionName $dupExt.ExtensionName -Publisher $dupExt.Publisher
            }
        }

        TraceInfo "Set VM extension $($ext.Publisher).$($ext.ExtensionName) with version $($ext.Version)"
        if([string]::IsNullOrEmpty($ext.PublicConfiguration))
        {
            if([string]::IsNullOrEmpty($ext.PrivateConfiguration))
            {
                $tmpVM = $tmpVM | `
                    Set-AzureVMExtension -ExtensionName $ext.ExtensionName -Publisher $ext.Publisher -Version $ext.Version
            }
            else
            {
                $tmpVM = $tmpVM | `
                    Set-AzureVMExtension -ExtensionName $ext.ExtensionName -Publisher $ext.Publisher -Version $ext.Version `
                        -PrivateConfiguration $ext.PrivateConfiguration
            }
        }
        else
        {
            if([string]::IsNullOrEmpty($ext.PrivateConfiguration))
            {
                $tmpVM = $tmpVM | `
                    Set-AzureVMExtension -ExtensionName $ext.ExtensionName -Publisher $ext.Publisher -Version $ext.Version `
                        -PublicConfiguration $ext.PublicConfiguration
            }
            else
            {
                $tmpVM = $tmpVM | `
                    Set-AzureVMExtension -ExtensionName $ext.ExtensionName -Publisher $ext.Publisher -Version $ext.Version `
                        -PublicConfiguration $ext.PublicConfiguration -PrivateConfiguration $ext.PrivateConfiguration
            }
        }
    }

    return $tmpVM
}

function NewVMConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName, 
        
        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 

        [Parameter(Mandatory=$true, ParameterSetName="WindowsDomainVM")]
        [PSCredential] $DomainUserCred,

        [Parameter(Mandatory=$true, ParameterSetName="WindowsDomainVM")]
        [String] $DomainFQDN,

        [Parameter(Mandatory=$true, ParameterSetName="WindowsDomainVM")]
        [Switch] $WindowsDomain,

        [Parameter(Mandatory=$true, ParameterSetName="LinuxVM")]
        [Switch] $Linux,

        [Parameter(Mandatory=$true, ParameterSetName="WindowsVM")]
        [Switch] $Windows,

        [Parameter(Mandatory=$false, ParameterSetName="LinuxVM")]
        [Object[]] $SSHKeyPairs = @(),

        [Parameter(Mandatory=$false, ParameterSetName="LinuxVM")]
        [Object[]] $SSHPublicKeys = @(),

        [Parameter(Mandatory=$false)]
        [ValidateSet("ReadOnly","ReadWrite")]
        [String] $HostCaching = "",

        [Parameter(Mandatory=$false)]
        [String] $CustomDataFile = "",

        [Parameter(Mandatory=$false)]
        [Switch] $DisableGuestAgent,

        [Parameter(Mandatory=$false)]
        [Object[]] $VMExtensions = @()
    )

    if([String]::IsNullOrEmpty($HostCaching))
    {
        $vm = New-AzureVMConfig -Name $VMName -InstanceSize $VMSize -ImageName $ImageName
    }
    else
    {
        $vm = New-AzureVMConfig -Name $VMName -InstanceSize $VMSize -HostCaching $HostCaching -ImageName $ImageName
    }

    if($Linux.IsPresent)
    {
        if ([String]::IsNullOrEmpty($CustomDataFile))
        {
            if(($null -eq $SSHKeyPairs) -or $SSHKeyPairs.Count -eq 0)
            {
                if(($null -eq $SSHPublicKeys) -or $SSHPublicKeys.Count -eq 0)
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -DisableGuestAgent:$DisableGuestAgent
                }
                else
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -DisableGuestAgent:$DisableGuestAgent -SSHPublicKeys $SSHPublicKeys
                }
            }
            else
            {
                if(($null -eq $SSHPublicKeys) -or $SSHPublicKeys.Count -eq 0)
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -DisableGuestAgent:$DisableGuestAgent -SSHKeyPairs $SSHKeyPairs
                }
                else
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -DisableGuestAgent:$DisableGuestAgent -SSHKeyPairs $SSHKeyPairs -SSHPublicKeys $SSHPublicKeys
                }
            }
        }
        else
        {
            if(($null -eq $SSHKeyPairs) -or $SSHKeyPairs.Count -eq 0)
            {
                if(($null -eq $SSHPublicKeys) -or $SSHPublicKeys.Count -eq 0)
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -CustomDataFile $CustomDataFile -DisableGuestAgent:$DisableGuestAgent
                }
                else
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -CustomDataFile $CustomDataFile -DisableGuestAgent:$DisableGuestAgent -SSHPublicKeys $SSHPublicKeys
                }
            }
            else
            {
                if(($null -eq $SSHPublicKeys) -or $SSHPublicKeys.Count -eq 0)
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -CustomDataFile $CustomDataFile -DisableGuestAgent:$DisableGuestAgent -SSHKeyPairs $SSHKeyPairs
                }
                else
                {
                    $vm = $vm | `
                          Add-AzureProvisioningConfig -Linux -LinuxUser $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                              -CustomDataFile $CustomDataFile -DisableGuestAgent:$DisableGuestAgent -SSHKeyPairs $SSHKeyPairs -SSHPublicKeys $SSHPublicKeys
                }
            }
        }
    }
    elseif($Windows.IsPresent)
    {
        if ([String]::IsNullOrEmpty($CustomDataFile))
        {
            $vm = $vm | `
                  Add-AzureProvisioningConfig -Windows -AdminUsername $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                      -DisableGuestAgent:$DisableGuestAgent
        }
        else
        {
            $vm = $vm | `
                  Add-AzureProvisioningConfig -Windows -AdminUsername $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                      -CustomDataFile $CustomDataFile -DisableGuestAgent:$DisableGuestAgent
        }
    }
    elseif($WindowsDomain.IsPresent)
    {
        $netbios = $DomainFQDN.Split(".")[0].ToUpper()
        $domainUserName = $DomainUserCred.UserName
        if($domainUserName.Contains("\"))
        {
            $domainUserName = $domainUserName.Split("\")[1]
        }
        if ([String]::IsNullOrEmpty($CustomDataFile))
        {
            $vm = $vm | `
                  Add-AzureProvisioningConfig -AdminUsername $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                      -Domain $netbios -DomainUserName $domainUserName -DomainPassword $DomainUserCred.GetNetworkCredential().Password `
                      -JoinDomain $DomainFQDN -WindowsDomain -DisableGuestAgent:$DisableGuestAgent
        }
        else
        {
            $vm = $vm | `
                  Add-AzureProvisioningConfig -AdminUsername $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                      -Domain $netbios -DomainUserName $domainUserName -DomainPassword $DomainUserCred.GetNetworkCredential().Password `
                      -JoinDomain $DomainFQDN -WindowsDomain -CustomDataFile $CustomDataFile -DisableGuestAgent:$DisableGuestAgent
        }
    }
    else
    {
        throw "You must specify one of 'Linux', 'Windows' and 'WindowsDomain' for VM $VMName."
    }

    $vm = $vm | Set-AzureSubnet -SubnetNames $SubnetName

    if(-not $DisableGuestAgent.IsPresent)
    {
        if(($null -ne $VMExtensions) -and ($VMExtensions.Count -gt 0))
        {
            $vm = $vm | SetAzureVMExtensions -VMExtensions $VMExtensions -SkipIfExists
        }
    }

    return $vm
}

function CreateAzureVM
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true, ValueFromPipeline = $true)]
        [Microsoft.WindowsAzure.Commands.ServiceManagement.Model.PersistentVM] $VM,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel,

        [Parameter(Mandatory=$false)]
        [Switch] $Wait,

        [Parameter(Mandatory=$false)]
        [Switch] $UseStaticInternalIP
    )

    $VMName = $VM.RoleName
    try
    {
        New-AzureVMWithRetry -VM $VM -ServiceName $ServiceName -VNetName $VNetName -DeploymentLabel $DeploymentLabel
    }
    catch
    {
        TraceError "Failed to create a VM $VMName in cloud service $ServiceName"
        throw
    }

    if($UseStaticInternalIP.IsPresent)
    {
        try
        {
            $createdVM = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $VMName"
            TraceError "Try to set static internal IP address $($createdVM.IpAddress) for the VM $VMName"
            $createdVM =  $createdVM | Set-AzureStaticVNetIP -IPAddress $createdVM.IpAddress
            Update-AzureVMWithRetry -Name $VMName -ServiceName $ServiceName -VM $createdVM.VM
        }
        catch
        {
            TraceError "Failed to set static internal IP address for the VM $VMName."
            throw
        }
    }

    if ($Wait.IsPresent)
    {
        WaitForVMReady -VMName $VMName -ServiceName $ServiceName
    }
}

function CreateDomainJoinedVM
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 
        
        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred,

        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 

        [Parameter(Mandatory=$false)]
        [ValidateSet("ReadOnly","ReadWrite")]
        [String] $HostCaching = "",

        [Parameter(Mandatory=$false)]
        [String] $CustomDataFile = "",

        [Parameter(Mandatory=$false)]
        [Object[]] $VMExtensions = @(),

        [Parameter(Mandatory=$false)]
        [Switch] $DisableGuestAgent,

        [Parameter(Mandatory=$false)]
        [Switch] $Wait,

        [Parameter(Mandatory=$false)]
        [Switch] $UseStaticInternalIP
    )

    TraceInfo "Creating a domain joined VM $VMName in Service $ServiceName `n Size:  `t$VMSize`n Image: `t$ImageName`n VNet:  `t$VNetName`n Subnet:`t$SubnetName`n Domain:`t$DomainFQDN"
    if([String]::IsNullOrEmpty($HostCaching))
    {
        $vm = NewVMConfig -WindowsDomain -VMName $VMName -VMSize $VMSize -ImageName $ImageName -SubnetName $SubnetName `
                  -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN -CustomDataFile $CustomDataFile `
                  -DisableGuestAgent:$DisableGuestAgent -VMExtensions $VMExtensions
    }
    else
    {
        $vm = NewVMConfig -WindowsDomain -VMName $VMName -VMSize $VMSize -ImageName $ImageName -SubnetName $SubnetName `
                  -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN -CustomDataFile $CustomDataFile `
                  -DisableGuestAgent:$DisableGuestAgent -VMExtensions $VMExtensions -HostCaching $HostCaching
    }

    if(IsVMSizeWithRDMA -VMSize $VMSize)
    {
        if($DisableGuestAgent.IsPresent)
        {
            TraceWarning "The RDMA network for $VMName cannot work because Azure VM extension 'HpcVmDrivers' is not enabled."
        }
        else
        {
            TraceInfo "CreateDomainJoinedVM: Set VM extension Microsoft.HpcCompute.HpcVmDrivers with version 1.* for the virtual machine $VMName."
            $ext = NewVMExtensionConfig -ExtensionName "HpcVmDrivers" -Publisher "Microsoft.HpcCompute" -Version "1.*"
            $vm = $vm | SetAzureVMExtensions -VMExtensions $ext -SkipIfExists
        }
    }

    CreateAzureVM -ServiceName $ServiceName -VM $vm -VNetName $VNetName -DeploymentLabel $DeploymentLabel -Wait:$Wait -UseStaticInternalIP:$UseStaticInternalIP
}

function ConvertSecureStrToPlain 
{
    param
    ( 
        [Parameter(Mandatory=$true)]
        [System.Security.SecureString] $SecurePassword
    )
    
    $pswPointer = [Runtime.InteropServices.Marshal]::SecureStringToBSTR($SecurePassword)
    $plainPsw = [Runtime.InteropServices.Marshal]::PtrToStringAuto($pswPointer)
    [Runtime.InteropServices.Marshal]::ZeroFreeBSTR($pswPointer)
    return $plainPsw
}

function GetAzureSubscription
{
    [CmdletBinding(DefaultParametersetName="GetByName")]
    param
    (
        [Parameter(Mandatory=$false, ParameterSetName="GetByName", Position=0)]
        [string] $SubscriptionName = "",

        [Parameter(Mandatory=$false, ParameterSetName="GetById", Position=0)]
        [string] $SubscriptionId = "",

        [Parameter(Mandatory=$false, ParameterSetName="GetByCurrent", Position=0)]
        [Switch] $Current
    )

    $retry = 0
    while($true)
    {
        try
        {
            if($Current.IsPresent)
            {
                $sub = Get-AzureSubscription -Current
            }
            elseif(-not [String]::IsNullOrEmpty($SubscriptionName))
            {
                $sub = InvokeAzureCmdWithRetry -Command "Get-AzureSubscription -SubscriptionName $subscriptionName"
            }
            elseif(-not [String]::IsNullOrEmpty($SubscriptionId))
            {
                $sub = InvokeAzureCmdWithRetry -Command "Get-AzureSubscription -SubscriptionId $SubscriptionId"
            }
            else
            {
                $sub = InvokeAzureCmdWithRetry -Command "Get-AzureSubscription"
            }
            
            return $sub
        }
        catch 
        {
            # NullReferenceException or IOException may be thrown due to access race condition to $env:APPDATA\Windows Azure Powershell\AzureProfile.json
            # NullReferenceException already handled in InvokeAzureCmdWithRetry, so we just handle IOException here
            # Simply retry several times in this case.
            if($_.Exception -is [System.IO.IOException] -or $_.Exception -is [System.NullReferenceException])
            {
                TraceWarning ("Failed to get subscription: " + ($_.Exception | Out-String))
                if($retry -lt 10)
                {
                    Start-Sleep -Seconds (Get-Random -Minimum 2 -Maximum 10)
                    $retry++
                    continue
                }
            }

            return
        }
    }
}

function SelectAzureSubscription
{
    [CmdletBinding(DefaultParametersetName="SelectByName")]
    param
    (
        [Parameter(Mandatory=$true, ParameterSetName="SelectByName", Position=0)]
        [string] $SubscriptionName,

        [Parameter(Mandatory=$true, ParameterSetName="SelectById", Position=0)]
        [string] $SubscriptionId
    )

    $retry = 0
    while($true)
    {
        try
        {
            if($PsCmdlet.ParameterSetName -eq "SelectByName")
            {
                Select-AzureSubscription -SubscriptionName $SubscriptionName -ErrorAction Stop | Out-Null
            }
            else
            {
                Select-AzureSubscription -SubscriptionId $SubscriptionId -ErrorAction Stop | Out-Null
            }

            break
        }
        catch 
        {
            # NullReferenceException or IOException may be thrown due to access race condition to $env:APPDATA\Windows Azure Powershell\AzureProfile.json
            # Simply retry several times in this case.
            if($_.Exception -is [System.NullReferenceException] -or $_.Exception -is [System.IO.IOException])
            {
                if($retry -lt 10)
                {
                    Start-Sleep -Seconds (Get-Random -Minimum 2 -Maximum 10)
                    $retry++
                    continue
                }
            }

            throw
        }
    }
}

function Get-AzureStorageContainerWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        $Context,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $retry = 0
    while($true)
    {
        try
        {
            $container = Get-AzureStorageContainer -Name $Name -Context $Context -ErrorAction Stop
            return $container
        }
        catch
        {
            if(IsCloudException $_.Exception)
            {
                $errCode = ErrorCodeOfCloudException $_.Exception
                if($errCode -eq "ResourceNotFound")
                {
                    return
                }
            }
            elseif($_.Exception -is [Microsoft.WindowsAzure.Commands.Storage.Common.ResourceNotFoundException])
            {
                return
            }

            if($retry -lt $RetryTimes)
            {
                TraceInfo ("Failed to get the storage container $Name :" + ($_ | Out-String))
                Start-Sleep -Seconds 10
                $retry++
            }
            else
            {
                TraceError ("Failed to get the storage container $Name :" + ($_ | Out-String))
                throw
            }
        }
    }
}

function New-AzureStorageContainerWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        $Context,

        [Parameter(Mandatory=$false)]
        $Permission = "Off",

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $stgName = $Context.StorageAccountName
    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Creating the container $Name in the storage account $stgName"
            $container = New-AzureStorageContainer -Name $Name -Context $Context -Permission $Permission -ErrorAction Stop
            TraceInfo "The container $Name was created in the storage account $stgName"
            return $container
        }
        catch
        {
            TraceWarning ("Failed to create container $Name in the storage account $stgName : " + ($_ | Out-String))
        }

        $container = Get-AzureStorageContainerWithRetry -Name $Name -Context $Context
        if($null -ne $container)
        {
            TraceInfo "The container $Name was already created in the storage account $stgName"
            return $container
        }
                
        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the container $Name in the storage account $stgName after 5 seconds"
            Start-Sleep -Seconds 5
            $retry++
        }
        else
        {
            throw "Failed to create the container $Name in the storage account $stgName after $RetryTimes retries"
        }
    }
}

function Start-AzureStorageBlobCopyWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true, ParameterSetName="BlobUri")]
        [alias("SrcUri")]
        [String] $AbsoluteUri,

        [Parameter(Mandatory=$true, ParameterSetName="BlobName")]
        [String] $SrcBlob,

        [Parameter(Mandatory=$true, ParameterSetName="BlobName")]
        [String] $SrcContainer,

        [Parameter(Mandatory=$true)]
        [String] $DestBlob,

        [Parameter(Mandatory=$true)]
        [String] $DestContainer,

        [Parameter(Mandatory=$true)]
        $Context,

        [Parameter(Mandatory=$true)]
        $DestContext,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $srcStgName = $Context.StorageAccountName
    $destStgName = $DestContext.StorageAccountName
    $retry = 0
    while($true)
    {
        try
        {
            if($PsCmdlet.ParameterSetName -eq "BlobUri")
            {            
                TraceInfo "Queuing the blob copy from $AbsoluteUri to [$destStgName, $DestContainer, $DestBlob]"
                $blobCopy = Start-AzureStorageBlobCopy -AbsoluteUri $AbsoluteUri -DestContainer $DestContainer -DestBlob $DestBlob `
                        -Context $Context -DestContext $DestContext -Force -ErrorAction Stop
                TraceInfo "The blob copy from $AbsoluteUri to [$destStgName, $DestContainer, $DestBlob] was successfully queued"
                return $blobCopy
            }
            else
            {
                TraceInfo "Queuing the blob copy from [$srcStgName, $SrcContainer, $SrcBlob] to [$destStgName, $DestContainer, $DestBlob]"
                $blobCopy = Start-AzureStorageBlobCopy -SrcContainer $SrcContainer -SrcBlob $SrcBlob -DestContainer $DestContainer -DestBlob $DestBlob `
                        -Context $Context -DestContext $DestContext -Force -ErrorAction Stop
                TraceInfo "The blob copy from [$srcStgName, $SrcContainer, $SrcBlob] to [$destStgName, $DestContainer, $DestBlob] was successfully queued"
                return $blobCopy                
            }
        }
        catch
        {
            $noRetry = $false
            if($_.Exception -is [Microsoft.WindowsAzure.Storage.StorageException])
            {
                $noRetry = $true
            }

            
            if((-not $noRetry) -and ($retry -lt $RetryTimes))
            {
                TraceWarning ("Failed to queue the blob copy to [$destStgName, $DestContainer, $DestBlob]: " + ($_ | Out-String))
                Start-Sleep -Seconds 5
                $retry++
            }
            else
            {
                TraceError ("Failed to queue the blob copy to [$destStgName, $DestContainer, $DestBlob]: " + ($_ | Out-String))
                throw
            }
        }
    }
}

# SIG # Begin signature block
# MIIaxwYJKoZIhvcNAQcCoIIauDCCGrQCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUJ0pR0EOxvK3TlzVRutHTIZHX
# GEugghWCMIIEwzCCA6ugAwIBAgITMwAAAHJp9pDCFDtJoQAAAAAAcjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUwMzIwMTczMjA0
# WhcNMTYwNjIwMTczMjA0WjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OjMxQzUtMzBCQS03QzkxMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoxPVJ2iBlQVE
# sJuqKGMDs9En3ra2ZyP2mieV2SfpWI7AXCdUfRHZ7B+712Ff5z/jWtu5iNakH+yw
# OsPZy911xkYMU2T9mm1IXZcHoKEMvj6SL5H3q2dq0SX5n3iqXJcxf4wwjDu9cKit
# 5v8t+rrQxeVIuseUK34MXOH/NYpAv1H0FU0vm0MjOh7ltNRzua86yAgrQBSj7gML
# NybNLvq+M9b3JBVF6cFLmII8syS9LSf7drlwWzPRJ71Q/6+TVq59UtTQO217Juvk
# 9BdtStzFmMyR+7VdVxybCwxPr6dW/2U7oouf4UxtszHDMtHw7XEnA2jg2iq6ODkZ
# frNXUQCMhQIDAQABo4IBCTCCAQUwHQYDVR0OBBYEFND9obr0IhS4jv6FL059LwBC
# pwFTMB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZI
# hvcNAQEFBQADggEBACJEd3U6Z3Bw2w+djrTKkrwYGl8Ys2vdoryZDDm9ybp6jjCk
# QCKLBGssgmRVJcEaYjUTkqGomLk5Mt4r5X6S+Zf1lgF3IKX9tqVbNlX1qMmRZUQ4
# 4Xfg7Ml8SuB1UqFU0DvuoIoQndg51sX2yA4Tahq0xD/ilyga1NHQ3jgbw5QKBZDt
# VbGGwfUgcNCwbZGDZCQtrRTcRmr4B7/tK4xFFOfvrKLEA8MByHdzmAYlIMsPM/fX
# YCl95S60wLsCjC5aJHBzM5NN2ZRQSLYLg0RKtrBoR5Zq+RlIBXKsS0hKgRbk9WAj
# uB5MvW3T/tl40KdnBlT9lCGh677+NM8EykDf1UIwggTsMIID1KADAgECAhMzAAAB
# Cix5rtd5e6asAAEAAAEKMA0GCSqGSIb3DQEBBQUAMHkxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBMB4XDTE1MDYwNDE3NDI0NVoXDTE2MDkwNDE3NDI0NVowgYMxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIx
# HjAcBgNVBAMTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJL8bza74QO5KNZG0aJhuqVG+2MWPi75R9LH7O3HmbEm
# UXW92swPBhQRpGwZnsBfTVSJ5E1Q2I3NoWGldxOaHKftDXT3p1Z56Cj3U9KxemPg
# 9ZSXt+zZR/hsPfMliLO8CsUEp458hUh2HGFGqhnEemKLwcI1qvtYb8VjC5NJMIEb
# e99/fE+0R21feByvtveWE1LvudFNOeVz3khOPBSqlw05zItR4VzRO/COZ+owYKlN
# Wp1DvdsjusAP10sQnZxN8FGihKrknKc91qPvChhIqPqxTqWYDku/8BTzAMiwSNZb
# /jjXiREtBbpDAk8iAJYlrX01boRoqyAYOCj+HKIQsaUCAwEAAaOCAWAwggFcMBMG
# A1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSJ/gox6ibN5m3HkZG5lIyiGGE3
# NDBRBgNVHREESjBIpEYwRDENMAsGA1UECxMETU9QUjEzMDEGA1UEBRMqMzE1OTUr
# MDQwNzkzNTAtMTZmYS00YzYwLWI2YmYtOWQyYjFjZDA1OTg0MB8GA1UdIwQYMBaA
# FMsR6MrStBZYAck3LjMWFrlMmgofMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY0NvZFNpZ1BDQV8w
# OC0zMS0yMDEwLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljQ29kU2lnUENBXzA4LTMx
# LTIwMTAuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCmqFOR3zsB/mFdBlrrZvAM2PfZ
# hNMAUQ4Q0aTRFyjnjDM4K9hDxgOLdeszkvSp4mf9AtulHU5DRV0bSePgTxbwfo/w
# iBHKgq2k+6apX/WXYMh7xL98m2ntH4LB8c2OeEti9dcNHNdTEtaWUu81vRmOoECT
# oQqlLRacwkZ0COvb9NilSTZUEhFVA7N7FvtH/vto/MBFXOI/Enkzou+Cxd5AGQfu
# FcUKm1kFQanQl56BngNb/ErjGi4FrFBHL4z6edgeIPgF+ylrGBT6cgS3C6eaZOwR
# XU9FSY0pGi370LYJU180lOAWxLnqczXoV+/h6xbDGMcGszvPYYTitkSJlKOGMIIF
# vDCCA6SgAwIBAgIKYTMmGgAAAAAAMTANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMTAwODMx
# MjIxOTMyWhcNMjAwODMxMjIyOTMyWjB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALJyWVwZMGS/HZpgICBC
# mXZTbD4b1m/My/Hqa/6XFhDg3zp0gxq3L6Ay7P/ewkJOI9VyANs1VwqJyq4gSfTw
# aKxNS42lvXlLcZtHB9r9Jd+ddYjPqnNEf9eB2/O98jakyVxF3K+tPeAoaJcap6Vy
# c1bxF5Tk/TWUcqDWdl8ed0WDhTgW0HNbBbpnUo2lsmkv2hkL/pJ0KeJ2L1TdFDBZ
# +NKNYv3LyV9GMVC5JxPkQDDPcikQKCLHN049oDI9kM2hOAaFXE5WgigqBTK3S9dP
# Y+fSLWLxRT3nrAgA9kahntFbjCZT6HqqSvJGzzc8OJ60d1ylF56NyxGPVjzBrAlf
# A9MCAwEAAaOCAV4wggFaMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFMsR6MrS
# tBZYAck3LjMWFrlMmgofMAsGA1UdDwQEAwIBhjASBgkrBgEEAYI3FQEEBQIDAQAB
# MCMGCSsGAQQBgjcVAgQWBBT90TFO0yaKleGYYDuoMW+mPLzYLTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTAfBgNVHSMEGDAWgBQOrIJgQFYnl+UlE/wq4QpTlVnk
# pDBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEE
# SDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2Nl
# cnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDANBgkqhkiG9w0BAQUFAAOCAgEAWTk+
# fyZGr+tvQLEytWrrDi9uqEn361917Uw7LddDrQv+y+ktMaMjzHxQmIAhXaw9L0y6
# oqhWnONwu7i0+Hm1SXL3PupBf8rhDBdpy6WcIC36C1DEVs0t40rSvHDnqA2iA6VW
# 4LiKS1fylUKc8fPv7uOGHzQ8uFaa8FMjhSqkghyT4pQHHfLiTviMocroE6WRTsgb
# 0o9ylSpxbZsa+BzwU9ZnzCL/XB3Nooy9J7J5Y1ZEolHN+emjWFbdmwJFRC9f9Nqu
# 1IIybvyklRPk62nnqaIsvsgrEA5ljpnb9aL6EiYJZTiU8XofSrvR4Vbo0HiWGFzJ
# NRZf3ZMdSY4tvq00RBzuEBUaAF3dNVshzpjHCe6FDoxPbQ4TTj18KUicctHzbMrB
# 7HCjV5JXfZSNoBtIA1r3z6NnCnSlNu0tLxfI5nI3EvRvsTxngvlSso0zFmUeDord
# EN5k9G/ORtTTF+l5xAS00/ss3x+KnqwK+xMnQK3k+eGpf0a7B2BHZWBATrBC7E7t
# s3Z52Ao0CW0cgDEf4g5U3eWh++VHEK1kmP9QFi58vwUheuKVQSdpw5OPlcmN2Jsh
# rg1cnPCiroZogwxqLbt2awAdlq3yFnv2FoMkuYjPaqhHMS+a3ONxPdcAfmJH0c6I
# ybgY+g5yjcGjPa8CQGr/aZuW4hCoELQ3UAjWwz0wggYHMIID76ADAgECAgphFmg0
# AAAAAAAcMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAX
# BgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290
# IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0wNzA0MDMxMjUzMDlaFw0yMTA0MDMx
# MzAzMDlaMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAf
# BgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQTCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJ+hbLHf20iSKnxrLhnhveLjxZlRI1Ctzt0YTiQP7tGn
# 0UytdDAgEesH1VSVFUmUG0KSrphcMCbaAGvoe73siQcP9w4EmPCJzB/LMySHnfL0
# Zxws/HvniB3q506jocEjU8qN+kXPCdBer9CwQgSi+aZsk2fXKNxGU7CG0OUoRi4n
# rIZPVVIM5AMs+2qQkDBuh/NZMJ36ftaXs+ghl3740hPzCLdTbVK0RZCfSABKR2YR
# JylmqJfk0waBSqL5hKcRRxQJgp+E7VV4/gGaHVAIhQAQMEbtt94jRrvELVSfrx54
# QTF3zJvfO4OToWECtR0Nsfz3m7IBziJLVP/5BcPCIAsCAwEAAaOCAaswggGnMA8G
# A1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFCM0+NlSRnAK7UD7dvuzK7DDNbMPMAsG
# A1UdDwQEAwIBhjAQBgkrBgEEAYI3FQEEAwIBADCBmAYDVR0jBIGQMIGNgBQOrIJg
# QFYnl+UlE/wq4QpTlVnkpKFjpGEwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5ghB5rRahSqClrUxzWPQHEy5lMFAGA1UdHwRJ
# MEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1
# Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRIMEYwRAYIKwYB
# BQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0Um9vdENlcnQuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEB
# BQUAA4ICAQAQl4rDXANENt3ptK132855UU0BsS50cVttDBOrzr57j7gu1BKijG1i
# uFcCy04gE1CZ3XpA4le7r1iaHOEdAYasu3jyi9DsOwHu4r6PCgXIjUji8FMV3U+r
# kuTnjWrVgMHmlPIGL4UD6ZEqJCJw+/b85HiZLg33B+JwvBhOnY5rCnKVuKE5nGct
# xVEO6mJcPxaYiyA/4gcaMvnMMUp2MT0rcgvI6nA9/4UKE9/CCmGO8Ne4F+tOi3/F
# NSteo7/rvH0LQnvUU3Ih7jDKu3hlXFsBFwoUDtLaFJj1PLlmWLMtL+f5hYbMUVbo
# nXCUbKw5TNT2eb+qGHpiKe+imyk0BncaYsk9Hm0fgvALxyy7z0Oz5fnsfbXjpKh0
# NbhOxXEjEiZ2CzxSjHFaRkMUvLOzsE1nyJ9C/4B5IYCeFTBm6EISXhrIniIh0EPp
# K+m79EjMLNTYMoBMJipIJF9a6lbvpt6Znco6b72BJ3QGEe52Ib+bgsEnVLaxaj2J
# oXZhtG6hE6a/qkfwEm/9ijJssv7fUciMI8lmvZ0dhxJkAj0tr1mPuOQh5bWwymO0
# eFQF1EEuUKyUsKV4q7OglnUa2ZKHE3UiLzKoCG6gW4wlv6DvhMoh1useT8ma7kng
# 9wFlb4kLfchpyOZu6qeXzjEp/w7FW1zYTRuh2Povnj8uVRZryROj/TGCBK8wggSr
# AgEBMIGQMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAh
# BgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBAhMzAAABCix5rtd5e6as
# AAEAAAEKMAkGBSsOAwIaBQCggcgwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFEIq
# BlNDkLBqgolG1WiY2gkqv297MGgGCisGAQQBgjcCAQwxWjBYoDaANABNAGkAYwBy
# AG8AcwBvAGYAdAAgAEgAUABDACAAUABhAGMAawAgADIAMAAxADIAIABSADKhHoAc
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL0hQQzANBgkqhkiG9w0BAQEFAASCAQB9
# tGnCaicGTJnxI8cZeuhYaV5s+E1Cag70pI+BKx8PPPx+EINKqLs0cj3U5IVHrOXp
# 9FQ77hCi4gpVkbLSaVS70yigbUunWwFuPRCeNWQse+0VecI66kT5Kg4LLPGjYcYL
# eReq4Z/4f20he3VGnCYnVp7ZoKhvzaUJA36mWkxlIo1yGSewl8shq4NPb7w9groA
# gmajEGCdtpcU0rNaV+7JWZIUAnSZvw+56lcq4uPD+yW0ru19RuQhuCgYHXXngc7C
# lczbOPCCSg2RCC4BC7EbxDZvBObqq8VCnPs9js9zXNWkJfnOY5oqH/NHCXfBD6u3
# VuhaPSVXwUSoO9xvvaRnoYICKDCCAiQGCSqGSIb3DQEJBjGCAhUwggIRAgEBMIGO
# MHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAfBgNVBAMT
# GE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQQITMwAAAHJp9pDCFDtJoQAAAAAAcjAJ
# BgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkDMQsGCSqGSIb3DQEHATAcBgkqhkiG9w0B
# CQUxDxcNMTUwNzA2MTcyMzExWjAjBgkqhkiG9w0BCQQxFgQUQ+BatjRL6ogHymoz
# ATzD8o/ZSnIwDQYJKoZIhvcNAQEFBQAEggEAHZYjgqMrOmzPKmHwlscfW+sUlqOn
# JC/9A3xde1uRKWxOY3O1cpMQ9t09S1Cz9wIvLRAxgZscsQ9YbRtzuTuIuilhm/ww
# vIsCEO0r3EYjPTgWcMbU0TwSMeloYfn871LZ8RHTDQuBr4E4VBozuZlk24qnPPh4
# 8661XhgMV2RslgFIyaiTfnNr891H2+7wZ6qOr0c4v1Wa98AxGBf7uTwH6Mjq6Bji
# 7RzTk9PoNoQpZSVUQQdOdydmzjsOk7RE/2rKtz9+q7vKmtMeivxJD702jDFebtUz
# aD99ZhfHPPgchhAg/VxO1DgP4QKIeXebHol1bBrlv2SnXuceitZGCUcd6A==
# SIG # End signature block
