# private
# Clear all reported progress.
function Clear-Progress()
{
    [Tracing]::states.Clear();
}

# private
# Add a datetime prefix before $log
function AppendDateTime($log)
{
    $date = Get-Date -Format 'MM/dd/yyyy HH:mm:ss' 
    "{0} - {1}" -f $date, $log
}    

# private
# Set the cursor to the saved end.
function Set-CursorEnd()
{
    if ([Tracing]::MaxLine -gt $host.UI.RawUI.CursorPosition.Y)
    {
        $old = $host.UI.RawUI.CursorPosition
        $old.Y = [Tracing]::MaxLine
        $host.UI.RawUI.CursorPosition = $old
    }
}

# private
# Save the current cursor end to a variable.
function Save-CursorEnd()
{
    if ([Tracing]::MaxLine -lt $host.UI.RawUI.CursorPosition.Y)
    {
        [Tracing]::MaxLine = $host.UI.RawUI.CursorPosition.Y
    }
}

# public
# Wait for all reported progress to complete, display the progress bar, progress notification to screen.
# This function will block until all progress complete.
function Wait-Progress
{
    [CmdletBinding(DefaultParametersetName="Jobs")] 
    Param
    (
        # No Wait
        [Parameter(Mandatory=$false, Position=0, ParameterSetName="NoWait")]
        [Switch] $NoWait,

        # Wait jobs
        [Parameter(Mandatory=$false, Position=0, ParameterSetName="Jobs")]
        [System.Management.Automation.Job[]] $Jobs,

        [Parameter(Mandatory=$false, Position=1, ParameterSetName="Jobs")]
        [Switch] $Any
    )

    Set-CursorEnd

    $windowWidth = $host.UI.RawUI.WindowSize.Width
    $dateWidth = 17
    $progressWidth = 5
    $statusWidth = 10
    $nameWidth = 38

    $textWidth = $windowWidth - $progressWidth - $statusWidth - $dateWidth - $nameWidth - 1

    $scroll = "/-\|/-\|"

    if ($NoWait)
    {
        $Jobs = @();
    }

    if ($Jobs -eq $null)
    {
        $Jobs = @([Tracing]::states.GetEnumerator() | Where-Object { $_.Value.Job -ne $null } | %{ $_.Value.Job })
    }

    $jobNum = $Jobs.Count

    do
    {
        Start-Sleep -Milliseconds 300

        $exitLoop = $false
        $doneJobs = @($Jobs | Where-Object { ($_.State -eq "Completed") -or ($_.State -eq "Failed") })
        if(($doneJobs.Count -eq $Jobs.Count) -or ($Any -and ($doneJobs.Count -gt 0)))
        {
            $exitLoop = $true
        }
        
        foreach($subProgress in ([Tracing]::states.GetEnumerator() | Where-Object { !$_.Value.Ended }))
        {
            $currentState = $subProgress.Value;

            $messageColor = [ConsoleColor]::White

            $errorInJob = $false
            if ($currentState.Job -ne $null)
            {
                $job = $currentState.Job;
                if ($job.JobStateInfo.State -eq [System.Management.Automation.JobState]::Completed -or
                    $job.JobStateInfo.State -eq [System.Management.Automation.JobState]::Failed)
                {
                    $currentState.Ended = $true;
                }

                $currentState.Message = ""

                if ($job.ChildJobs[0].JobStateInfo.Reason -ne $null)
                {
                    $currentState.Message = $job.ChildJobs[0].JobStateInfo.Reason.Message
                }

                if (!$currentState.Message)
                {
                    if($job.ChildJobs[0].Error.Count -gt 0)
                    {
                        $currentState.Message = $job.ChildJobs[0].Error[-1] | Out-String
                    }
                    else
                    {
                        $currentState.Message = $job.ChildJobs[0].Verbose | ?{ $_.Message -ne $null -and $_.Message.StartsWith("[Error]") } | %{$_.Message} | select -Last(1)
                        if ($currentState.Message)
                        {
                            $start = $currentState.Message.IndexOf("-");
                            if ($start -gt 0 -and $currentState.Message.Length -gt $start + 2)
                            {
                                $currentState.Message = $currentState.Message.Substring($start + 2);
                            }
                        }
                    }
                }

                if (!$currentState.Message)
                {
                    $messageColor = [ConsoleColor]::White
                    $currentState.Message = $job.ChildJobs[0].Verbose | ?{ $_.Message -ne $null -and $_.Message -match "^\[(Information|Warning)\]" } | %{$_.Message} | select -Last(1)
                    if ($currentState.Message)
                    {
                        if ($currentState.Message.StartsWith("[Warning]"))
                        {
                            $messageColor = [ConsoleColor]::Yellow
                        }

                        $start = $currentState.Message.IndexOf("-");
                        if ($start -gt 0 -and $currentState.Message.Length -gt $start + 2)
                        {
                            $currentState.Message = $currentState.Message.Substring($start + 2);
                        }
                    }
                }
                else
                {
                    $messageColor = [ConsoleColor]::Red
                    $errorInJob = $true
                }

                $currentState.State = $job.State;
            }
            else
            {
                if(($currentState.State -eq "Passed") -or ($currentState.State -eq "Failed"))
                {
                    $currentState.Ended = $true;
                }
            }
                
            if (!$currentState.Message)
            {
                $messageColor = [ConsoleColor]::White
                $currentState.Message = "---"
            }

            if ($currentState.State -eq "Completed")
            {
                $currentState.State = "Passed";
            }

            $currentState.Message = $currentState.Message -replace "`t|`n|`r", " "

            if ($currentState.Message.Length -gt $textWidth)
            {
                $currentState.Message = $currentState.Message.Substring(0, $textWidth - 5) + " ... "
            }

            if ($currentState.CursorPosition -eq $null)
            {
                # new item, move to end
                Set-CursorEnd

                # save cursor
                $currentState.CursorPosition = $host.UI.RawUI.CursorPosition;
            }
            else
            {
                # old item, move to saved place.
                $host.UI.RawUI.CursorPosition = $currentState.CursorPosition;
            }

            # OUTPUT Starts
            # Date
            Write-Host $currentState.StartTime.ToString('MM/dd HH:mm:ss') -f White -NoNewline
            if ($currentState.State -eq "Running")
            {
                $progressSign = $scroll[$currentState.ScrollIndex];
                Write-Host " $progressSign " -f Cyan -NoNewline
            }
            else
            {
                Write-Host " - " -f White -NoNewline
            }

            # Progress
            if ($currentState.State -eq "Queued")
            {
                Write-Host "     " -f White -NoNewline
            }
            else
            {
                $now = Get-Date
                $totalSeconds = ($now - $currentState.StartTime).TotalSeconds
                if ($totalSeconds -gt 86400)
                {
                    Write-Host (" > 1d") -NoNewline
                }
                elseif ($totalSeconds -ge 600)
                {
                    # only show m.
                    $totalSeconds /= 60
                    Write-Host ("{0,4:0}m" -f $totalSeconds) -NoNewline
                }
                elseif ($totalSeconds -ge 120)
                {
                    Write-Host ("{0,5}" -f ("{0:0}m{1:0}s" -f ([Math]::Floor($totalSeconds / 60)), ($totalSeconds % 60))) -NoNewline
                }
                else
                {
                    Write-Host ("{0,4:0}s" -f $totalSeconds) -NoNewline
                }
            }

            $currentState.ScrollIndex ++
            if ($currentState.ScrollIndex -ge $scroll.Length)
            {
                $currentState.ScrollIndex = 0
            }

            # Status
            Write-Host "[" -f White -NoNewline
            $color = [ConsoleColor]::Cyan
            if ($currentState.State -eq "Failed")
            {
                $color = [ConsoleColor]::Red
            }
            elseif($errorInJob)
            {
                # if there is error in the job, we shall mark the status with Yellow color to alert user
                $color = [ConsoleColor]::Yellow
            }
            elseif ($currentState.State -eq "Running")
            {
                $color = [ConsoleColor]::Cyan
            }
            elseif ($currentState.State -eq "Passed")
            {
                $color = [ConsoleColor]::Green
            }

            $statusTextWidth = $statusWidth - 3
            Write-Host ("{0, -$statusTextWidth}" -f $currentState.State) -f $color -NoNewline
            Write-Host '] ' -f White -NoNewline

            # Name
            Write-Host ("{0, -$nameWidth}" -f $currentState.Name) -f Gray -NoNewline

            # Message
            Write-Host ("{0, -$textWidth}" -f $currentState.Message) -f $messageColor -NoNewline

            Write-Host

            Save-CursorEnd
        }

        $completed = ([Tracing]::states.GetEnumerator() | Where-Object { $_.Value.Ended } | measure).Count
        $total = [Tracing]::SubProgressCount
        $mainProgress = [Tracing]::MainProgress

        # old item, move to saved place.
        # "$host.UI.RawUI.CursorPosition = [Tracing]::CursorPosition" throws object unbox exception in Win10 Preview
        $savedCurP = [Tracing]::CursorPosition
        $host.UI.RawUI.CursorPosition = $savedCurP

        AppendDateTime $mainProgress | Write-Host -f White -NoNewline
        Write-Host " - ( $completed / $total )" -f Gray

        Set-CursorEnd
    } while (-not $exitLoop)
    
    # return the done jobs if the switch 'Any' is specified
    if($Any)
    {
        return $doneJobs    
    }
}

# public
# Register a main progress and set its sub progress count
# This will clear all registered progress if the previous main progress is not the same as the current.
# It will do nothing if the previous is the same as current.
function Start-Progress($mainProgress, $subProgressCount)
{
    Clear-Progress
    [Tracing]::MainProgress = $mainProgress
    $host.UI.RawUI.WindowTitle = $mainProgress
    Set-CursorEnd
    Write-Host

    [Tracing]::CursorPosition = $host.UI.RawUI.CursorPosition

    # Date
    AppendDateTime $mainProgress | Write-Host -f White -NoNewline
    Write-Host " - ( 0 / $subProgressCount )" -f Gray

    $width = $host.UI.RawUI.WindowSize.Width - 2
    Write-Host (New-Object System.String '-', $width) -f Gray

    [Tracing]::SubProgressCount = $subProgressCount

    Save-CursorEnd
}

# Public
# Report a sub progress with sub progress name, status, and message.
# If the status is reported as string, the progress will be displayed without the progress notification sign, and it
# won't cause the Wait-Progress to block.
# If the status is reported as Job, the progress will be displayed with a progress notification sign.
# and it will block the Wait-Progress call until the Job is completed or failed.
function Report-Progress
{
    [CmdletBinding(DefaultParametersetName="Jobs")] 
    Param
    (
        [Parameter(Position = 0, Mandatory = $true)]
        [System.String]$subProgress,

        [Parameter(Position = 1, Mandatory = $true, ParameterSetName = "Status")]
        [ValidateSet("Queued", "Failed", "Passed", "Running")]
        [System.String]$status,

        [Parameter(Position = 1, Mandatory = $true, ParameterSetName = "Job")]
        [System.Management.Automation.Job]$job,

        [Parameter(Position = 2, Mandatory = $false)]
        [System.String]$message
    )

    if ($message -eq $null)
    {
        $message = "---";
    }

    if ($PSCmdlet.ParameterSetName -eq "Job")
    {
        [Tracing]::ReportProgress($subProgress, $job, $message);
    }
    else
    {
        [Tracing]::ReportProgress($subProgress, $status, $message);
    }

    # false means not block, just output to the screen once.
    Wait-Progress -NoWait
}

# Public
# Write a string to standard warning stream 3>
function Warning($log)
{
    Set-CursorEnd
    AppendDateTime $log | Write-Warning
    Save-CursorEnd
}

# Public
# Write a string to standard Error stream 2>
function Error($log)
{
    Set-CursorEnd
    AppendDateTime $log | Write-Error
    Save-CursorEnd
}

# Public
# Write a string to standard verbose stream 4>
function Verbose($log)
{
    Set-CursorEnd
    AppendDateTime $log | Write-Verbose
    Save-CursorEnd
}

# Public
# Initialize before reporting progress.
function Initialize-Progress()
{
    $TraceSource = @"
    using System;
    using System.Collections.Generic;
    using System.Management.Automation.Host;
    using System.Management.Automation;

    public static class Tracing
    {
        public static string OrigConsoleTitle
        {
            get;
            set;
        }

        public static int SubProgressCount
        {
            get;
            set;
        }

        public static int MaxLine
        {
            get;
            set;
        }

        public static string MainProgress
        {
            get;
            set;
        }

        public static Coordinates? CursorPosition { get; set; }

        public class Progress
        {
            public DateTime StartTime { get; set; }
            public string Name { get; set; }
            public string Message { get; set; }
            public string State   { get; set; }
            public int ScrollIndex { get; set; }
            public Coordinates? CursorPosition { get; set; }
            public Job Job { get; set; }
            public bool Ended { get; set; }
        }

        public static Dictionary<string, Progress> states = new Dictionary<string, Progress>();
        public static void ReportProgress(string subProgress, Job job, string message)
        {
            Progress progress = null;
            if (!states.TryGetValue(subProgress, out progress))
            {
                progress = new Progress ()
                {
                    Name = subProgress, 
                    Message = message,
                    Job = job,
                    StartTime = DateTime.Now,
                    ScrollIndex = 0,
                };

                states[subProgress] = progress;
            }
            else
            {
                // If the previous state is Queued, reset the StartTime
                if(progress.State.Equals("Queued", StringComparison.OrdinalIgnoreCase))
                {
                    progress.StartTime = DateTime.Now;
                }
            }

            progress.Job = job;
            progress.Ended = false;
            progress.Message = message;
        }

        public static void ReportProgress(string subProgress, string state, string message)
        {
            Progress progress = null;
            if (!states.TryGetValue(subProgress, out progress))
            {
                progress = new Progress ()
                {
                    Name = subProgress,
                    Message = message,
                    State = state,
                    StartTime = DateTime.Now,
                    ScrollIndex = 0,
                };

                states[subProgress] = progress;
            }
            else
            {
                // If the previous state is Queued, reset the StartTime
                if(progress.State.Equals("Queued", StringComparison.OrdinalIgnoreCase))
                {
                    progress.StartTime = DateTime.Now;
                }
            }

            progress.Ended = false;
            progress.State = state;
            progress.Message = message;
        }
    }
"@

    Add-Type -TypeDefinition $TraceSource
    Clear-Progress
    Save-CursorEnd

    # Save the original Title
    [Tracing]::OrigConsoleTitle = $host.UI.RawUI.WindowTitle

    $preferWidth = 160
    $preferHeight = 60
    $bufferSize = $host.UI.RawUI.BufferSize
    # Set the buffer width to a large value to test whether the maximum supported window width can be 160
    $bufferSize.Width = $preferWidth + 15 
    $host.UI.RawUI.BufferSize = $bufferSize
    $maxWinSize = $host.UI.RawUI.Get_MaxWindowSize()

    $width = $maxWinSize.Width - 10
    $height = $maxWinSize.Height - 4
    if($width -gt $preferWidth)
    {
        $width = $preferWidth
    }
    if($height -gt $preferHeight )
    {
        $height = $preferHeight 
    }

    $bufferSize.Width = $width

    # Set the buffer height to the maximum value to minimize the probability of window scroll.
    $bufferHeight = 9999
    while($bufferHeight -gt 200)
    {
        try
        {
            $bufferSize.Height = $bufferHeight
            $host.UI.RawUI.BufferSize = $bufferSize
            break
        }
        catch
        {
            $bufferHeight = [Math]::Ceiling($bufferHeight/2)
        }
    }
    
    try
    {
        # Clear the buffer, and reset the saved [Tracing]::MaxLine to let the next line start from the beginning.
        # The .Net call is faster according to some post, so don't use cls or Clear-Host
        if($host.UI.RawUI.BufferSize.Height -lt ($host.UI.RawUI.CursorPosition.Y + 200))
        {
            [System.Console]::Clear()
            [Tracing]::MaxLine = $host.UI.RawUI.CursorPosition.Y
        }
    }
    catch
    {
        # Ignore the exception
    }

    $windowSize = $bufferSize
    $windowSize.Height = $height
    $host.UI.RawUI.WindowSize = $windowSize
}

# Public 
# Un-initialize after done.
function Uninitialize-Progress()
{
    Clear-Progress
    [Tracing]::MainProgress = $null
    $host.UI.RawUI.WindowTitle = [Tracing]::OrigConsoleTitle
    Set-CursorEnd
}

# SIG # Begin signature block
# MIIaxwYJKoZIhvcNAQcCoIIauDCCGrQCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUolOUWnUwdiba0HIAMgelFnsz
# ysOgghWCMIIEwzCCA6ugAwIBAgITMwAAAHJp9pDCFDtJoQAAAAAAcjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUwMzIwMTczMjA0
# WhcNMTYwNjIwMTczMjA0WjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OjMxQzUtMzBCQS03QzkxMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoxPVJ2iBlQVE
# sJuqKGMDs9En3ra2ZyP2mieV2SfpWI7AXCdUfRHZ7B+712Ff5z/jWtu5iNakH+yw
# OsPZy911xkYMU2T9mm1IXZcHoKEMvj6SL5H3q2dq0SX5n3iqXJcxf4wwjDu9cKit
# 5v8t+rrQxeVIuseUK34MXOH/NYpAv1H0FU0vm0MjOh7ltNRzua86yAgrQBSj7gML
# NybNLvq+M9b3JBVF6cFLmII8syS9LSf7drlwWzPRJ71Q/6+TVq59UtTQO217Juvk
# 9BdtStzFmMyR+7VdVxybCwxPr6dW/2U7oouf4UxtszHDMtHw7XEnA2jg2iq6ODkZ
# frNXUQCMhQIDAQABo4IBCTCCAQUwHQYDVR0OBBYEFND9obr0IhS4jv6FL059LwBC
# pwFTMB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZI
# hvcNAQEFBQADggEBACJEd3U6Z3Bw2w+djrTKkrwYGl8Ys2vdoryZDDm9ybp6jjCk
# QCKLBGssgmRVJcEaYjUTkqGomLk5Mt4r5X6S+Zf1lgF3IKX9tqVbNlX1qMmRZUQ4
# 4Xfg7Ml8SuB1UqFU0DvuoIoQndg51sX2yA4Tahq0xD/ilyga1NHQ3jgbw5QKBZDt
# VbGGwfUgcNCwbZGDZCQtrRTcRmr4B7/tK4xFFOfvrKLEA8MByHdzmAYlIMsPM/fX
# YCl95S60wLsCjC5aJHBzM5NN2ZRQSLYLg0RKtrBoR5Zq+RlIBXKsS0hKgRbk9WAj
# uB5MvW3T/tl40KdnBlT9lCGh677+NM8EykDf1UIwggTsMIID1KADAgECAhMzAAAB
# Cix5rtd5e6asAAEAAAEKMA0GCSqGSIb3DQEBBQUAMHkxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBMB4XDTE1MDYwNDE3NDI0NVoXDTE2MDkwNDE3NDI0NVowgYMxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIx
# HjAcBgNVBAMTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJL8bza74QO5KNZG0aJhuqVG+2MWPi75R9LH7O3HmbEm
# UXW92swPBhQRpGwZnsBfTVSJ5E1Q2I3NoWGldxOaHKftDXT3p1Z56Cj3U9KxemPg
# 9ZSXt+zZR/hsPfMliLO8CsUEp458hUh2HGFGqhnEemKLwcI1qvtYb8VjC5NJMIEb
# e99/fE+0R21feByvtveWE1LvudFNOeVz3khOPBSqlw05zItR4VzRO/COZ+owYKlN
# Wp1DvdsjusAP10sQnZxN8FGihKrknKc91qPvChhIqPqxTqWYDku/8BTzAMiwSNZb
# /jjXiREtBbpDAk8iAJYlrX01boRoqyAYOCj+HKIQsaUCAwEAAaOCAWAwggFcMBMG
# A1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSJ/gox6ibN5m3HkZG5lIyiGGE3
# NDBRBgNVHREESjBIpEYwRDENMAsGA1UECxMETU9QUjEzMDEGA1UEBRMqMzE1OTUr
# MDQwNzkzNTAtMTZmYS00YzYwLWI2YmYtOWQyYjFjZDA1OTg0MB8GA1UdIwQYMBaA
# FMsR6MrStBZYAck3LjMWFrlMmgofMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY0NvZFNpZ1BDQV8w
# OC0zMS0yMDEwLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljQ29kU2lnUENBXzA4LTMx
# LTIwMTAuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCmqFOR3zsB/mFdBlrrZvAM2PfZ
# hNMAUQ4Q0aTRFyjnjDM4K9hDxgOLdeszkvSp4mf9AtulHU5DRV0bSePgTxbwfo/w
# iBHKgq2k+6apX/WXYMh7xL98m2ntH4LB8c2OeEti9dcNHNdTEtaWUu81vRmOoECT
# oQqlLRacwkZ0COvb9NilSTZUEhFVA7N7FvtH/vto/MBFXOI/Enkzou+Cxd5AGQfu
# FcUKm1kFQanQl56BngNb/ErjGi4FrFBHL4z6edgeIPgF+ylrGBT6cgS3C6eaZOwR
# XU9FSY0pGi370LYJU180lOAWxLnqczXoV+/h6xbDGMcGszvPYYTitkSJlKOGMIIF
# vDCCA6SgAwIBAgIKYTMmGgAAAAAAMTANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMTAwODMx
# MjIxOTMyWhcNMjAwODMxMjIyOTMyWjB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALJyWVwZMGS/HZpgICBC
# mXZTbD4b1m/My/Hqa/6XFhDg3zp0gxq3L6Ay7P/ewkJOI9VyANs1VwqJyq4gSfTw
# aKxNS42lvXlLcZtHB9r9Jd+ddYjPqnNEf9eB2/O98jakyVxF3K+tPeAoaJcap6Vy
# c1bxF5Tk/TWUcqDWdl8ed0WDhTgW0HNbBbpnUo2lsmkv2hkL/pJ0KeJ2L1TdFDBZ
# +NKNYv3LyV9GMVC5JxPkQDDPcikQKCLHN049oDI9kM2hOAaFXE5WgigqBTK3S9dP
# Y+fSLWLxRT3nrAgA9kahntFbjCZT6HqqSvJGzzc8OJ60d1ylF56NyxGPVjzBrAlf
# A9MCAwEAAaOCAV4wggFaMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFMsR6MrS
# tBZYAck3LjMWFrlMmgofMAsGA1UdDwQEAwIBhjASBgkrBgEEAYI3FQEEBQIDAQAB
# MCMGCSsGAQQBgjcVAgQWBBT90TFO0yaKleGYYDuoMW+mPLzYLTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTAfBgNVHSMEGDAWgBQOrIJgQFYnl+UlE/wq4QpTlVnk
# pDBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEE
# SDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2Nl
# cnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDANBgkqhkiG9w0BAQUFAAOCAgEAWTk+
# fyZGr+tvQLEytWrrDi9uqEn361917Uw7LddDrQv+y+ktMaMjzHxQmIAhXaw9L0y6
# oqhWnONwu7i0+Hm1SXL3PupBf8rhDBdpy6WcIC36C1DEVs0t40rSvHDnqA2iA6VW
# 4LiKS1fylUKc8fPv7uOGHzQ8uFaa8FMjhSqkghyT4pQHHfLiTviMocroE6WRTsgb
# 0o9ylSpxbZsa+BzwU9ZnzCL/XB3Nooy9J7J5Y1ZEolHN+emjWFbdmwJFRC9f9Nqu
# 1IIybvyklRPk62nnqaIsvsgrEA5ljpnb9aL6EiYJZTiU8XofSrvR4Vbo0HiWGFzJ
# NRZf3ZMdSY4tvq00RBzuEBUaAF3dNVshzpjHCe6FDoxPbQ4TTj18KUicctHzbMrB
# 7HCjV5JXfZSNoBtIA1r3z6NnCnSlNu0tLxfI5nI3EvRvsTxngvlSso0zFmUeDord
# EN5k9G/ORtTTF+l5xAS00/ss3x+KnqwK+xMnQK3k+eGpf0a7B2BHZWBATrBC7E7t
# s3Z52Ao0CW0cgDEf4g5U3eWh++VHEK1kmP9QFi58vwUheuKVQSdpw5OPlcmN2Jsh
# rg1cnPCiroZogwxqLbt2awAdlq3yFnv2FoMkuYjPaqhHMS+a3ONxPdcAfmJH0c6I
# ybgY+g5yjcGjPa8CQGr/aZuW4hCoELQ3UAjWwz0wggYHMIID76ADAgECAgphFmg0
# AAAAAAAcMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAX
# BgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290
# IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0wNzA0MDMxMjUzMDlaFw0yMTA0MDMx
# MzAzMDlaMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAf
# BgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQTCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJ+hbLHf20iSKnxrLhnhveLjxZlRI1Ctzt0YTiQP7tGn
# 0UytdDAgEesH1VSVFUmUG0KSrphcMCbaAGvoe73siQcP9w4EmPCJzB/LMySHnfL0
# Zxws/HvniB3q506jocEjU8qN+kXPCdBer9CwQgSi+aZsk2fXKNxGU7CG0OUoRi4n
# rIZPVVIM5AMs+2qQkDBuh/NZMJ36ftaXs+ghl3740hPzCLdTbVK0RZCfSABKR2YR
# JylmqJfk0waBSqL5hKcRRxQJgp+E7VV4/gGaHVAIhQAQMEbtt94jRrvELVSfrx54
# QTF3zJvfO4OToWECtR0Nsfz3m7IBziJLVP/5BcPCIAsCAwEAAaOCAaswggGnMA8G
# A1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFCM0+NlSRnAK7UD7dvuzK7DDNbMPMAsG
# A1UdDwQEAwIBhjAQBgkrBgEEAYI3FQEEAwIBADCBmAYDVR0jBIGQMIGNgBQOrIJg
# QFYnl+UlE/wq4QpTlVnkpKFjpGEwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5ghB5rRahSqClrUxzWPQHEy5lMFAGA1UdHwRJ
# MEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1
# Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRIMEYwRAYIKwYB
# BQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0Um9vdENlcnQuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEB
# BQUAA4ICAQAQl4rDXANENt3ptK132855UU0BsS50cVttDBOrzr57j7gu1BKijG1i
# uFcCy04gE1CZ3XpA4le7r1iaHOEdAYasu3jyi9DsOwHu4r6PCgXIjUji8FMV3U+r
# kuTnjWrVgMHmlPIGL4UD6ZEqJCJw+/b85HiZLg33B+JwvBhOnY5rCnKVuKE5nGct
# xVEO6mJcPxaYiyA/4gcaMvnMMUp2MT0rcgvI6nA9/4UKE9/CCmGO8Ne4F+tOi3/F
# NSteo7/rvH0LQnvUU3Ih7jDKu3hlXFsBFwoUDtLaFJj1PLlmWLMtL+f5hYbMUVbo
# nXCUbKw5TNT2eb+qGHpiKe+imyk0BncaYsk9Hm0fgvALxyy7z0Oz5fnsfbXjpKh0
# NbhOxXEjEiZ2CzxSjHFaRkMUvLOzsE1nyJ9C/4B5IYCeFTBm6EISXhrIniIh0EPp
# K+m79EjMLNTYMoBMJipIJF9a6lbvpt6Znco6b72BJ3QGEe52Ib+bgsEnVLaxaj2J
# oXZhtG6hE6a/qkfwEm/9ijJssv7fUciMI8lmvZ0dhxJkAj0tr1mPuOQh5bWwymO0
# eFQF1EEuUKyUsKV4q7OglnUa2ZKHE3UiLzKoCG6gW4wlv6DvhMoh1useT8ma7kng
# 9wFlb4kLfchpyOZu6qeXzjEp/w7FW1zYTRuh2Povnj8uVRZryROj/TGCBK8wggSr
# AgEBMIGQMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAh
# BgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBAhMzAAABCix5rtd5e6as
# AAEAAAEKMAkGBSsOAwIaBQCggcgwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFLae
# w1Pk3I8VDddLzm7COQhw8vIVMGgGCisGAQQBgjcCAQwxWjBYoDaANABNAGkAYwBy
# AG8AcwBvAGYAdAAgAEgAUABDACAAUABhAGMAawAgADIAMAAxADIAIABSADKhHoAc
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL0hQQzANBgkqhkiG9w0BAQEFAASCAQBl
# /a8wNRGUT01nIJ7mgzekZqqC5vFFwhlcFVo9/dPFdSqqUzPGaAPjXCeXnHVg7ds5
# Jxw3jdLN7j7PPKPAxeMgxrJfAOW895RyNtEQYH0VRxmwqNGyHupkI/KwOfyNrcl0
# LRAOBQuCj0+/us2YghgARsyqB3aCXQb09MZPuRJ7ofVmxhoSdvtTYteJu+NH3JXW
# 7Ix7qjpbXuIP4cjPuhgyne14otZ6N5cJSTUF1Kz87wZOMDG5n8YvN8NUe3Hg5iAn
# QSdIuq4q3ifpFEGRiESt9/vSWjAACzvgRQFSZPgx8wt1zW8eEkaRbfnDbYVhlynt
# 22XeYqt3ct60TTk1XK3loYICKDCCAiQGCSqGSIb3DQEJBjGCAhUwggIRAgEBMIGO
# MHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAfBgNVBAMT
# GE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQQITMwAAAHJp9pDCFDtJoQAAAAAAcjAJ
# BgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkDMQsGCSqGSIb3DQEHATAcBgkqhkiG9w0B
# CQUxDxcNMTUwNzA2MTcyMzA4WjAjBgkqhkiG9w0BCQQxFgQUsCDmLCjYrQBW9cwS
# KbJsU85hc9YwDQYJKoZIhvcNAQEFBQAEggEAhKmScJv5tnHbMimkqCl5WQQkck4/
# xBmlF6V5AWi3dUchUP4P1OWEW13IENgO7jNDx+bGLP5RYHg6mid+7+GmnxjQKjQb
# 2FksSai1iHJFDOdMDFkl3GqRzYzgamCwA8j9seA1uZOYqBnb9T0V/GA2iZw/R2sC
# fYJOv4pNf075Px6GNUBag68WHABPGtgJeuTBM2k4C3VgWn/ttzP5It/8Fcl4X8ur
# 8jwk8HJtNOIkqTi0cqZTPJSIAA8hak+G8NO5XiqMwRnU3ih6yTIeMrsKvl+9JD6D
# pxnftSpEvXN6pGiiH6sqmcloi8VOMcZRINRALN/cZCz9rS8Qad6S+0n1eg==
# SIG # End signature block
