﻿<#
.SYNOPSIS
   Removes a Microsoft HPC Pack cluster from Microsoft Azure
.DESCRIPTION
   This script removes a Microsoft HPC cluster from Microsoft Azure

   Author :  Microsoft HPC Pack team
   Version:  4.4.0
.EXAMPLE
   .\Remove-HPCIaasCluster.ps1 -ConfigFile ".\HPCIaaSCluster.xml"
.EXAMPLE
   .\Remove-HPCIaasCluster.ps1 -ConfigFile ".\HPCIaaSCluster.xml" -KeepImage
#>
[CmdletBinding(SupportsShouldProcess = $true, ConfirmImpact = "High")]
Param
(
    [Parameter(Mandatory=$true)]
    [String] $ConfigFile,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepImage,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepVNet,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepAffinityGroup,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepStorageAccount,

    [Parameter(Mandatory=$false)]
    [Switch] $Force,

    [Parameter(Mandatory=$false)]
    [ValidateRange(0, 900)]
    [Int] $SecondsWaitingForDiskDeletion = 180
)

# Check the Windows Azure Powershell version
$azureModule = Get-Module -ListAvailable -Name Azure 
if ($null -eq $azureModule)
{
    throw "Azure Powershell not found. Install the latest version from http://azure.microsoft.com/downloads/#cmd-line-tools"
}
if ($azureModule.Version -lt "0.8.10")
{
    $ver = $azureModule.Version
    throw "The Azure Powershell Version $ver is too old. Install the latest version from http://azure.microsoft.com/downloads/#cmd-line-tools"
}

$removeImage = -not $KeepImage.IsPresent
$removeVNet = -not $KeepVNet.IsPresent
$removeAG = -not $KeepAffinityGroup.IsPresent
$removeStoreAcc = -not $KeepStorageAccount.IsPresent

"Checking config file"
$clusterConfig = [xml](Get-Content $ConfigFile -ErrorAction Stop)
$iaasConfig = $clusterConfig.IaaSClusterConfig

if($null -ne $iaasConfig.Subscription["SubscriptionName"])
{
    $subName = $iaasConfig.Subscription.SubscriptionName.Trim()
    $subId = $null
}
else
{
    $subId = $iaasConfig.Subscription.SubscriptionId.Trim()
    $subName = $null
}

$storeAcc = $iaasConfig.Subscription.StorageAccount
$AGName = $iaasConfig.AffinityGroup
$VNetName = $iaasConfig.VNet.VNetName

$roleList = @()
$DNSName = $null

$newDC =  'NewDC' -eq $iaasConfig.Domain.DCOption
$headNodeAsDC = 'HeadNodeAsDC' -eq $iaasConfig.Domain.DCOption
if ($newDC)
{
    # used to remove the DNS server setting
    $DNSName = $iaasConfig.Domain.DomainController.VMName

    $DCSvc = $iaasConfig.Domain.DomainController.ServiceName
    $DCAG = $iaasConfig.Domain.DomainController.AffinityGroup
    $roleList += @{ RoleName = "Domain Controller"; Svcs = @($DCSvc); AG = $DCAG }
}
elseif ($headNodeAsDC)
{
    # HeadNodeAsDC also adds DNS server setting
    $DNSName = $iaasConfig.HeadNode.VMName
}

$newDB = 'NewRemoteDB' -eq $iaasConfig.Database.DBOption
if ($newDB)
{
    $DBSvc = $iaasConfig.Database.DBServer.ServiceName
    $DBAG = $iaasConfig.Database.DBServer.AffinityGroup
    $roleList += @{ RoleName = "DB Server"; Svcs = @($DBSvc); AG = $DBAG }
}

$HNSvc = $iaasConfig.HeadNode.ServiceName
$HNAG = $iaasConfig.HeadNode.AffinityGroup
$roleList += @{ RoleName = "Head Node"; Svcs = @($HNSvc); AG = $HNAG }

$hasCN = $null -ne $iaasConfig.ComputeNodes -and 0 -ne $iaasConfig.ComputeNodes.NodeCount
if ($hasCN)
{
    $CNAG = $iaasConfig.ComputeNodes.AffinityGroup
    $CNHPCInstalled = ($null -ne $iaasConfig.ComputeNodes.ImageName) -and ($iaasConfig.ComputeNodes.ImageName.HPCPackInstalled -eq $true)
    if ($CNHPCInstalled -eq $true)
    {
        $CNImageName = $iaasConfig.ComputeNodes.ImageName.InnerText.Trim()
    }

    $CNSvc = $iaasConfig.ComputeNodes.ServiceName
    if ($null -ne $CNSvc)
    {
        # single service
        $roleList += @{ RoleName = "Compute Node"; Svcs = @($CNSvc); AG = $CNAG }
        if ($CNHPCInstalled -eq $true)
        {
            # only if the image was copied to storeAcc by the script
            $CNImagePattern = "${CNImageName}_$storeAcc"
        }
    }
    else
    {
        # service pattern
        $count = [Math]::Ceiling($iaasConfig.ComputeNodes.NodeCount / $iaasConfig.ComputeNodes.MaxNodeCountPerService)

        $svcPattern = $iaasConfig.ComputeNodes.ServiceNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $svcRoot = $svcPattern[0]
        $svcStartsFrom = [Int]::Parse($svcPattern[1])
        $svcDigitWidth = $svcPattern[1].Length

        $storePattern = $iaasConfig.ComputeNodes.StorageAccountNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $storeRoot = $storePattern[0]
        $storeStartsFrom = [Int]::Parse($storePattern[1])
        $storeDigitWidth = $storePattern[1].Length

        if ($CNHPCInstalled -eq $true)
        {
            if ($CNImageName -match "_$storeRoot[0-9]{$storeDigitWidth}$")
            {
                $CNImagePattern = $CNImageName.Substring(0, $CNImageName.Length - $storeDigitWidth)
            }
            else
            {
                $CNImagePattern = "${CNImageName}_$storeRoot"
            }
        }

        $CNSvcs = @()
        $CNStores = @()

        for ($i = 0; $i -lt $count; $i++)
        {
            $CNSvcs += "$svcRoot$(($svcStartsFrom + $i).ToString("D$svcDigitWidth"))"
            $CNStores += "$storeRoot$(($storeStartsFrom + $i).ToString("D$storeDigitWidth"))"
        }

        $roleList += @{ RoleName = "Compute Node"; Svcs = $CNSvcs; Stores = $CNStores; AG = $CNAG }
    }
}

$hasBN = $null -ne $iaasConfig.BrokerNodes -and 0 -ne $iaasConfig.BrokerNodes.NodeCount
if ($hasBN)
{
    $BNHPCInstalled = ($null -ne $iaasConfig.BrokerNodes.ImageName) -and ($iaasConfig.BrokerNodes.ImageName.HPCPackInstalled -eq $true)
    if ($BNHPCInstalled -eq $true)
    {
        $BNImageName = $iaasConfig.BrokerNodes.ImageName.InnerText.Trim()
        $BNImagePattern = "${BNImageName}_$storeAcc"
    }

    $BNSvc = $iaasConfig.BrokerNodes.ServiceName
    $BNAG = $iaasConfig.BrokerNodes.AffinityGroup
    $roleList += @{ RoleName = "Broker Node"; Svcs = @($BNSvc); AG = $BNAG }
}

$hasLN = $null -ne $iaasConfig.LinuxComputeNodes -and 0 -ne $iaasConfig.LinuxComputeNodes.NodeCount
if ($hasLN)
{
    $LNAG = $iaasConfig.LinuxComputeNodes.AffinityGroup

    $LNSvc = $iaasConfig.LinuxComputeNodes.ServiceName
    if ($null -ne $LNSvc)
    {
        # single service
        $roleList += @{ RoleName = "Linux Compute Node"; Svcs = @($LNSvc); AG = $LNAG }
    }
    else
    {
        # service pattern
        $count = [Math]::Ceiling($iaasConfig.LinuxComputeNodes.NodeCount / $iaasConfig.LinuxComputeNodes.MaxNodeCountPerService)

        $svcPattern = $iaasConfig.LinuxComputeNodes.ServiceNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $svcRoot = $svcPattern[0]
        $svcStartsFrom = [Int]::Parse($svcPattern[1])
        $svcDigitWidth = $svcPattern[1].Length

        $storePattern = $iaasConfig.LinuxComputeNodes.StorageAccountNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $storeRoot = $storePattern[0]
        $storeStartsFrom = [Int]::Parse($storePattern[1])
        $storeDigitWidth = $storePattern[1].Length

        $LNSvcs = @()
        $LNStores = @()

        for ($i = 0; $i -lt $count; $i++)
        {
            $LNSvcs += "$svcRoot$(($svcStartsFrom + $i).ToString("D$svcDigitWidth"))"
            $LNStores += "$storeRoot$(($storeStartsFrom + $i).ToString("D$storeDigitWidth"))"
        }

        $roleList += @{ RoleName = "Linux Compute Node"; Svcs = $LNSvcs; Stores = $LNStores; AG = $LNAG }
    }
}

"Scanning Azure deployment"
if([String]::IsNullOrEmpty($subName))
{
    Select-AzureSubscription -SubscriptionId $subId | Out-Null
}
else
{
    Select-AzureSubscription -SubscriptionName $subName | Out-Null
}

$listAGs = @()
$listSvcs = @()
$listImgs = @()
$listStores = @()
$VNetExisting = $false
$DNSExisting = $false

$createdSvcs = @(Get-AzureService | % { $_.ServiceName })
$createdStores = $null
if ($removeStoreAcc) { $createdStores = @(Get-AzureStorageAccount -WarningAction SilentlyContinue | % { $_.StorageAccountName }) }
$createdAGs = $null
if ($removeAG) { $createdAGs = @(Get-AzureAffinityGroup | % { $_.Name }) }

foreach ($role in $roleList)
{
    foreach ($svcName in $role.Svcs)
    {
        if ($svcName -notin $createdSvcs)
        {
            "$($role.RoleName)'s Cloud Service $svcName does not exist"
        }
        else
        {
            if ($listSvcs -notcontains $svcName)
            {
                $listSvcs += $svcName
            }
        }
    }

    if ($removeStoreAcc)
    {
        foreach ($storeName in $role.Stores)
        {
            if ($storeName -notin $createdStores)
            {
                "$($role.RoleName)'s Storage Account $storeName does not exist"
            }
            else
            {
                if ($listStores -notcontains $storeName)
                {
                    $listStores += $storeName
                }
            }
        }
    }

    if ($removeAG -and -not [string]::IsNullOrEmpty($role.AG))
    {
        if ($role.AG -notin $createdAGs)
        {
            "$($role.RoleName)'s Affinity Group $($role.AG) does not exist"
        }
        else
        {
            if ($listAGs -notcontains $role.AG)
            {
                $listAGs += $role.AG
            }
        }
    }
}

if ($removeImage)
{
    $imgs = Get-AzureVMImage
    if($hasCN)
    {
        if ($CNHPCInstalled -eq $true)
        {
            if ($null -ne $CNSvc)
            {
                # $pattern contains the expected full name in single service case
                $ReplicatedCNImage = $imgs | ? { $_.ImageName -eq $CNImagePattern -and [string]::IsNullOrWhiteSpace($_.ServiceName) }
                if ($null -ne $ReplicatedCNImage)
                {
                    $listImgs += $ReplicatedCNImage.ImageName
                }
            }
            else
            {
                # $pattern contains such foamrt: hpcInstalledImage_stgroot (no suffix digits)
                $ReplicatedCNImages = @($imgs | ? { $_.ImageName.StartsWith($CNImagePattern) -and [string]::IsNullOrWhiteSpace($_.ServiceName) })
                $listImgs += $ReplicatedCNImages | % { $_.ImageName } | ? { $_ -ne $CNImageName }
            }
        }
        else
        {
            $svcList = @()
            if ($null -ne $CNSvc)
            {
                $svcList += $CNSvc
            }
            else
            {
                $svcList += $CNSvcs
            }

            $rootImages = @($imgs | ? { $_.ImageName -like "HPC-Pack-ComputeNode_*" -and $_.ServiceName -in $svcList })

            if ($rootImages.Count -eq 0)
            {
                "Compute Node Image does not exist"
            }
            else
            {
                $svcsHostingMultiImages = @($rootImages | % { $_.ServiceName } | Group-Object | ? { $_.Count -gt 1 })
                if ($svcsHostingMultiImages.Count -gt 0)
                {
                    Write-Warning "More than one Compute Node Images captured from the same Cloud Service are found. (e.g. from $($svcsHostingMultiImages[0].Name))"
                    Write-Warning "Skipping Compute Node Image"
                }
                else
                {
                    $imgNames = @($rootImages | % { $_.ImageName })
                    foreach ($name in $imgNames)
                    {
                        $listImgs += $imgs | % { $_.ImageName } | ? { $_.StartsWith($name) }
                    }
                }
            }
        }
    }

    if($hasBN)
    {
        if ($BNHPCInstalled -eq $true)
        {
            $ReplicatedBNImage = $imgs | ? { $_.ImageName -eq $BNImagePattern -and [string]::IsNullOrWhiteSpace($_.ServiceName) }
            if ($null -ne $ReplicatedBNImage)
            {
                $listImgs += $ReplicatedBNImage.ImageName
            }
        }
        else
        {
            $BNImage = @($imgs | ? { $_.ImageName -like "HPC-Pack-BrokerNode_*" -and $_.ServiceName -eq $BNSvc })
            if ($BNImage.Count -gt 1)
            {
                Write-Warning "More than one Broker Node Images are found. Skipping Broker Node Image"
            }
            elseif ($BNImage.Count -eq 0)
            {
                "Broker Node Image does not exist"
            }
            else
            {
                $listImgs += $BNImage[0].ImageName
            }
        }
    }
}

if ($removeStoreAcc)
{
    if ($storeAcc -notin $createdStores)
    {
        "Config-wide Storage Account $storeAcc does not exist"
    }
    else
    {
        if ($listStores -notcontains $storeAcc)
        {
            $listStores += $storeAcc
        }
    }
}

if ($removeAG -and -not [string]::IsNullOrEmpty($AGName))
{
    if ($AGName -notin $createdAGs)
    {
        "Config-wide Affinity Group $AGName does not exist"
    }
    else
    {
        if ($listAGs -notcontains $AGName)
        {
            $listAGs += $AGName
        }
    }
}

if ($removeVNet -or $null -ne $DNSName)
{
    $VNConfig = [xml](Get-AzureVNetConfig).XMLConfiguration
}
if ($removeVNet)
{
    $sites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
    if ($null -ne $sites)
    {
        $targetSite = $sites.ChildNodes | ? { $_.Name -eq $VNetName }
        if ($null -eq $targetSite)
        {
            "Virtual Network $VNetName does not exist"
        }
        else
        {
            $VNetExisting = $true
        }
    }
}

if ($null -ne $DNSName)
{
    $dnsServers = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers
    if ($null -ne $dnsServers)
    {
        $targetDns = $dnsServers.ChildNodes | ? { $_.Name -eq $DNSName }
        if ($null -eq $targetDns)
        {
            "DNS Server setting for $DNSName does not exist"
        }
        else
        {
            $ns = New-Object System.Xml.XmlNamespaceManager($VNConfig.NameTable)
            $ns.AddNamespace("ns", $VNConfig.NetworkConfiguration.NamespaceURI)
            $sitesUsingThisDNS = @($VNConfig.SelectNodes("//ns:NetworkConfiguration/ns:VirtualNetworkConfiguration/ns:VirtualNetworkSites/ns:VirtualNetworkSite[ns:DnsServersRef/ns:DnsServerRef/@name='$DNSName']", $ns)) | % { $_.name }
            $okToRemoveDNS = $true
            foreach ($siteName in $sitesUsingThisDNS)
            {
                if ($siteName -ne $VNetName)
                {
                    $okToRemoveDNS = $false
                    Write-Warning "DNS Server $DNSName (Domain Controller of this cluster) is referenced by non-HPC VNet(s). (e.g. $siteName)"
                    Write-Warning "Skipping DNS Server setting"
                    break;
                }
            }
            if ($okToRemoveDNS)
            {
                $DNSExisting = $true
            }
        }
    }
}

if (-not $VNetExisting -and -not $DNSExisting -and $listStores.Count -eq 0 -and $listImgs.Count -eq 0 -and $listAGs.Count -eq 0 -and $listSvcs.Count -eq 0)
{
    "All the resources have already been removed"
    return
}

"The following Azure resource(s) will be removed:"
if ($listSvcs.Count -gt 0)
{
    "Cloud Service: $($listSvcs -join ',')"
}
if ($listImgs.Count -gt 0)
{
    "Image: $($listImgs -join ',')"
}
if ($listAGs.Count -gt 0)
{
    "Affinity Group: $($listAGs -join ',')"
}
if ($listStores.Count -gt 0)
{
    "Storage Account: $($listStores -join ',')"
}
if ($VNetExisting)
{
    "Virtual Network: $VNetName"
}
if ($DNSExisting)
{
    "DNS Server Setting: $DNSName"
}

if (-not $Force.IsPresent)
{
    if ($PSCmdlet.ShouldProcess("The HPC Cluster will be removed from Azure", "Continue?", "Confirm:") -eq $false)
    {
        "Remove operation aborted"
        return
    }
}

if ($listSvcs.Count -gt 0)
{
    "Removing Cloud Services"
    $listSvcs | % { Remove-AzureService -ServiceName $_ -Force -DeleteAll -ErrorAction Stop | Out-Null }

    "Checking Cloud Services Removed"
    $timeout = 900
    $beginTime = [DateTime]::Now
    while ($true)
    {
        if (([DateTime]::Now - $beginTime).TotalSeconds -ge $timeout)
        {
            throw "Failed to remove Cloud Services"
        }
    
        $remainingSvcs = @(Get-AzureService | % { $_.ServiceName })

        Start-Sleep -Seconds 15
        $listSvcs | % { if ($svc -in $remainingSvcs) { continue } }

        break
    }
}

if ($listImgs.Count -gt 0)
{
    "Removing VM Images"
    $listImgs | % { Remove-AzureVMImage -ImageName $_ -DeleteVHD -ErrorAction Stop | Out-Null }
}

if (($listSvcs.Count -gt 0 -or $listImgs.Count -gt 0) -and $SecondsWaitingForDiskDeletion -gt 0)
{
    "Waiting $SecondsWaitingForDiskDeletion seconds for Azure to delete the Services/Images' associated vhds"
    Start-Sleep -Seconds $SecondsWaitingForDiskDeletion
}

"Checking Azure Disks"
$retryLimit = 5
$retryCount = 0

while ($retryCount -lt $retryLimit)
{
    $allDisks = Get-AzureDisk
    $listDisks = @()

    if ($newDC)
    {
        $DCDisks = @($allDisks | ? { $_.DiskName.StartsWith("$DCSvc-") })
        if ($DCDisks.Count -gt 0)
        {
            $listDisks += $DCDisks
        }
    }

    if ($newDB)
    {
        $DBDisks = @($allDisks | ? { $_.DiskName.StartsWith("$DBSvc-") })
        if ($DBDisks.Count -gt 0)
        {
            $listDisks += $DBDisks
        }
    }

    $HNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$HNSvc-") })
    if ($HNDisks.Count -gt 0)
    {
        $listDisks += $HNDisks
    }

    if ($hasCN)
    {
        $svcList = @()
        if ($null -ne $CNSvc)
        {
            $svcList += $CNSvc
        }
        else
        {
            $svcList += $CNSvcs
        }

        foreach ($svc in $svcList)
        {
            $CNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$svc-") })
            if ($CNDisks.Count -gt 0)
            {
                $listDisks += $CNDisks
            }
        }
    }

    if ($hasBN)
    {
        $BNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$BNSvc-") })
        if ($BNDisks.Count -gt 0)
        {
            $listDisks += $BNDisks
        }
    }
    
    if ($hasLN)
    {
        $svcList = @()
        if ($null -ne $LNSvc)
        {
            $svcList += $LNSvc
        }
        else
        {
            $svcList += $LNSvcs
        }

        foreach ($svc in $svcList)
        {
            $LNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$svc-") })
            if ($LNDisks.Count -gt 0)
            {
                $listDisks += $LNDisks
            }
        }
    }

    if ($listDisks.Count -eq 0)
    {
        break
    }

    Start-Sleep -Seconds 30
    $retryCount++
}

if ($listDisks.Count -gt 0)
{
    "Removing Azure Disks"
    $listDisks | Remove-AzureDisk -DeleteVHD -ErrorAction Stop | Out-Null
    "Waiting $SecondsWaitingForDiskDeletion seconds for Azure to delete the remaining VHDs detected in the last step"
    Start-Sleep -Seconds $SecondsWaitingForDiskDeletion
}

$storesFailedToRemove = @()

if ($listStores.Count -gt 0)
{
    "Removing Storage Account"
    foreach ($store in $listStores)
    {
        Remove-AzureStorageAccount -StorageAccountName $store | Out-Null
        if (-not $?)
        {
            $storesFailedToRemove += $store
        }
    }
}

$agsFailedToRemove = @()

if ($listAGs.Count -gt 0)
{
    "Removing Affinity Groups"
    foreach ($agToRemove in $listAGs)
    {
        Remove-AzureAffinityGroup -Name $agToRemove | Out-Null
        if (-not $?)
        {
            $agsFailedToRemove += $agToRemove
        }
    }
}

$netCfgUpdated = $true

if ($DNSExisting -or $VNetExisting)
{
    "Retrieving Azure Virtual Network Configuration"
    $VNConfig = [xml](Get-AzureVNetConfig).XMLConfiguration

    if ($DNSExisting)
    {
        "Removing DNS Server settings from the configuration"
        $dnsServers = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers
        $targetDns = $dnsServers.ChildNodes | ? { $_.Name -eq $DNSName }
        $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers.RemoveChild($targetDns) | Out-Null

        if ($VNetExisting -eq $false)
        {
            $vNetSites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
            if ($null -ne $vNetSites)
            {
                $dnsSite = $vNetSites.ChildNodes | ? { $_.Name -eq $VNetName }
                if ($null -ne $dnsSite -and $null -ne $dnsSite.DnsServersRef)
                {
                    $refedDns = $dnsSite.DnsServersRef.ChildNodes | ? { $_.Name -eq $DNSName }
                    if ($null -ne $refedDns)
                    {
                        $dnsSite.DnsServersRef.RemoveChild($refedDns) | Out-Null
                    }
                }
            }
        }
    }

    if ($VNetExisting)
    {
        "Removing Virtual Network from the configuration"
        $sites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
        $targetSite = $sites.ChildNodes | ? { $_.Name -eq $VNetName }
        $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites.RemoveChild($targetSite) | Out-Null
    }

    $fileName = "$PSScriptRoot\RemoveCluster_NetCfg_$(Get-Random -Minimum 100000 -Maximum 1000000).xml"
    "Temp NetCfg file is saved to $fileName"
    $VNConfig.Save($fileName)
    "Updating the configuration to Azure"
    Set-AzureVNetConfig -ConfigurationPath $fileName | Out-Null
    if (-not $?)
    {
        $netCfgUpdated = $false
    }
    Remove-Item -Path $fileName
}

if ($storesFailedToRemove.Count -gt 0 -or $agsFailedToRemove.Count -gt 0 -or $netCfgUpdated -eq $false)
{
    Write-Warning "Errors occured while removing the following resources:"
    if ($storesFailedToRemove.Count -gt 0) { Write-Warning "Storage Account: $($storesFailedToRemove -join ',')" }
    if ($agsFailedToRemove.Count -gt 0) { Write-Warning "Affinity Group: $($agsFailedToRemove -join ',')" }
    if (-not $netCfgUpdated) { Write-Warning "VNet and/or DNS settings" }
    Write-Warning "Check whether they are held by other Azure resources (e.g. images) and try again"
}

"Finished removing IaaS cluster"
# SIG # Begin signature block
# MIIaxwYJKoZIhvcNAQcCoIIauDCCGrQCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUijJUPCdI/Sog0Ui+cN6IgZhy
# vj6gghWCMIIEwzCCA6ugAwIBAgITMwAAAHJp9pDCFDtJoQAAAAAAcjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUwMzIwMTczMjA0
# WhcNMTYwNjIwMTczMjA0WjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OjMxQzUtMzBCQS03QzkxMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoxPVJ2iBlQVE
# sJuqKGMDs9En3ra2ZyP2mieV2SfpWI7AXCdUfRHZ7B+712Ff5z/jWtu5iNakH+yw
# OsPZy911xkYMU2T9mm1IXZcHoKEMvj6SL5H3q2dq0SX5n3iqXJcxf4wwjDu9cKit
# 5v8t+rrQxeVIuseUK34MXOH/NYpAv1H0FU0vm0MjOh7ltNRzua86yAgrQBSj7gML
# NybNLvq+M9b3JBVF6cFLmII8syS9LSf7drlwWzPRJ71Q/6+TVq59UtTQO217Juvk
# 9BdtStzFmMyR+7VdVxybCwxPr6dW/2U7oouf4UxtszHDMtHw7XEnA2jg2iq6ODkZ
# frNXUQCMhQIDAQABo4IBCTCCAQUwHQYDVR0OBBYEFND9obr0IhS4jv6FL059LwBC
# pwFTMB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZI
# hvcNAQEFBQADggEBACJEd3U6Z3Bw2w+djrTKkrwYGl8Ys2vdoryZDDm9ybp6jjCk
# QCKLBGssgmRVJcEaYjUTkqGomLk5Mt4r5X6S+Zf1lgF3IKX9tqVbNlX1qMmRZUQ4
# 4Xfg7Ml8SuB1UqFU0DvuoIoQndg51sX2yA4Tahq0xD/ilyga1NHQ3jgbw5QKBZDt
# VbGGwfUgcNCwbZGDZCQtrRTcRmr4B7/tK4xFFOfvrKLEA8MByHdzmAYlIMsPM/fX
# YCl95S60wLsCjC5aJHBzM5NN2ZRQSLYLg0RKtrBoR5Zq+RlIBXKsS0hKgRbk9WAj
# uB5MvW3T/tl40KdnBlT9lCGh677+NM8EykDf1UIwggTsMIID1KADAgECAhMzAAAB
# Cix5rtd5e6asAAEAAAEKMA0GCSqGSIb3DQEBBQUAMHkxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBMB4XDTE1MDYwNDE3NDI0NVoXDTE2MDkwNDE3NDI0NVowgYMxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIx
# HjAcBgNVBAMTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJL8bza74QO5KNZG0aJhuqVG+2MWPi75R9LH7O3HmbEm
# UXW92swPBhQRpGwZnsBfTVSJ5E1Q2I3NoWGldxOaHKftDXT3p1Z56Cj3U9KxemPg
# 9ZSXt+zZR/hsPfMliLO8CsUEp458hUh2HGFGqhnEemKLwcI1qvtYb8VjC5NJMIEb
# e99/fE+0R21feByvtveWE1LvudFNOeVz3khOPBSqlw05zItR4VzRO/COZ+owYKlN
# Wp1DvdsjusAP10sQnZxN8FGihKrknKc91qPvChhIqPqxTqWYDku/8BTzAMiwSNZb
# /jjXiREtBbpDAk8iAJYlrX01boRoqyAYOCj+HKIQsaUCAwEAAaOCAWAwggFcMBMG
# A1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSJ/gox6ibN5m3HkZG5lIyiGGE3
# NDBRBgNVHREESjBIpEYwRDENMAsGA1UECxMETU9QUjEzMDEGA1UEBRMqMzE1OTUr
# MDQwNzkzNTAtMTZmYS00YzYwLWI2YmYtOWQyYjFjZDA1OTg0MB8GA1UdIwQYMBaA
# FMsR6MrStBZYAck3LjMWFrlMmgofMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY0NvZFNpZ1BDQV8w
# OC0zMS0yMDEwLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljQ29kU2lnUENBXzA4LTMx
# LTIwMTAuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCmqFOR3zsB/mFdBlrrZvAM2PfZ
# hNMAUQ4Q0aTRFyjnjDM4K9hDxgOLdeszkvSp4mf9AtulHU5DRV0bSePgTxbwfo/w
# iBHKgq2k+6apX/WXYMh7xL98m2ntH4LB8c2OeEti9dcNHNdTEtaWUu81vRmOoECT
# oQqlLRacwkZ0COvb9NilSTZUEhFVA7N7FvtH/vto/MBFXOI/Enkzou+Cxd5AGQfu
# FcUKm1kFQanQl56BngNb/ErjGi4FrFBHL4z6edgeIPgF+ylrGBT6cgS3C6eaZOwR
# XU9FSY0pGi370LYJU180lOAWxLnqczXoV+/h6xbDGMcGszvPYYTitkSJlKOGMIIF
# vDCCA6SgAwIBAgIKYTMmGgAAAAAAMTANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMTAwODMx
# MjIxOTMyWhcNMjAwODMxMjIyOTMyWjB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALJyWVwZMGS/HZpgICBC
# mXZTbD4b1m/My/Hqa/6XFhDg3zp0gxq3L6Ay7P/ewkJOI9VyANs1VwqJyq4gSfTw
# aKxNS42lvXlLcZtHB9r9Jd+ddYjPqnNEf9eB2/O98jakyVxF3K+tPeAoaJcap6Vy
# c1bxF5Tk/TWUcqDWdl8ed0WDhTgW0HNbBbpnUo2lsmkv2hkL/pJ0KeJ2L1TdFDBZ
# +NKNYv3LyV9GMVC5JxPkQDDPcikQKCLHN049oDI9kM2hOAaFXE5WgigqBTK3S9dP
# Y+fSLWLxRT3nrAgA9kahntFbjCZT6HqqSvJGzzc8OJ60d1ylF56NyxGPVjzBrAlf
# A9MCAwEAAaOCAV4wggFaMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFMsR6MrS
# tBZYAck3LjMWFrlMmgofMAsGA1UdDwQEAwIBhjASBgkrBgEEAYI3FQEEBQIDAQAB
# MCMGCSsGAQQBgjcVAgQWBBT90TFO0yaKleGYYDuoMW+mPLzYLTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTAfBgNVHSMEGDAWgBQOrIJgQFYnl+UlE/wq4QpTlVnk
# pDBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEE
# SDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2Nl
# cnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDANBgkqhkiG9w0BAQUFAAOCAgEAWTk+
# fyZGr+tvQLEytWrrDi9uqEn361917Uw7LddDrQv+y+ktMaMjzHxQmIAhXaw9L0y6
# oqhWnONwu7i0+Hm1SXL3PupBf8rhDBdpy6WcIC36C1DEVs0t40rSvHDnqA2iA6VW
# 4LiKS1fylUKc8fPv7uOGHzQ8uFaa8FMjhSqkghyT4pQHHfLiTviMocroE6WRTsgb
# 0o9ylSpxbZsa+BzwU9ZnzCL/XB3Nooy9J7J5Y1ZEolHN+emjWFbdmwJFRC9f9Nqu
# 1IIybvyklRPk62nnqaIsvsgrEA5ljpnb9aL6EiYJZTiU8XofSrvR4Vbo0HiWGFzJ
# NRZf3ZMdSY4tvq00RBzuEBUaAF3dNVshzpjHCe6FDoxPbQ4TTj18KUicctHzbMrB
# 7HCjV5JXfZSNoBtIA1r3z6NnCnSlNu0tLxfI5nI3EvRvsTxngvlSso0zFmUeDord
# EN5k9G/ORtTTF+l5xAS00/ss3x+KnqwK+xMnQK3k+eGpf0a7B2BHZWBATrBC7E7t
# s3Z52Ao0CW0cgDEf4g5U3eWh++VHEK1kmP9QFi58vwUheuKVQSdpw5OPlcmN2Jsh
# rg1cnPCiroZogwxqLbt2awAdlq3yFnv2FoMkuYjPaqhHMS+a3ONxPdcAfmJH0c6I
# ybgY+g5yjcGjPa8CQGr/aZuW4hCoELQ3UAjWwz0wggYHMIID76ADAgECAgphFmg0
# AAAAAAAcMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAX
# BgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290
# IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0wNzA0MDMxMjUzMDlaFw0yMTA0MDMx
# MzAzMDlaMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAf
# BgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQTCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJ+hbLHf20iSKnxrLhnhveLjxZlRI1Ctzt0YTiQP7tGn
# 0UytdDAgEesH1VSVFUmUG0KSrphcMCbaAGvoe73siQcP9w4EmPCJzB/LMySHnfL0
# Zxws/HvniB3q506jocEjU8qN+kXPCdBer9CwQgSi+aZsk2fXKNxGU7CG0OUoRi4n
# rIZPVVIM5AMs+2qQkDBuh/NZMJ36ftaXs+ghl3740hPzCLdTbVK0RZCfSABKR2YR
# JylmqJfk0waBSqL5hKcRRxQJgp+E7VV4/gGaHVAIhQAQMEbtt94jRrvELVSfrx54
# QTF3zJvfO4OToWECtR0Nsfz3m7IBziJLVP/5BcPCIAsCAwEAAaOCAaswggGnMA8G
# A1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFCM0+NlSRnAK7UD7dvuzK7DDNbMPMAsG
# A1UdDwQEAwIBhjAQBgkrBgEEAYI3FQEEAwIBADCBmAYDVR0jBIGQMIGNgBQOrIJg
# QFYnl+UlE/wq4QpTlVnkpKFjpGEwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5ghB5rRahSqClrUxzWPQHEy5lMFAGA1UdHwRJ
# MEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1
# Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRIMEYwRAYIKwYB
# BQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0Um9vdENlcnQuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEB
# BQUAA4ICAQAQl4rDXANENt3ptK132855UU0BsS50cVttDBOrzr57j7gu1BKijG1i
# uFcCy04gE1CZ3XpA4le7r1iaHOEdAYasu3jyi9DsOwHu4r6PCgXIjUji8FMV3U+r
# kuTnjWrVgMHmlPIGL4UD6ZEqJCJw+/b85HiZLg33B+JwvBhOnY5rCnKVuKE5nGct
# xVEO6mJcPxaYiyA/4gcaMvnMMUp2MT0rcgvI6nA9/4UKE9/CCmGO8Ne4F+tOi3/F
# NSteo7/rvH0LQnvUU3Ih7jDKu3hlXFsBFwoUDtLaFJj1PLlmWLMtL+f5hYbMUVbo
# nXCUbKw5TNT2eb+qGHpiKe+imyk0BncaYsk9Hm0fgvALxyy7z0Oz5fnsfbXjpKh0
# NbhOxXEjEiZ2CzxSjHFaRkMUvLOzsE1nyJ9C/4B5IYCeFTBm6EISXhrIniIh0EPp
# K+m79EjMLNTYMoBMJipIJF9a6lbvpt6Znco6b72BJ3QGEe52Ib+bgsEnVLaxaj2J
# oXZhtG6hE6a/qkfwEm/9ijJssv7fUciMI8lmvZ0dhxJkAj0tr1mPuOQh5bWwymO0
# eFQF1EEuUKyUsKV4q7OglnUa2ZKHE3UiLzKoCG6gW4wlv6DvhMoh1useT8ma7kng
# 9wFlb4kLfchpyOZu6qeXzjEp/w7FW1zYTRuh2Povnj8uVRZryROj/TGCBK8wggSr
# AgEBMIGQMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAh
# BgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBAhMzAAABCix5rtd5e6as
# AAEAAAEKMAkGBSsOAwIaBQCggcgwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFMdN
# J3xJ3Q+gZf8Rx557P9uHZgrQMGgGCisGAQQBgjcCAQwxWjBYoDaANABNAGkAYwBy
# AG8AcwBvAGYAdAAgAEgAUABDACAAUABhAGMAawAgADIAMAAxADIAIABSADKhHoAc
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL0hQQzANBgkqhkiG9w0BAQEFAASCAQB4
# 9/niaYG6BbzExYzmjL9bcuvsat19eRm1BngCrM9IXul1aXIY6OXXlucwDH17ZO2l
# /Vghpp+1Yng8g3vLs6FKGgwCi4uTFelGMHaL3PGIzKNg1JNf+RL+a22Blz4aPpxE
# 8NC72zCjUB/Hx3Qxqf6lLZAteQhKqjKq776pfRo6l/mxG7Yr0wgVyxYkYq9tciQ5
# 25VJG/MVknMOEEYT5ezU/2Z7+v7+hqf1yyd3JF+cTle9UvoQxlUi/vk68jiOkITZ
# SVjKmnlw7yjOs+vibzZaP3aZzqdo3Xyy6MgGA0/gOZv8+6t/7uDkIekVveiQFclp
# iS1wzKnBrBjZFf1YKbkQoYICKDCCAiQGCSqGSIb3DQEJBjGCAhUwggIRAgEBMIGO
# MHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdS
# ZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAfBgNVBAMT
# GE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQQITMwAAAHJp9pDCFDtJoQAAAAAAcjAJ
# BgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkDMQsGCSqGSIb3DQEHATAcBgkqhkiG9w0B
# CQUxDxcNMTUwNzA2MTcyMzExWjAjBgkqhkiG9w0BCQQxFgQUWLJOPrr2CdF8MiI1
# 5yqlZgwpXgIwDQYJKoZIhvcNAQEFBQAEggEAIDBiRWWi7q8+pu7o85/HaoIuiVDt
# Q27UCMAHN0TN3WscJGXkfsIpfCx2UodoYwsqKUaF2u3JaRsLUACxmHcKgvcFu1dP
# RXC9693sluplmjmM6NJEsdv/MuYOujHG3lgjkD6TjAKkh4r03uLr74/WCYJFhp2R
# epdaLZ1SP3n/5w6UQ0x3Cl5DBT+1fetzYN4LGA88pPVhbWeXX782jmBiZnYPNVJ/
# 6FhviMHubwaFVotFy63QM2vNtjucI2ntYtLSZqW0rHdYPlCJ+9ZX+parWUMaDwwm
# HXe/X1GKXx2nedQHChDog4J//tzmvFfD+jjoXN1lcRzLo7OLGE0LZZovwg==
# SIG # End signature block
